@@ -2,18 +2,20 @@
 
 open Stdune
 open Dune_engine
+module Caml = Stdlib
 
 let config =
   { Scheduler.Config.concurrency = 1
   ; display = { verbosity = Short; status_line = false }
-  ; rpc = None
   ; stats = None
+  ; insignificant_changes = `React
+  ; signal_watcher = `No
   }
 
 let setup =
   lazy
     (Path.set_root (Path.External.cwd ());
-     Path.Build.set_build_dir (Path.Build.Kind.of_string "_build"))
+     Path.Build.set_build_dir (Path.Outside_build_dir.of_string "_build"))
 
 let prog = Option.value_exn (Bin.which ~path:(Env.path Env.initial) "true")
 
