@@ -1,30 +1,59 @@
 open Stdune
 open Import
 
+let synopsis =
+  [ `P "The installation directories used are defined by priority:"
+  ; `Noblank
+  ; `P
+      "- directories set on the command line of $(i,dune install), or \
+       corresponding environment variables"
+  ; `Noblank
+  ; `P
+      "- directories set in dune binary. They are setup before the compilation \
+       of dune with $(i,./configure)"
+  ; `Noblank
+  ; `P
+      "- inferred from the environment variable $(i,OPAM_SWITCH_PREFIX) if \
+       present"
+  ]
+
 let print_line fmt =
   Printf.ksprintf (fun s -> Console.print [ Pp.verbatim s ]) fmt
 
 let interpret_destdir ~destdir path =
   match destdir with
   | None -> path
-  | Some prefix ->
-    Path.append_local (Path.of_string prefix) (Path.local_part path)
+  | Some destdir -> Path.append_local destdir (Path.local_part path)
 
-let get_dirs context ~prefix_from_command_line ~libdir_from_command_line =
-  match prefix_from_command_line with
-  | Some p ->
-    let prefix = Path.of_string p in
-    let dir = Option.value ~default:"lib" libdir_from_command_line in
-    Fiber.return (prefix, Some (Path.relative prefix dir))
-  | None ->
-    let open Fiber.O in
-    let* prefix = Context.install_prefix context in
-    let+ libdir =
-      match libdir_from_command_line with
-      | None -> Memo.Build.run (Context.install_ocaml_libdir context)
-      | Some l -> Fiber.return (Some (Path.relative prefix l))
-    in
-    (prefix, libdir)
+let get_dirs context ~prefix_from_command_line ~from_command_line =
+  let module Roots = Install.Section.Paths.Roots in
+  let prefix_from_command_line =
+    Option.map ~f:Path.of_string prefix_from_command_line
+  in
+  let roots =
+    match prefix_from_command_line with
+    | Some prefix ->
+      Roots.opam_from_prefix prefix |> Roots.map ~f:(fun s -> Some s)
+    | None -> Context.roots context
+  in
+  let roots = Roots.first_has_priority from_command_line roots in
+  let must_be_defined name v =
+    match v with
+    | Some v -> v
+    | None ->
+      User_error.raise
+        [ Pp.textf "The %s installation directory is unknown." name ]
+        ~hints:[ Pp.textf "It could be specified with --%s" name ]
+  in
+  { Roots.lib_root = must_be_defined "libdir" roots.lib_root
+  ; libexec_root = must_be_defined "libexecdir" roots.libexec_root
+  ; bin = must_be_defined "bindir" roots.bin
+  ; sbin = must_be_defined "sbindir" roots.sbin
+  ; etc_root = must_be_defined "etcdir" roots.etc_root
+  ; doc_root = must_be_defined "docdir" roots.doc_root
+  ; share_root = must_be_defined "datadir" roots.share_root
+  ; man = must_be_defined "mandir" roots.man
+  }
 
 module Workspace = struct
   type t =
@@ -33,8 +62,8 @@ module Workspace = struct
     }
 
   let get () =
-    let open Memo.Build.O in
-    Memo.Build.run
+    let open Memo.O in
+    Memo.run
       (let+ conf = Dune_rules.Dune_load.load ()
        and+ contexts = Context.DB.all () in
        { packages = conf.packages; contexts })
@@ -47,7 +76,8 @@ module Workspace = struct
       let dir = Package.dir p in
       Ok
         (Path.Source.relative dir
-           (Dune_engine.Utils.install_file ~package:name ~findlib_toolchain:None))
+           (Dune_rules.Install_rules.install_file ~package:name
+              ~findlib_toolchain:None))
 end
 
 let resolve_package_install workspace pkg =
@@ -84,6 +114,10 @@ module Special_file = struct
     | _ -> None
 end
 
+type rmdir_mode =
+  | Fail
+  | Warn
+
 (** Operations that act on real files or just pretend to (for --dry-run) *)
 module type File_operations = sig
   val copy_file :
@@ -99,7 +133,7 @@ module type File_operations = sig
 
   val remove_file_if_exists : Path.t -> unit
 
-  val remove_dir_if_exists_and_empty : Path.t -> unit
+  val remove_dir_if_exists : if_non_empty:rmdir_mode -> Path.t -> unit
 end
 
 module type Workspace = sig
@@ -120,8 +154,11 @@ module File_ops_dry_run : File_operations = struct
   let remove_file_if_exists path =
     print_line "Removing (if it exists) %s" (Path.to_string_maybe_quoted path)
 
-  let remove_dir_if_exists_and_empty path =
-    print_line "Removing directory (if empty) %s"
+  let remove_dir_if_exists ~if_non_empty path =
+    print_line "Removing directory (%s if not empty) %s"
+      (match if_non_empty with
+      | Fail -> "fail"
+      | Warn -> "warn")
       (Path.to_string_maybe_quoted path)
 end
 
@@ -150,11 +187,11 @@ module File_ops_real (W : Workspace) : File_operations = struct
           let* packages =
             match Package.Name.Map.find workspace.packages package with
             | None -> Fiber.return None
-            | Some package -> Memo.Build.run (get_vcs (Package.dir package))
+            | Some package -> Memo.run (get_vcs (Package.dir package))
           in
           match packages with
           | None -> Fiber.return None
-          | Some vcs -> Memo.Build.run (Dune_engine.Vcs.describe vcs)
+          | Some vcs -> Memo.run (Dune_engine.Vcs.describe vcs)
         else Fiber.return None
       in
       let ppf = Format.formatter_of_out_channel oc in
@@ -162,6 +199,7 @@ module File_ops_real (W : Workspace) : File_operations = struct
       Format.pp_print_flush ppf ();
       Fiber.return ()
     | None -> plain_copy ()
+    (* XXX should we really be catching everything here? *)
     | exception _ ->
       User_warning.emit ~loc:(Loc.in_file src)
         [ Pp.text
@@ -265,27 +303,30 @@ module File_ops_real (W : Workspace) : File_operations = struct
   let copy_file ~src ~dst ~executable ~special_file ~package
       ~(conf : Dune_rules.Artifact_substitution.conf) =
     let chmod = if executable then fun _ -> 0o755 else fun _ -> 0o644 in
-    let ic, oc = Io.setup_copy ~chmod ~src ~dst () in
-    Fiber.finalize
-      ~finally:(fun () ->
-        Io.close_both (ic, oc);
-        Fiber.return ())
-      (fun () ->
-        match (special_file : Special_file.t option) with
-        | Some META -> copy_special_file ~src ~package ~ic ~oc ~f:process_meta
-        | Some Dune_package ->
-          copy_special_file ~src ~package ~ic ~oc
-            ~f:(process_dune_package ~get_location:conf.get_location)
-        | None ->
-          Dune_rules.Artifact_substitution.copy ~conf ~input_file:src
-            ~input:(input ic) ~output:(output oc))
+    match (special_file : Special_file.t option) with
+    | Some sf ->
+      let ic, oc = Io.setup_copy ~chmod ~src ~dst () in
+      Fiber.finalize
+        ~finally:(fun () ->
+          Io.close_both (ic, oc);
+          Fiber.return ())
+        (fun () ->
+          let f =
+            match sf with
+            | META -> process_meta
+            | Dune_package ->
+              process_dune_package ~get_location:conf.get_location
+          in
+          copy_special_file ~src ~package ~ic ~oc ~f)
+    | None ->
+      Dune_rules.Artifact_substitution.copy_file ~conf ~src ~dst ~chmod ()
 
   let remove_file_if_exists dst =
     if Path.exists dst then (
       print_line "Deleting %s" (Path.to_string_maybe_quoted dst);
       print_unix_error (fun () -> Path.unlink dst))
 
-  let remove_dir_if_exists_and_empty dir =
+  let remove_dir_if_exists ~if_non_empty dir =
     if Path.exists dir then
       match Path.readdir_unsorted dir with
       | Ok [] ->
@@ -294,11 +335,18 @@ module File_ops_real (W : Workspace) : File_operations = struct
         print_unix_error (fun () -> Path.rmdir dir)
       | Error (e, _, _) ->
         User_message.prerr (User_error.make [ Pp.text (Unix.error_message e) ])
-      | _ ->
-        User_error.raise
-          [ Pp.textf "Please delete non-empty directory %s manually."
-              (Path.to_string_maybe_quoted dir)
-          ]
+      | _ -> (
+        let dir = Path.to_string_maybe_quoted dir in
+        match if_non_empty with
+        | Warn ->
+          User_message.prerr
+            (User_error.make
+               [ Pp.textf "Directory %s is not empty, cannot delete (ignoring)."
+                   dir
+               ])
+        | Fail ->
+          User_error.raise
+            [ Pp.textf "Please delete non-empty directory %s manually." dir ])
 
   let mkdir_p p =
     (* CR-someday amokhov: We should really change [Path.mkdir_p dir] to fail if
@@ -321,7 +369,7 @@ module Sections = struct
     | All
     | Only of Section.Set.t
 
-  let sections_conv : Section.t list Cmdliner.Arg.converter =
+  let sections_conv =
     let all =
       Section.all |> Section.Set.to_list
       |> List.map ~f:(fun section -> (Section.to_string section, section))
@@ -353,7 +401,7 @@ let file_operations ~dry_run ~workspace : (module File_operations) =
 
 let package_is_vendored (pkg : Dune_engine.Package.t) =
   let dir = Package.dir pkg in
-  Memo.Build.run (Dune_engine.Source_tree.is_vendored dir)
+  Memo.run (Dune_engine.Source_tree.is_vendored dir)
 
 type what =
   | Install
@@ -370,52 +418,104 @@ let cmd_what = function
 let install_uninstall ~what =
   let doc = Format.asprintf "%a packages." pp_what what in
   let name_ = Arg.info [] ~docv:"PACKAGE" in
+  let absolute_path =
+    ( (fun path ->
+        if Filename.is_relative path then
+          `Error "the path must be absolute to avoid ambiguity"
+        else `Ok path)
+    , snd Arg.string )
+  in
   let term =
     let+ common = Common.term
     and+ prefix_from_command_line =
       Arg.(
         value
         & opt (some string) None
-        & info [ "prefix" ] ~docv:"PREFIX"
+        & info [ "prefix" ]
+            ~env:(Cmd.Env.info "DUNE_INSTALL_PREFIX")
+            ~docv:"PREFIX"
             ~doc:
               "Directory where files are copied. For instance binaries are \
                copied into $(i,\\$prefix/bin), library files into \
-               $(i,\\$prefix/lib), etc... It defaults to the current opam \
-               prefix if opam is available and configured, otherwise it uses \
-               the same prefix as the ocaml compiler.")
-    and+ libdir_from_command_line =
-      Arg.(
-        value
-        & opt (some string) None
-        & info [ "libdir" ] ~docv:"PATH"
-            ~doc:
-              "Directory where library files are copied, relative to \
-               $(b,prefix) or absolute. If $(b,--prefix) is specified the \
-               default is $(i,\\$prefix/lib), otherwise it is the output of \
-               $(b,ocamlfind printconf destdir)")
+               $(i,\\$prefix/lib), etc...")
     and+ destdir =
       Arg.(
         value
         & opt (some string) None
-        & info [ "destdir" ] ~env:(env_var "DESTDIR") ~docv:"PATH"
+        & info [ "destdir" ] ~env:(Cmd.Env.info "DESTDIR") ~docv:"PATH"
+            ~doc:"This directory is prepended to all installed paths.")
+    and+ libdir_from_command_line =
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "libdir" ] ~docv:"PATH"
             ~doc:
-              "When passed, this directory is prepended to all installed paths.")
-    and+ mandir =
+              "Directory where library files are copied, relative to \
+               $(b,prefix) or absolute. If $(b,--prefix) is specified the \
+               default is $(i,\\$prefix/lib). Only absolute path accepted.")
+    and+ mandir_from_command_line =
       let doc =
-        "When passed, manually override the directory to install man pages"
+        "Manually override the directory to install man pages. Only absolute \
+         path accepted."
       in
-      Arg.(value & opt (some string) None & info [ "mandir" ] ~docv:"PATH" ~doc)
-    and+ docdir =
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "mandir" ] ~docv:"PATH" ~doc)
+    and+ docdir_from_command_line =
       let doc =
-        "When passed, manually override the directory to install documentation"
+        "Manually override the directory to install documentation files. Only \
+         absolute path accepted."
       in
-      Arg.(value & opt (some string) None & info [ "docdir" ] ~docv:"PATH" ~doc)
-    and+ etcdir =
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "docdir" ] ~docv:"PATH" ~doc)
+    and+ etcdir_from_command_line =
       let doc =
-        "When passed, manually override the directory to install configuration \
-         files"
+        "Manually override the directory to install configuration files. Only \
+         absolute path accepted."
       in
-      Arg.(value & opt (some string) None & info [ "etcdir" ] ~docv:"PATH" ~doc)
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "etcdir" ] ~docv:"PATH" ~doc)
+    and+ bindir_from_command_line =
+      let doc =
+        "Manually override the directory to install public binaries. Only \
+         absolute path accepted."
+      in
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "bindir" ] ~docv:"PATH" ~doc)
+    and+ sbindir_from_command_line =
+      let doc =
+        "Manually override the directory to install files from sbin section. \
+         Only absolute path accepted."
+      in
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "sbindir" ] ~docv:"PATH" ~doc)
+    and+ datadir_from_command_line =
+      let doc =
+        "Manually override the directory to install files from share section. \
+         Only absolute path accepted."
+      in
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "datadir" ] ~docv:"PATH" ~doc)
+    and+ libexecdir_from_command_line =
+      let doc =
+        "Manually override the directory to install executable library files. \
+         Only absolute path accepted."
+      in
+      Arg.(
+        value
+        & opt (some absolute_path) None
+        & info [ "libexecdir" ] ~docv:"PATH" ~doc)
     and+ dry_run =
       Arg.(
         value & flag
@@ -427,7 +527,8 @@ let install_uninstall ~what =
         & info [ "relocatable" ]
             ~doc:
               "Make the binaries relocatable (the installation directory can \
-               be moved).")
+               be moved). The installation directory must be specified with \
+               --prefix")
     and+ create_install_files =
       Arg.(
         value & flag
@@ -446,6 +547,7 @@ let install_uninstall ~what =
               "Select context to install from. By default, install files from \
                all defined contexts.")
     and+ sections = Sections.term in
+    let common = Common.forbid_builds common in
     let config = Common.init ~log_file:No_log_file common in
     Scheduler.go ~common ~config (fun () ->
         let open Fiber.O in
@@ -473,7 +575,7 @@ let install_uninstall ~what =
                 package_is_vendored pkg >>| function
                 | true -> None
                 | false -> Some (Package.name pkg))
-            >>| List.filter_map ~f:Fun.id
+            >>| List.filter_opt
           | l -> Fiber.return l
         in
         let install_files, missing_install_files =
@@ -492,7 +594,7 @@ let install_uninstall ~what =
             ; Pp.enumerate missing_install_files ~f:(fun p ->
                   Pp.text (Path.to_string p))
             ]
-            ~hints:[ Pp.text "try running: dune build @install" ];
+            ~hints:[ Pp.text "try running: dune build [-p <pkg>] @install" ];
         (match
            (contexts, prefix_from_command_line, libdir_from_command_line)
          with
@@ -535,50 +637,52 @@ let install_uninstall ~what =
                  (context, entries_per_package))
         in
         let destdir =
-          if create_install_files then
-            Some (Option.value ~default:"_destdir" destdir)
-          else destdir
+          Option.map ~f:Path.of_string
+            (if create_install_files then
+             Some (Option.value ~default:"_destdir" destdir)
+            else destdir)
         in
+        let relocatable =
+          if relocatable then
+            match prefix_from_command_line with
+            | Some dir -> Some (Path.of_string dir)
+            | None ->
+              User_error.raise
+                [ Pp.text "Option --prefix is needed with --relocation" ]
+          else None
+        in
+
         let open Fiber.O in
         let (module Ops) = file_operations ~dry_run ~workspace in
         let files_deleted_in = ref Path.Set.empty in
+        let from_command_line =
+          let open Install.Section.Paths.Roots in
+          { lib_root = libdir_from_command_line
+          ; etc_root = etcdir_from_command_line
+          ; doc_root = docdir_from_command_line
+          ; man = mandir_from_command_line
+          ; bin = bindir_from_command_line
+          ; sbin = sbindir_from_command_line
+          ; libexec_root = libexecdir_from_command_line
+          ; share_root = datadir_from_command_line
+          }
+          |> map ~f:(Option.map ~f:Path.of_string)
+          |> complete
+        in
         let+ () =
-          let mandir =
-            Option.map ~f:Path.of_string
-              (match mandir with
-              | Some _ -> mandir
-              | None -> Dune_rules.Setup.mandir)
-          in
-          let docdir =
-            Option.map ~f:Path.of_string
-              (match docdir with
-              | Some _ -> docdir
-              | None -> Dune_rules.Setup.docdir)
-          in
-          let etcdir =
-            Option.map ~f:Path.of_string
-              (match etcdir with
-              | Some _ -> etcdir
-              | None -> Dune_rules.Setup.etcdir)
-          in
           Fiber.sequential_iter install_files_by_context
             ~f:(fun (context, entries_per_package) ->
-              let* prefix, libdir =
-                get_dirs context ~prefix_from_command_line
-                  ~libdir_from_command_line
+              let roots =
+                get_dirs context ~prefix_from_command_line ~from_command_line
               in
               let conf =
                 Dune_rules.Artifact_substitution.conf_for_install ~relocatable
                   ~default_ocamlpath:context.default_ocamlpath
-                  ~stdlib_dir:context.stdlib_dir ~prefix ~libdir ~mandir ~docdir
-                  ~etcdir
+                  ~stdlib_dir:context.stdlib_dir ~roots ~context
               in
               Fiber.sequential_iter entries_per_package
                 ~f:(fun (package, entries) ->
-                  let paths =
-                    Install.Section.Paths.make ~package ~destdir:prefix ?libdir
-                      ?mandir ?docdir ?etcdir ()
-                  in
+                  let paths = Install.Section.Paths.make ~package ~roots in
                   let+ entries =
                     Fiber.sequential_map entries ~f:(fun entry ->
                         let special_file = Special_file.of_entry entry in
@@ -593,9 +697,12 @@ let install_uninstall ~what =
                             match special_file with
                             | _ when not create_install_files ->
                               Fiber.return true
-                            | None ->
-                              Dune_rules.Artifact_substitution.test_file
-                                ~src:entry.src ()
+                            | None -> (
+                              let open Dune_rules.Artifact_substitution in
+                              let+ status = test_file ~src:entry.src () in
+                              match status with
+                              | Some_substitution -> true
+                              | No_substitution -> false)
                             | Some Special_file.META
                             | Some Special_file.Dune_package ->
                               Fiber.return true
@@ -607,9 +714,9 @@ let install_uninstall ~what =
                           if copy then
                             let* () =
                               (match Path.is_directory dst with
-                              | true -> Ops.remove_dir_if_exists_and_empty dst
-                              | false | (exception _) ->
-                                Ops.remove_file_if_exists dst);
+                              | true ->
+                                Ops.remove_dir_if_exists ~if_non_empty:Fail dst
+                              | false -> Ops.remove_file_if_exists dst);
                               print_line "%s %s" msg
                                 (Path.to_string_maybe_quoted dst);
                               Ops.mkdir_p dir;
@@ -635,9 +742,12 @@ let install_uninstall ~what =
         (* This [List.rev] is to ensure we process children directories before
            their parents *)
         |> List.rev
-        |> List.iter ~f:Ops.remove_dir_if_exists_and_empty)
+        |> List.iter ~f:(Ops.remove_dir_if_exists ~if_non_empty:Warn))
   in
-  (term, Cmdliner.Term.info (cmd_what what) ~doc ~man:Common.help_secs)
+  Cmd.v
+    (Cmd.info (cmd_what what) ~doc
+       ~man:Manpage.(`S s_synopsis :: (synopsis @ Common.help_secs)))
+    term
 
 let install = install_uninstall ~what:Install
 
