@@ -13,18 +13,19 @@ let send_shutdown cli =
   | Error e -> raise (Dune_rpc_private.Version_error.E e)
 
 let exec common =
-  let where = Rpc.wait_for_server common in
-  Dune_rpc_impl.Run.client where
+  let open Fiber.O in
+  let where = Rpc.active_server common in
+  let* conn = Dune_rpc_impl.Client.Connection.connect_exn where in
+  Dune_rpc_impl.Client.client conn ~f:send_shutdown
     (Dune_rpc_private.Initialize.Request.create
        ~id:(Dune_rpc_private.Id.make (Sexp.Atom "shutdown_cmd")))
-    ~f:send_shutdown
 
 let info =
   let doc = "cancel and shutdown any builds in the current workspace" in
-  Term.info "shutdown" ~doc
+  Cmd.info "shutdown" ~doc
 
 let term =
   let+ (common : Common.t) = Common.term in
   Rpc.client_term common exec
 
-let command = (term, info)
+let command = Cmd.v info term
