@@ -11,32 +11,38 @@ let local_libraries =
   ; ("vendor/csexp/src", Some "Csexp", false, None)
   ; ("otherlibs/stdune", Some "Stdune", false, None)
   ; ("src/dune_graph", Some "Dune_graph", false, None)
-  ; ("src/dune_lang", Some "Dune_lang", false, None)
   ; ("vendor/incremental-cycles/src", Some "Incremental_cycles", false, None)
   ; ("src/dag", Some "Dag", false, None)
   ; ("src/fiber", Some "Fiber", false, None)
+  ; ("src/dune_console", Some "Dune_console", false, None)
   ; ("src/memo", Some "Memo", false, None)
+  ; ("src/dune_metrics", Some "Dune_metrics", false, None)
+  ; ("src/dune_digest", Some "Dune_digest", false, None)
+  ; ("src/dune_sexp", Some "Dune_sexp", false, None)
+  ; ("src/ocaml-config", Some "Ocaml_config", false, None)
+  ; ("src/ocaml", Some "Ocaml", false, None)
+  ; ("vendor/re/src", Some "Dune_re", false, None)
+  ; ("otherlibs/dune-glob/src", Some "Dune_glob", false, None)
   ; ("otherlibs/xdg", Some "Xdg", false, None)
+  ; ("otherlibs/dune-rpc/private", Some "Dune_rpc_private", false, None)
+  ; ("vendor/build_path_prefix_map/src", Some "Build_path_prefix_map", false,
+    None)
   ; ("src/dune_util", Some "Dune_util", false, None)
+  ; ("src/dune_lang", Some "Dune_lang", false, None)
   ; ("src/fiber_util", Some "Fiber_util", false, None)
   ; ("src/dune_cache_storage", Some "Dune_cache_storage", false, None)
   ; ("src/dune_cache", Some "Dune_cache", false, None)
-  ; ("vendor/re/src", Some "Dune_re", false, None)
   ; ("vendor/opam-file-format/src", None, false, None)
-  ; ("otherlibs/dune-glob/src", Some "Dune_glob", false, None)
-  ; ("src/ocaml-config", Some "Ocaml_config", false, None)
   ; ("otherlibs/action-plugin/src", Some "Dune_action_plugin", false, None)
-  ; ("src/chrome_trace", Some "Chrome_trace", false, None)
+  ; ("otherlibs/chrome-trace/src", Some "Chrome_trace", false, None)
   ; ("vendor/spawn/src", Some "Spawn", false, None)
   ; ("src/dune_stats", Some "Dune_stats", false, None)
-  ; ("src/meta_parser", Some "Dune_meta_parser", false, None)
   ; ("src/section", Some "Dune_section", false, None)
-  ; ("vendor/build_path_prefix_map/src", Some "Build_path_prefix_map", false,
-    None)
-  ; ("otherlibs/dune-rpc/private", Some "Dune_rpc_private", false, None)
+  ; ("otherlibs/site/src/private", Some "Dune_site_private", false, None)
+  ; ("src/meta_parser", Some "Dune_meta_parser", false, None)
   ; ("src/dune_rpc_server", Some "Dune_rpc_server", false, None)
   ; ("src/thread_worker", Some "Thread_worker", false, None)
-  ; ("src/ocamlc_loc", Some "Ocamlc_loc", false, None)
+  ; ("otherlibs/ocamlc_loc/src", Some "Ocamlc_loc", false, None)
   ; ("src/fsevents", Some "Fsevents", false, None)
   ; ("vendor/ocaml-inotify/src", Some "Ocaml_inotify", false, None)
   ; ("src/async_inotify_for_dune", Some "Async_inotify_for_dune", false,
@@ -60,4 +66,12 @@ let link_flags =
     ; "-cclib"
     ; "-framework CoreServices"
     ])
+  ; ("win32",
+    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
+  ; ("win64",
+    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
+  ; ("mingw",
+    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
+  ; ("mingw64",
+    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
   ]
