@@ -1,4 +1,3 @@
-open! Stdune
 open Import
 open Fiber.O
 
@@ -40,27 +39,16 @@ module Config = struct
     let console_backend t =
       match t.status_line with
       | false -> Console.Backend.dumb
-      | true -> Console.Backend.progress
+      | true -> Console.Backend.progress ()
   end
 
   type t =
     { concurrency : int
     ; display : Display.t
-    ; rpc : Dune_rpc.Where.t option
     ; stats : Dune_stats.t option
+    ; insignificant_changes : [ `Ignore | `React ]
+    ; signal_watcher : [ `Yes | `No ]
     }
-
-  let add_to_env t env =
-    match t.rpc with
-    | None -> env
-    | Some where ->
-      if true then env
-      else
-        (* This is disabled because setting DUNE_RPC in tests breaks inner dune
-           invocations. We should come up with a better design for this feature,
-           or a way to cleanly work around it and prevent "unauthorized" access
-           where the test shuts down a dune that's running it. *)
-        Dune_rpc.Where.add_to_env where env
 end
 
 type job =
@@ -113,8 +101,20 @@ module Signal = struct
   let name t = Signal.name (to_int t)
 end
 
+module Shutdown = struct
+  module Signal = Signal
+
+  module Reason = struct
+    type t =
+      | Requested
+      | Signal of Signal.t
+  end
+
+  exception E of Reason.t
+end
+
 module Thread : sig
-  val spawn : (unit -> 'a) -> unit
+  val spawn : signal_watcher:[ `Yes | `No ] -> (unit -> 'a) -> unit
 
   val delay : float -> unit
 
@@ -127,17 +127,26 @@ end = struct
       (let signos = List.map Signal.all ~f:Signal.to_int in
        ignore (Unix.sigprocmask SIG_BLOCK signos : int list))
 
-  let create =
+  let create ~signal_watcher =
     if Sys.win32 then Thread.create
     else
       (* On unix, we make sure to block signals globally before starting a
          thread so that only the signal watcher thread can receive signals. *)
       fun f x ->
-      Lazy.force block_signals;
+      let () =
+        match signal_watcher with
+        | `Yes -> Lazy.force block_signals
+        | `No -> ()
+      in
       Thread.create f x
 
-  let spawn f =
-    let (_ : Thread.t) = create f () in
+  let () =
+    Fdecl.set Console.Backend.spawn_thread (fun f ->
+        let (_ : Thread.t) = create ~signal_watcher:`Yes f () in
+        ())
+
+  let spawn ~signal_watcher f =
+    let (_ : Thread.t) = create ~signal_watcher f () in
     ()
 end
 
@@ -515,7 +524,7 @@ module Process_watcher : sig
   (** Initialize the process watcher thread. *)
   type t
 
-  val init : Event.Queue.t -> t
+  val init : signal_watcher:[ `Yes | `No ] -> Event.Queue.t -> t
 
   (** Register a new running job. *)
   val register_job : t -> job -> unit
@@ -641,7 +650,7 @@ end = struct
       wait t
     done
 
-  let init events =
+  let init ~signal_watcher events =
     let t =
       { mutex = Mutex.create ()
       ; something_is_running = Condition.create ()
@@ -650,7 +659,7 @@ end = struct
       ; running_count = 0
       }
     in
-    Thread.spawn (fun () -> run t);
+    Thread.spawn ~signal_watcher (fun () -> run t);
     t
 end
 
@@ -705,26 +714,31 @@ end = struct
         if n = 3 then sys_exit 1
     done
 
-  let init q = Thread.spawn (fun () -> run q)
+  let init q = Thread.spawn ~signal_watcher:`Yes (fun () -> run q)
 end
 
 type status =
   | (* We are not doing a build. Just accumulating invalidations until the next
        build starts. *)
-      Standing_by of
+    Standing_by of
       { invalidation : Memo.Invalidation.t
       ; saw_insignificant_changes : bool
-            (* Whether we saw build input changes that are insignificant for the
-               build. We need to track this because we still want to start a new
-               build in this case, even if we know it's going to be a no-op. We
-               do that so that RPC clients can observe that Dune reacted to the
-               change. *)
+            (* When [insignificant_changes = `Ignore], this field is always
+               false.
+
+               When [insignificant_changes = `React], we do the following:
+
+               Whether we saw build input changes that are insignificant for
+               the build. We need to track this because we still want to start
+               a new build in this case, even if we know it's going to be a
+               no-op. We do that so that RPC clients can observe that Dune
+               reacted to the change. *)
       }
   | (* Running a build *)
-      Building of Fiber_util.Cancellation.t
+    Building of Fiber.Cancel.t
   | (* Cancellation requested. Build jobs are immediately rejected in this
        state *)
-      Restarting_build of
+    Restarting_build of
       Memo.Invalidation.t
 
 module Build_outcome = struct
@@ -748,7 +762,8 @@ end
 module Alarm_clock : sig
   type t
 
-  val create : Event.Queue.t -> frequency:float -> t
+  val create :
+    signal_watcher:[ `Yes | `No ] -> Event.Queue.t -> frequency:float -> t
 
   type alarm
 
@@ -810,11 +825,11 @@ end = struct
     t.alarms <- [];
     Mutex.unlock t.mutex
 
-  let create events ~frequency =
+  let create ~signal_watcher events ~frequency =
     let t =
       { events; active = true; alarms = []; frequency; mutex = Mutex.create () }
     in
-    Thread.spawn (polling_loop t);
+    Thread.spawn ~signal_watcher (polling_loop t);
     t
 
   let sleep t duration =
@@ -833,10 +848,10 @@ end = struct
     Mutex.unlock t.mutex
 end
 
-(** All fields of [t] must be immutable. This is because we re-create [t]
-    everytime we start a new build to locally set the [cancellation] field.
-    However, all instances of [t] must share all other fields, in particular the
-    references such as [status].
+(** All fields of [t] must be immutable. This is because we re-create [t] every
+    time we start a new build to locally set the [cancel] field. However, all
+    instances of [t] must share all other fields, in particular the references
+    such as [status].
 
     Another option would be to split [t] in two records such as:
 
@@ -845,7 +860,7 @@ end
 
       type t =
         { shared : shared
-        ; cancellation : Fiber_util.Cancellation.t
+        ; cancel : Fiber.Cancel.t
         }
     ]}
 
@@ -867,7 +882,7 @@ type t =
   ; file_watcher : Dune_file_watcher.t option
   ; fs_syncs : unit Fiber.Ivar.t Dune_file_watcher.Sync_id.Table.t
   ; wait_for_build_input_change : unit Fiber.Ivar.t option ref
-  ; cancellation : Fiber_util.Cancellation.t
+  ; cancel : Fiber.Cancel.t
   }
 
 let t : t Fiber.Var.t = Fiber.Var.create ()
@@ -878,6 +893,10 @@ let t_opt () = Fiber.Var.get t
 
 let t () = Fiber.Var.get_exn t
 
+let stats () =
+  let+ t = t () in
+  t.config.stats
+
 let running_jobs_count t = Event.Queue.pending_jobs t.events
 
 let yield_if_there_are_pending_events () =
@@ -889,8 +908,7 @@ exception Build_cancelled
 
 let cancelled () = raise (Memo.Non_reproducible Build_cancelled)
 
-let check_cancelled t =
-  if Fiber_util.Cancellation.fired t.cancellation then cancelled ()
+let check_cancelled t = if Fiber.Cancel.fired t.cancel then cancelled ()
 
 let abort_if_build_was_cancelled = t_opt () >>| Option.iter ~f:check_cancelled
 
@@ -911,14 +929,14 @@ let with_job_slot f =
   let* t = t () in
   Fiber.Throttle.run t.job_throttle ~f:(fun () ->
       check_cancelled t;
-      f t.config)
+      f t.cancel t.config)
 
 (* We use this version privately in this module whenever we can pass the
    scheduler explicitly *)
 let wait_for_process t pid =
   let+ res, outcome =
-    Fiber_util.Cancellation.with_handler t.cancellation
-      ~on_cancellation:(fun () ->
+    Fiber.Cancel.with_handler t.cancel
+      ~on_cancel:(fun () ->
         Process_watcher.killall t.process_watcher Sys.sigkill;
         Fiber.return ())
       (fun () ->
@@ -961,12 +979,15 @@ let prepare (config : Config.t) ~(handler : Handler.t) =
   (* We return the scheduler in chunks to resolve the dependency cycle
      (scheduler wants to know the file_watcher, file_watcher wants to send
      events to scheduler) *)
+  let signal_watcher = config.signal_watcher in
   ( events
   , fun ~file_watcher ->
       (* The signal watcher must be initialized first so that signals are
          blocked in all threads. *)
-      Signal_watcher.init events;
-      let process_watcher = Process_watcher.init events in
+      (match signal_watcher with
+      | `Yes -> Signal_watcher.init events
+      | `No -> ());
+      let process_watcher = Process_watcher.init ~signal_watcher events in
       { status =
           (* Slightly weird initialization happening here: for polling mode we
              initialize in "Building" state, immediately switch to Standing_by
@@ -974,7 +995,7 @@ let prepare (config : Config.t) ~(handler : Handler.t) =
              "Stand_by" from the start. We can't "just" switch the initial value
              here because then the non-polling mode would run in "Standing_by"
              mode, which is even weirder. *)
-          ref (Building (Fiber_util.Cancellation.create ()))
+          ref (Building (Fiber.Cancel.create ()))
       ; job_throttle = Fiber.Throttle.create config.concurrency
       ; process_watcher
       ; events
@@ -983,18 +1004,19 @@ let prepare (config : Config.t) ~(handler : Handler.t) =
       ; file_watcher
       ; fs_syncs = Dune_file_watcher.Sync_id.Table.create 64
       ; wait_for_build_input_change = ref None
-      ; alarm_clock = lazy (Alarm_clock.create events ~frequency:0.1)
-      ; cancellation =
+      ; alarm_clock =
+          lazy (Alarm_clock.create ~signal_watcher events ~frequency:0.1)
+      ; cancel =
           (* This cancellation will never be fired, so this field could instead
              be an [option]. We use a dummy cancellation rather than an option
              to keep the code simpler. *)
-          Fiber_util.Cancellation.create ()
+          Fiber.Cancel.create ()
       } )
 
 module Run_once : sig
   type run_error =
     | Already_reported
-    | Shutdown_requested
+    | Shutdown_requested of Shutdown.Reason.t
     | Exn of Exn_with_backtrace.t
 
   (** Run the build and clean up after it (kill any stray processes etc). *)
@@ -1002,7 +1024,7 @@ module Run_once : sig
 end = struct
   type run_error =
     | Already_reported
-    | Shutdown_requested
+    | Shutdown_requested of Shutdown.Reason.t
     | Exn of Exn_with_backtrace.t
 
   exception Abort of run_error
@@ -1037,56 +1059,64 @@ end = struct
       | Some ivar ->
         Dune_file_watcher.Sync_id.Table.remove t.fs_syncs id;
         [ Fill (ivar, ()) ])
-    | Build_inputs_changed events -> (
-      let invalidation =
-        (handle_invalidation_events events : Memo.Invalidation.t)
-      in
-      let fills =
-        if Memo.Invalidation.is_empty invalidation then
-          match !(t.status) with
-          | Standing_by prev ->
-            t.status :=
-              Standing_by { prev with saw_insignificant_changes = true };
-            []
-          | _ -> []
-        else
-          match !(t.status) with
-          | Restarting_build prev_invalidation ->
-            t.status :=
-              Restarting_build
-                (Memo.Invalidation.combine prev_invalidation invalidation);
-            []
-          | Standing_by prev ->
-            t.status :=
-              Standing_by
-                { prev with
-                  invalidation =
-                    Memo.Invalidation.combine prev.invalidation invalidation
-                };
-            []
-          | Building cancellation ->
-            t.handler t.config Build_interrupted;
-            t.status := Restarting_build invalidation;
-            Fiber_util.Cancellation.fire' cancellation
-      in
-      match !(t.wait_for_build_input_change) with
-      | None -> (
-        match Nonempty_list.of_list fills with
-        | None -> iter t
-        | Some fills -> fills)
-      | Some ivar ->
-        t.wait_for_build_input_change := None;
-        Fill (ivar, ()) :: fills)
+    | Build_inputs_changed events -> build_input_change t events
     | Timer fill | Worker_task fill -> [ fill ]
     | File_system_watcher_terminated ->
       filesystem_watcher_terminated ();
       raise (Abort Already_reported)
     | Yield ivar -> [ Fill (ivar, ()) ]
-    | Shutdown signal -> (
+    | Shutdown signal ->
       got_signal signal;
-      match signal with
-      | Shutdown -> raise (Abort Shutdown_requested)
-      | _ -> raise (Abort Already_reported))
+      raise
+      @@ Abort
+           (Shutdown_requested
+              (match signal with
+              | Shutdown -> Requested
+              | Signal s -> Signal s))
+
+  and build_input_change (t : t) events =
+    let invalidation = handle_invalidation_events events in
+    let significant_changes = not (Memo.Invalidation.is_empty invalidation) in
+    let insignificant_changes =
+      match t.config.insignificant_changes with
+      | `Ignore -> false
+      | `React -> not significant_changes
+    in
+    let fills =
+      match !(t.status) with
+      | Restarting_build prev_invalidation ->
+        t.status :=
+          Restarting_build
+            (Memo.Invalidation.combine prev_invalidation invalidation);
+        []
+      | Standing_by prev ->
+        t.status :=
+          Standing_by
+            { invalidation =
+                Memo.Invalidation.combine prev.invalidation invalidation
+            ; saw_insignificant_changes =
+                prev.saw_insignificant_changes || insignificant_changes
+            };
+        []
+      | Building cancellation -> (
+        match significant_changes with
+        | false -> []
+        | true ->
+          t.handler t.config Build_interrupted;
+          t.status := Restarting_build invalidation;
+          Fiber.Cancel.fire' cancellation)
+    in
+    match
+      Nonempty_list.of_list
+      @@
+      match !(t.wait_for_build_input_change) with
+      | Some ivar when significant_changes || insignificant_changes ->
+        t.wait_for_build_input_change := None;
+        Fiber.Fill (ivar, ()) :: fills
+      | _ -> fills
+    with
+    | None -> iter t
+    | Some fills -> fills
 
   let run t f : _ result =
     let fiber =
@@ -1124,8 +1154,11 @@ module Worker = struct
   let stop t = Thread_worker.stop t.worker
 
   let create () =
-    let worker = Thread_worker.create ~spawn_thread:Thread.spawn in
     let+ scheduler = t () in
+    let worker =
+      let signal_watcher = scheduler.config.signal_watcher in
+      Thread_worker.create ~spawn_thread:(Thread.spawn ~signal_watcher)
+    in
     { worker; events = scheduler.events }
 
   let task (t : t) ~f =
@@ -1178,6 +1211,8 @@ let wait_for_build_input_change t =
 module Run = struct
   exception Build_cancelled = Build_cancelled
 
+  module Shutdown = Shutdown
+
   type file_watcher =
     | Automatic
     | No_watcher
@@ -1187,14 +1222,14 @@ module Run = struct
   module Event = Handler.Event
 
   let rec poll_iter t step ~invalidation =
-    let cancellation = Fiber_util.Cancellation.create () in
-    t.status := Building cancellation;
+    let cancel = Fiber.Cancel.create () in
+    t.status := Building cancel;
     (if Memo.Invalidation.is_empty invalidation then Memo.Perf_counters.reset ()
     else
       let details_hum = Memo.Invalidation.details_hum invalidation in
       t.handler t.config (Source_files_changed { details_hum });
       Memo.reset invalidation);
-    let* res = set { t with cancellation } (fun () -> step) in
+    let* res = set { t with cancel } (fun () -> step) in
     match !(t.status) with
     | Standing_by _ ->
       (* We just finished a build, so there's no way this was set *)
@@ -1272,8 +1307,6 @@ module Run = struct
     in
     loop ()
 
-  exception Shutdown_requested
-
   let go config ?timeout ?(file_watcher = No_watcher)
       ~(on_event : Config.t -> Handler.Event.t -> unit) run =
     let events, prepare = prepare config ~handler:on_event in
@@ -1284,7 +1317,8 @@ module Run = struct
         Some
           (Dune_file_watcher.create_default
              ~scheduler:
-               { spawn_thread = Thread.spawn
+               { spawn_thread =
+                   Thread.spawn ~signal_watcher:config.signal_watcher
                ; thread_safe_send_emit_events_job =
                    (fun job -> Event_queue.send_file_watcher_task events job)
                }
@@ -1313,7 +1347,7 @@ module Run = struct
       in
       match Run_once.run_and_cleanup t run with
       | Ok a -> Result.Ok a
-      | Error Shutdown_requested -> Error (Shutdown_requested, None)
+      | Error (Shutdown_requested reason) -> Error (Shutdown.E reason, None)
       | Error Already_reported ->
         Error (Dune_util.Report_error.Already_reported, None)
       | Error (Exn exn_with_bt) ->
@@ -1374,10 +1408,6 @@ let sleep duration =
     (* cancellation mechanism isn't exposed to the user *)
     assert false
 
-let flush_file_watcher () =
-  let* t = t () in
-  flush_file_watcher t
-
 let wait_for_build_input_change () =
   let* t = t () in
   wait_for_build_input_change t
