@@ -1,13 +1,10 @@
-open! Dune_engine
-open! Stdune
 open Import
-module Mapper = Action_mapper.Make (Action_dune_lang) (Action_dune_lang)
 
-(* So that we can comfortably use both the [Action_builder.O] and [Memo.BUild.O]
-   monad at the same time *)
-let ( let+! ) = Memo.Build.O.( let+ )
+(* So that we can comfortably use both the [Action_builder.O] and [Memo.O]
+   monads at the same time. *)
+let ( let+! ) = Memo.O.( let+ )
 
-let ( let*! ) = Memo.Build.O.( let* )
+let ( let*! ) = Memo.O.( let* )
 
 let as_in_build_dir ~what ~loc p =
   match Path.as_in_build_dir p with
@@ -19,6 +16,15 @@ let as_in_build_dir ~what ~loc p =
           (Path.to_string_maybe_quoted p)
       ]
 
+let validate_target_dir ~targets_dir ~loc targets path =
+  if Path.Build.(parent_exn path <> targets_dir) then
+    User_error.raise ~loc
+      [ Pp.text
+          "This action has targets in a different directory than the current \
+           one, this is not allowed by dune at the moment:"
+      ; Targets.pp targets
+      ]
+
 module Action_expander : sig
   (* An applicative to help write action expansion. It is similar to
      [Action_builder.With_targets.t] but with some differences. The differences
@@ -43,7 +49,10 @@ module Action_expander : sig
   val set_env : var:string -> value:string t -> (value:string -> 'a) t -> 'a t
 
   val run :
-    'a t -> expander:Expander.t -> 'a Action_builder.With_targets.t Memo.Build.t
+       'a t
+    -> targets_dir:Path.Build.t option
+    -> expander:Expander.t
+    -> 'a Action_builder.With_targets.t Memo.t
 
   (* String with vars expansion *)
   module E : sig
@@ -88,7 +97,7 @@ end = struct
   type deps = Path.Set.t Action_builder.t
 
   type collector =
-    { file_targets : Path.Build.Set.t  (** We only infer file targets *)
+    { file_targets : Loc.t Path.Build.Map.t  (** We only infer file targets *)
     ; deps : deps
     ; deps_if_exist : deps
     }
@@ -99,9 +108,9 @@ end = struct
     ; dir : Path.Build.t
     }
 
-  type 'a t = env -> collector -> ('a Action_builder.t * collector) Memo.Build.t
+  type 'a t = env -> collector -> ('a Action_builder.t * collector) Memo.t
 
-  let return x _env acc = Memo.Build.return (Action_builder.return x, acc)
+  let return x _env acc = Memo.return (Action_builder.return x, acc)
 
   let map t ~f env acc =
     let+! b, acc = t env acc in
@@ -110,13 +119,13 @@ end = struct
   let both a b env acc =
     let*! a, acc = a env acc in
     let*! b, acc = b env acc in
-    Memo.Build.return (Action_builder.both a b, acc)
+    Memo.return (Action_builder.both a b, acc)
 
   let all =
     let rec loop res l env acc =
       match l with
       | [] ->
-        Memo.Build.return
+        Memo.return
           (Action_builder.map (Action_builder.all res) ~f:List.rev, acc)
       | t :: l ->
         let*! x, acc = t env acc in
@@ -124,13 +133,13 @@ end = struct
     in
     fun l env acc -> loop [] l env acc
 
-  let run t ~expander =
+  let run t ~targets_dir ~expander =
     let deps = Action_builder.return Path.Set.empty in
     let acc =
-      { file_targets = Path.Build.Set.empty; deps; deps_if_exist = deps }
+      { file_targets = Path.Build.Map.empty; deps; deps_if_exist = deps }
     in
     let env = { expander; infer = true; dir = Expander.dir expander } in
-    Memo.Build.map (t env acc) ~f:(fun (b, acc) ->
+    Memo.map (t env acc) ~f:(fun (b, acc) ->
         let { file_targets; deps; deps_if_exist } = acc in
         (* A file can be inferred as both a dependency and a target, for
            instance:
@@ -138,7 +147,7 @@ end = struct
            {[ (progn (copy a b) (copy b c)) ]} *)
         let remove_targets =
           let file_targets =
-            Path.Build.Set.to_list file_targets
+            Path.Build.Map.keys file_targets
             |> Path.Set.of_list_map ~f:Path.build
           in
           fun deps -> Path.Set.diff deps file_targets
@@ -152,7 +161,13 @@ end = struct
               >>> Action_builder.if_file_exists f ~then_:(Action_builder.path f)
                     ~else_:(Action_builder.return ()))
         in
-        let targets = Targets.Files.create file_targets in
+        let targets =
+          let file_targets = Path.Build.Set.of_keys file_targets in
+          Targets.Files.create file_targets
+        in
+        Option.iter targets_dir ~f:(fun targets_dir ->
+            Path.Build.Map.iteri file_targets ~f:(fun path loc ->
+                validate_target_dir ~targets_dir ~loc targets path));
         Action_builder.with_targets ~targets
           (let+ () = deps >>= Action_builder.path_set
            and+ () = deps_if_exist >>= action_builder_path_set_if_exist
@@ -179,7 +194,7 @@ end = struct
 
   let set_env ~var ~value t env acc =
     let*! value, acc = value env acc in
-    let value = Action_builder.memoize "env var" value in
+    let value = Action_builder.memoize ~cutoff:String.equal "env var" value in
     let env =
       { env with
         expander = Expander.set_local_env_var env.expander ~var ~value
@@ -205,7 +220,7 @@ end = struct
     let ( >>> ) a b env acc =
       let*! a, acc = a env acc in
       let*! b, acc = b env acc in
-      Memo.Build.return (Action_builder.O.( >>> ) a b, acc)
+      Memo.return (Action_builder.O.( >>> ) a b, acc)
   end
 
   module E = struct
@@ -219,10 +234,10 @@ end = struct
         Action_builder.Expander.expand ~dir:(Path.build env.dir) ~mode template
           ~f:(Expander.expand_pform env.expander)
 
-      let expand_path env sw =
-        let+ v = expand env ~mode:Single sw in
-        Value.to_path v ~error_loc:(String_with_vars.loc sw)
-          ~dir:(Path.build env.dir)
+      let expand_path t sw =
+        let+ v = expand t ~mode:Single sw in
+        Value.to_path_in_build_or_external v
+          ~error_loc:(String_with_vars.loc sw) ~dir:t.dir
 
       let expand_string env sw =
         let+ v = expand env ~mode:Single sw in
@@ -243,8 +258,8 @@ end = struct
 
         let expand_path env sw =
           let+! v = expand env ~mode:Single sw in
-          Value.to_path v ~error_loc:(String_with_vars.loc sw)
-            ~dir:(Path.build env.dir)
+          Value.to_path_in_build_or_external v
+            ~error_loc:(String_with_vars.loc sw) ~dir:env.dir
 
         let expand_string env sw =
           let+! v = expand env ~mode:Single sw in
@@ -252,13 +267,11 @@ end = struct
       end
     end
 
-    let string sw env acc =
-      Memo.Build.return (Expander.expand_string env sw, acc)
+    let string sw env acc = Memo.return (Expander.expand_string env sw, acc)
 
-    let strings sw env acc =
-      Memo.Build.return (Expander.expand_strings env sw, acc)
+    let strings sw env acc = Memo.return (Expander.expand_strings env sw, acc)
 
-    let path sw env acc = Memo.Build.return (Expander.expand_path env sw, acc)
+    let path sw env acc = Memo.return (Expander.expand_path env sw, acc)
 
     module At_rule_eval_stage = struct
       let make ~expand sw ~f env acc =
@@ -271,7 +284,7 @@ end = struct
     end
 
     let register_dep x ~f env acc =
-      Memo.Build.return
+      Memo.return
         (if not env.infer then (x, acc)
         else
           let x = Action_builder.memoize "dep" x in
@@ -290,11 +303,13 @@ end = struct
       register_dep fn ~f:Option.some env acc
 
     let dep_if_exists sw env acc =
-      Memo.Build.return
+      Memo.return
         (let fn = Expander.expand_path env sw in
          if not env.infer then (fn, acc)
          else
-           let fn = Action_builder.memoize "dep_if_exists" fn in
+           let fn =
+             Action_builder.memoize ~cutoff:Path.equal "dep_if_exists" fn
+           in
            ( fn
            , { acc with
                deps_if_exist =
@@ -305,20 +320,22 @@ end = struct
 
     let add_or_remove_target ~what ~f sw env acc =
       if not env.infer then
-        Memo.Build.return
+        Memo.return
           ( (let+ p = Expander.expand_path env sw in
              as_in_build_dir ~what ~loc:(loc sw) p)
           , acc )
       else
         let+! p = Expander.No_deps.expand_path env sw in
-        let p = as_in_build_dir p ~what ~loc:(loc sw) in
+        let loc = loc sw in
+        let p = as_in_build_dir p ~what ~loc in
         ( Action_builder.return p
-        , { acc with file_targets = f acc.file_targets p } )
+        , { acc with file_targets = f acc.file_targets p loc } )
 
     let consume_file =
-      add_or_remove_target ~what:"File" ~f:Path.Build.Set.remove
+      add_or_remove_target ~what:"File" ~f:(fun map p _loc ->
+          Path.Build.Map.remove map p)
 
-    let target = add_or_remove_target ~what:"Target" ~f:Path.Build.Set.add
+    let target = add_or_remove_target ~what:"Target" ~f:Path.Build.Map.set
 
     let prog_and_args sw env acc =
       let b =
@@ -338,7 +355,7 @@ end = struct
             | Relative_to_current_dir | Absolute ->
               Action_builder.return (Ok (Path.relative dir s))
             | In_path ->
-              Action_builder.memo_build
+              Action_builder.of_memo
                 (Artifacts.Bin.binary ~loc:(Some loc)
                    (Expander.artifacts env.expander)
                    s))
@@ -353,7 +370,7 @@ end = struct
   end
 end
 
-let rec expand (t : Action_dune_lang.t) : Action.t Action_expander.t =
+let rec expand (t : Dune_lang.Action.t) : Action.t Action_expander.t =
   let module A = Action_expander in
   let module E = Action_expander.E in
   let open Action_expander.O in
@@ -403,9 +420,9 @@ let rec expand (t : Action_dune_lang.t) : Action.t Action_expander.t =
     let+ l = A.all (List.map xs ~f:E.strings) in
     let l = List.concat l in
     O.Echo l
-  | Cat fn ->
-    let+ fn = E.dep fn in
-    O.Cat fn
+  | Cat xs ->
+    let+ xs = A.all (List.map xs ~f:E.dep) in
+    O.Cat xs
   | Copy (x, y) ->
     let+ x = E.dep x
     and+ y = E.target y in
@@ -414,14 +431,10 @@ let rec expand (t : Action_dune_lang.t) : Action.t Action_expander.t =
     let+ x = E.dep x
     and+ y = E.target y in
     O.Symlink (x, y)
-  | Hardlink (x, y) ->
-    let+ x = E.dep x
-    and+ y = E.target y in
-    O.Hardlink (x, y)
   | Copy_and_add_line_directive (x, y) ->
     let+ x = E.dep x
     and+ y = E.target y in
-    O.Copy_and_add_line_directive (x, y)
+    Copy_line_directive.action x y
   | System x ->
     let+ x = E.string x in
     O.System x
@@ -432,30 +445,20 @@ let rec expand (t : Action_dune_lang.t) : Action.t Action_expander.t =
     let+ fn = E.target fn
     and+ s = E.string s in
     O.Write_file (fn, perm, s)
-  | Rename (x, _) ->
-    (* [Rename] is not part of the syntax so this case is not reachable. The
-       reason it is not exposed to the user is because we can't easily decide
-       what [x] is. i.e. it might be a file generated by the action and renamed
-       immediately, or it could be a dependency. *)
-    Code_error.raise ~loc:(String_with_vars.loc x)
-      "Rename is not allowed in unexpanded actions" []
-  | Remove_tree x ->
-    (* Same kind of stuff for [Remove_tree] *)
-    Code_error.raise ~loc:(String_with_vars.loc x)
-      "Remove_tree is not allowed in unexpanded actions" []
-  | Mkdir x ->
+  | Mkdir x -> (
     (* This code path should in theory be unreachable too, but we don't delete
        it to remember about the check in in case we expose [mkdir] in the syntax
        one day. *)
     let+ path = E.path x in
-    if not (Path.is_managed path) then
+    match Path.as_in_build_dir path with
+    | Some path -> O.Mkdir path
+    | None ->
       User_error.raise ~loc:(String_with_vars.loc x)
         [ Pp.text
             "(mkdir ...) is not supported for paths outside of the workspace:"
         ; Pp.seq (Pp.verbatim "  ")
             (Dune_lang.pp (List [ Dune_lang.atom "mkdir"; Dpath.encode path ]))
-        ];
-    O.Mkdir path
+        ])
   | Diff { optional; file1; file2; mode } ->
     let+ file1 = E.dep_if_exists file1
     and+ file2 =
@@ -465,22 +468,13 @@ let rec expand (t : Action_dune_lang.t) : Action.t Action_expander.t =
         as_in_build_dir p ~loc:(String_with_vars.loc file2) ~what:"File"
     in
     O.Diff { optional; file1; file2; mode }
-  | Merge_files_into (sources, extras, target) ->
-    let+ sources = A.all (List.map sources ~f:E.dep)
-    and+ extras = A.all (List.map extras ~f:E.string)
-    and+ target = E.target target in
-    O.Merge_files_into (sources, extras, target)
   | No_infer t -> A.no_infer (expand t)
   | Pipe (outputs, l) ->
     let+ l = A.all (List.map l ~f:expand) in
     O.Pipe (outputs, l)
-  | Format_dune_file (ver, src, dst) ->
-    let+ src = E.dep src
-    and+ dst = E.target dst in
-    O.Format_dune_file (ver, src, dst)
   | Cram script ->
     let+ script = E.dep script in
-    O.Cram script
+    Cram_exec.action script
 
 let expand_no_targets t ~loc ~deps:deps_written_by_user ~expander ~what =
   let open Action_builder.O in
@@ -491,7 +485,8 @@ let expand_no_targets t ~loc ~deps:deps_written_by_user ~expander ~what =
     Expander.set_expanding_what expander (User_action_without_targets { what })
   in
   let* { Action_builder.With_targets.build; targets } =
-    Action_builder.memo_build (Action_expander.run (expand t) ~expander)
+    Action_builder.of_memo
+      (Action_expander.run (expand t) ~targets_dir:None ~expander)
   in
   if not (Targets.is_empty targets) then
     User_error.raise ~loc
@@ -524,7 +519,7 @@ let expand t ~loc ~deps:deps_written_by_user ~targets_dir
                 | One -> Target
                 | Multiple -> Targets))
              (Expander.Deps.Without
-                (Memo.Build.return
+                (Memo.return
                    (Value.L.paths
                       (List.map targets
                          ~f:(fun (target, (_ : Targets_spec.Kind.t)) ->
@@ -534,7 +529,7 @@ let expand t ~loc ~deps:deps_written_by_user ~targets_dir
     Expander.set_expanding_what expander (User_action targets_written_by_user)
   in
   let+! { Action_builder.With_targets.build; targets } =
-    Action_expander.run (expand t) ~expander
+    Action_expander.run (expand t) ~targets_dir:(Some targets_dir) ~expander
   in
   let targets =
     match (targets_written_by_user : _ Targets_spec.t) with
@@ -542,13 +537,7 @@ let expand t ~loc ~deps:deps_written_by_user ~targets_dir
     | Static { targets = targets_written_by_user; multiplicity = _ } ->
       let files, dirs =
         List.partition_map targets_written_by_user ~f:(fun (path, kind) ->
-            if Path.Build.(parent_exn path <> targets_dir) then
-              User_error.raise ~loc
-                [ Pp.text
-                    "This action has targets in a different directory than the \
-                     current one, this is not allowed by dune at the moment:"
-                ; Targets.pp targets
-                ];
+            validate_target_dir ~targets_dir ~loc targets path;
             match kind with
             | File -> Left path
             | Directory -> Right path)
@@ -565,6 +554,6 @@ let expand t ~loc ~deps:deps_written_by_user ~targets_dir
   in
   Action_builder.with_targets ~targets build
 
-(* We re-export [Action_dune_lang] in the end to avoid polluting the inferred
+(* We re-export [Dune_lang.Action] in the end to avoid polluting the inferred
    types in this module with all the various t's *)
-include Action_dune_lang
+include Dune_lang.Action
