@@ -1,10 +1,9 @@
 (** Generate bootstrap info *)
-open! Dune_engine
 
 (** Generate an OCaml file containing a description of the dune sources for the
     bootstrap procedure *)
 
-open Stdune
+open Import
 
 (** Generate the rules to handle the stanza *)
 val gen_rules :
@@ -12,4 +11,4 @@ val gen_rules :
   -> Dune_file.Executables.t
   -> dir:Path.Build.t
   -> Lib.Compile.t
-  -> unit Memo.Build.t
+  -> unit Memo.t
