@@ -1,4 +1,3 @@
-open! Dune_engine
 open Import
 
 let dev_files =
@@ -8,7 +7,7 @@ let dev_files =
       (let open Dyn in
       variant "dev_files" (List.map ~f:string exts))
   in
-  Predicate.create ~id ~f:(fun p ->
+  Predicate_with_id.create ~id ~f:(fun p ->
       let ext = Filename.extension p in
       List.mem exts ext ~equal:String.equal)
 
@@ -21,11 +20,17 @@ let add_obj_dir sctx ~obj_dir =
     Rules.Produce.Alias.add_deps
       (Alias.check ~dir:(Obj_dir.dir obj_dir))
       (Action_builder.paths_matching_unit ~loc:(Loc.of_pos __POS__) dir_glob)
-  else Memo.Build.return ()
+  else Memo.return ()
 
 let add_files sctx ~dir files =
   if (Super_context.context sctx).merlin then
     let alias = Alias.check ~dir in
     let files = Path.Set.of_list files in
     Rules.Produce.Alias.add_deps alias (Action_builder.path_set files)
-  else Memo.Build.return ()
+  else Memo.return ()
+
+let add_cycle_check sctx ~dir modules =
+  if (Super_context.context sctx).merlin then
+    let alias = Alias.check ~dir in
+    Rules.Produce.Alias.add_deps alias (Action_builder.ignore modules)
+  else Memo.return ()
