@@ -1,5 +1,4 @@
-open! Stdune
-open! Dune_engine
+open Import
 
 let header_file_content =
   {|
@@ -19,7 +18,7 @@ CCOMP
 let rules ~sctx ~dir =
   let file = Path.Build.relative dir Cxx_flags.preprocessed_filename in
   let ocfg = (Super_context.context sctx).ocaml_config in
-  let open Memo.Build.O in
+  let open Memo.O in
   let* prog =
     Super_context.resolve_program sctx ~dir ~loc:None
       (Ocaml_config.c_compiler ocfg)
