@@ -1,6 +1,7 @@
 module Main = Main
 module Context = Context
 module Super_context = Super_context
+module Compilation_context = Compilation_context
 module Findlib = Findlib
 module Colors = Colors
 module Profile = Profile
@@ -9,13 +10,16 @@ module Dune_package = Dune_package
 module Dep_conf = Dep_conf
 module Dir_contents = Dir_contents
 module Expander = Expander
-module Watermarks = Watermarks
 module Lib = Lib
+module Lib_flags = Lib_flags
 module Lib_info = Lib_info
 module Modules = Modules
+module Module_compilation = Module_compilation
 module Exe_rules = Exe_rules
+module Lib_rules = Lib_rules
 module Obj_dir = Obj_dir
-module Merlin_server = Merlin_server
+module Merlin_ident = Merlin_ident
+module Merlin = Merlin
 module Ml_sources = Ml_sources
 module Scope = Scope
 module Module = Module
@@ -29,15 +33,32 @@ module Utop = Utop
 module Setup = Setup
 module Meta = Meta
 module Toplevel = Toplevel
+module Top_module = Top_module
 module Global = Global
 module Only_packages = Only_packages
 module Resolve = Resolve
+module Ocamldep = Ocamldep
+module Dep_rules = Dep_rules
+module Dep_graph = Dep_graph
+module Preprocess = Preprocess
+module Coq_rules = Coq_rules
+module Coq_module = Coq_module
+module Coq_sources = Coq_sources
+module Coq_stanza = Coq_stanza
+module Coq_lib = Coq_lib
+module Command = Command
+module Install = Install
+module Lib_name = Lib_name
+module Diff = Dune_lang.Action.Diff
+
+module Install_rules = struct
+  let install_file = Install_rules.install_file
+end
 
 (* Only for tests *)
 module Scheme = Scheme
 module Lib_config = Lib_config
 module Dynlink_supported = Dynlink_supported
-module Ocaml_version = Ocaml_version
 module Lib_dep = Lib_dep
 module Ocamlobjinfo = Ocamlobjinfo
 module Action_unexpanded = Action_unexpanded
