@@ -1,11 +1,6 @@
-open! Dune_engine
 open Import
 open Action_builder.O
 
-type any_package =
-  | Local of Package.t
-  | Installed of Dune_package.t
-
 module Expanding_what = struct
   type t =
     | Nothing_special
@@ -17,31 +12,31 @@ end
 module Deps = struct
   module T = struct
     type 'a t =
-      | Without of 'a Memo.Build.t
+      | Without of 'a Memo.t
       | With of 'a Action_builder.t
 
-    let return x = Without (Memo.Build.return x)
+    let return x = Without (Memo.return x)
 
     let map t ~f =
       match t with
-      | Without t -> Without (Memo.Build.map t ~f)
+      | Without t -> Without (Memo.map t ~f)
       | With t -> With (Action_builder.map t ~f)
 
     let both a b =
       match (a, b) with
-      | Without a, Without b -> Without (Memo.Build.both a b)
+      | Without a, Without b -> Without (Memo.both a b)
       | With a, With b -> With (Action_builder.both a b)
       | Without a, With b ->
-        With (Action_builder.both (Action_builder.memo_build a) b)
+        With (Action_builder.both (Action_builder.of_memo a) b)
       | With a, Without b ->
-        With (Action_builder.both a (Action_builder.memo_build b))
+        With (Action_builder.both a (Action_builder.of_memo b))
   end
 
   include T
   include Applicative.Make (T)
 
   let action_builder = function
-    | Without x -> Action_builder.memo_build x
+    | Without x -> Action_builder.of_memo x
     | With x -> x
 end
 
@@ -60,11 +55,11 @@ type t =
   ; c_compiler : string
   ; context : Context.t
   ; lookup_artifacts :
-      (dir:Path.Build.t -> Ml_sources.Artifacts.t Memo.Build.t) option
+      (dir:Path.Build.t -> Ml_sources.Artifacts.t Memo.t) option
   ; foreign_flags :
          dir:Path.Build.t
-      -> string list Action_builder.t Foreign_language.Dict.t Memo.Build.t
-  ; find_package : Package.Name.t -> any_package option Memo.Build.t
+      -> string list Action_builder.t Foreign_language.Dict.t Memo.t
+  ; sites : Sites.t
   ; expanding_what : Expanding_what.t
   }
 
@@ -109,7 +104,7 @@ let add_bindings_full t ~bindings =
 let add_bindings t ~bindings =
   add_bindings_full t
     ~bindings:
-      (Pform.Map.map bindings ~f:(fun v -> Deps.Without (Memo.Build.return v)))
+      (Pform.Map.map bindings ~f:(fun v -> Deps.Without (Memo.return v)))
 
 let path p = [ Value.Path p ]
 
@@ -132,7 +127,7 @@ let expand_version { scope; _ } ~source s =
       (Dune_project.packages project)
       (Package.Name.of_string s)
   with
-  | Some p -> Memo.Build.return (value_from_version p.version)
+  | Some p -> Memo.return (value_from_version p.version)
   | None when Dune_project.dune_version project < (2, 9) ->
     User_error.raise ~loc:source.Dune_lang.Template.Pform.loc
       [ Pp.textf "Package %S doesn't exist in the current project." s ]
@@ -151,7 +146,7 @@ let expand_version { scope; _ } ~source s =
             "Library names are not allowed in this position. Only package \
              names are allowed"
         ];
-    let open Memo.Build.O in
+    let open Memo.O in
     Lib.DB.find (Scope.libs scope) libname >>| function
     | Some lib -> value_from_version (Lib_info.version (Lib.info lib))
     | None ->
@@ -162,14 +157,14 @@ let expand_version { scope; _ } ~source s =
             s
         ])
 
-let isn't_allowed_in_this_position_message ~source =
-  User_error.make ~loc:source.Dune_lang.Template.Pform.loc
-    [ Pp.textf "%s isn't allowed in this position."
-        (Dune_lang.Template.Pform.describe source)
-    ]
-
 let isn't_allowed_in_this_position ~source =
-  raise (User_error.E (isn't_allowed_in_this_position_message ~source))
+  let exn =
+    User_error.make ~loc:source.Dune_lang.Template.Pform.loc
+      [ Pp.textf "%s isn't allowed in this position."
+          (Dune_lang.Template.Pform.describe source)
+      ]
+  in
+  raise (User_error.E exn)
 
 let expand_artifact ~source t a s =
   match t.lookup_artifacts with
@@ -181,7 +176,7 @@ let expand_artifact ~source t a s =
     let does_not_exist ~loc ~what name =
       User_error.raise ~loc [ Pp.textf "%s %s does not exist." what name ]
     in
-    let* artifacts = Action_builder.memo_build (lookup ~dir) in
+    let* artifacts = Action_builder.of_memo (lookup ~dir) in
     match a with
     | Pform.Artifact.Mod kind -> (
       let name =
@@ -195,7 +190,7 @@ let expand_artifact ~source t a s =
           ~what:"Module"
           (Module_name.to_string name)
       | Some (t, m) -> (
-        match Obj_dir.Module.cm_file t m ~kind with
+        match Obj_dir.Module.cm_file t m ~kind:(Ocaml kind) with
         | None -> Action_builder.return [ Value.String "" ]
         | Some path -> dep (Path.build path)))
     | Lib mode -> (
@@ -216,7 +211,7 @@ let expand_artifact ~source t a s =
                Value.Path fn))))
 
 let cc t =
-  Memo.Build.map (t.foreign_flags ~dir:t.dir) ~f:(fun cc ->
+  Memo.map (t.foreign_flags ~dir:t.dir) ~f:(fun cc ->
       Foreign_language.Dict.map cc ~f:(fun cc ->
           let+ flags = cc in
           strings (t.c_compiler :: flags)))
@@ -237,7 +232,7 @@ type nonrec expansion_result =
   | Direct of value
   | Need_full_expander of (t -> value)
 
-let static v = Direct (Without (Memo.Build.return v))
+let static v = Direct (Without (Memo.return v))
 
 let[@inline never] invalid_use_of_target_variable t
     ~(source : Dune_lang.Template.Pform.t) ~var_multiplicity =
@@ -265,7 +260,7 @@ let[@inline never] invalid_use_of_target_variable t
 let expand_read_macro ~dir ~source s ~read ~pack =
   let path = relative ~source dir s in
   let read =
-    let open Memo.Build.O in
+    let open Memo.O in
     let+ x = Build_system.read_file path ~f:read in
     pack x
   in
@@ -277,7 +272,7 @@ let expand_read_macro ~dir ~source s ~read ~pack =
         (* To prevent it from working in certain position before Dune 3.0. It'd
            be nice if we could invite the user to upgrade to (lang dune 3.0),
            but this is a bigger refactoring. *)
-        With (Action_builder.memo_build read))
+        With (Action_builder.of_memo read))
 
 let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
     (pform : Pform.t) : expansion_result =
@@ -314,15 +309,15 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
       | Ocamlc -> static (path context.ocamlc)
       | Ocamlopt -> static (get_prog context.ocamlopt)
       | Make ->
-        let open Memo.Build.O in
+        let open Memo.O in
         Direct
           (Without
-             (context.which "make" >>| function
+             (Context.make context >>| function
+              | Some p -> path p
               | None ->
                 Utils.program_not_found ~context:context.name
                   ~loc:(Some (Dune_lang.Template.Pform.loc source))
-                  "make"
-              | Some p -> path p))
+                  "make"))
       | Cpp -> static (strings (c_compiler_and_flags context @ [ "-E" ]))
       | Pa_cpp ->
         static
@@ -366,19 +361,18 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
         Need_full_expander
           (fun t ->
             Without
-              (Memo.Build.return
-                 [ Value.Dir (Path.build (Scope.root t.scope)) ]))
+              (Memo.return [ Value.Dir (Path.build (Scope.root t.scope)) ]))
       | Cc ->
         Need_full_expander
           (fun t ->
             With
-              (let* cc = Action_builder.memo_build (cc t) in
+              (let* cc = Action_builder.of_memo (cc t) in
                cc.c))
       | Cxx ->
         Need_full_expander
           (fun t ->
             With
-              (let* cc = Action_builder.memo_build (cc t) in
+              (let* cc = Action_builder.of_memo (cc t) in
                cc.cxx))
       | Ccomp_type ->
         static
@@ -427,7 +421,7 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                    string v)
               | None ->
                 Without
-                  (Memo.Build.return
+                  (Memo.return
                      (string (Option.value ~default (Env.get t.env var))))))
       | Version ->
         Need_full_expander (fun t -> Without (expand_version t ~source s))
@@ -447,7 +441,7 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
           (fun t ->
             With
               (let* prog =
-                 Action_builder.memo_build
+                 Action_builder.of_memo
                    (Artifacts.Bin.binary
                       ~loc:(Some (Dune_lang.Template.Pform.loc source))
                       t.bin_artifacts_host s)
@@ -467,7 +461,7 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                in
                let scope = if lib_exec then t.scope_host else t.scope in
                let p =
-                 let open Resolve.Build.O in
+                 let open Resolve.Memo.O in
                  if lib_private then
                    let* lib =
                      Lib.DB.resolve (Scope.libs scope)
@@ -481,10 +475,10 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                      Option.equal Dune_project.equal (Some current_project)
                        referenced_project
                    then
-                     Resolve.Build.return
+                     Resolve.Memo.return
                        (Path.relative (Lib_info.src_dir (Lib.info lib)) file)
                    else
-                     Resolve.Build.fail
+                     Resolve.Memo.fail
                        (User_error.make
                           ~loc:(Dune_lang.Template.Pform.loc source)
                           [ Pp.textf
@@ -511,8 +505,8 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                      ~lib ~file
                in
                let p =
-                 let open Memo.Build.O in
-                 Resolve.Build.peek p >>| function
+                 let open Memo.O in
+                 Resolve.Memo.peek p >>| function
                  | Ok p -> (
                    match file with
                    | "" | "." ->
@@ -557,17 +551,17 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                  | Error () ->
                    let p =
                      if lib_private then
-                       Resolve.Build.map p ~f:(fun _ -> assert false)
+                       Resolve.Memo.map p ~f:(fun _ -> assert false)
                      else
-                       let open Resolve.Build.O in
+                       let open Resolve.Memo.O in
                        let* available =
-                         Resolve.Build.lift_memo
+                         Resolve.Memo.lift_memo
                            (Lib.DB.available (Scope.libs scope) lib)
                        in
                        match available with
                        | false -> p >>| fun _ -> assert false
                        | true ->
-                         Resolve.Build.fail
+                         Resolve.Memo.fail
                            (User_error.make
                               ~loc:(Dune_lang.Template.Pform.loc source)
                               [ Pp.textf
@@ -579,9 +573,9 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                                   (if lib_exec then "exec" else "")
                               ])
                    in
-                   Resolve.Build.read p
+                   Resolve.Memo.read p
                in
-               Action_builder.memo_build_join p))
+               Action_builder.of_memo_join p))
       | Lib_available ->
         Need_full_expander
           (fun t ->
@@ -590,14 +584,14 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                  Lib_name.parse_string_exn
                    (Dune_lang.Template.Pform.loc source, s)
                in
-               let open Memo.Build.O in
+               let open Memo.O in
                let+ available = Lib.DB.available (Scope.libs t.scope) lib in
                available |> string_of_bool |> string))
       | Bin_available ->
         Need_full_expander
           (fun t ->
             Without
-              (let open Memo.Build.O in
+              (let open Memo.O in
               let+ b = Artifacts.Bin.binary_available t.bin_artifacts_host s in
               b |> string_of_bool |> string))
       | Read -> expand_read_macro ~dir ~source s ~read:Io.read_file ~pack:string
@@ -616,7 +610,31 @@ let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source
                            "This file must be a list of lines escaped using \
                             OCaml's conventions"
                        ]
-                   | Ok s -> s)))))
+                   | Ok s -> s)))
+      | Coq_config ->
+        Need_full_expander
+          (fun t ->
+            Without
+              (let open Memo.O in
+              let* bin =
+                Artifacts.Bin.binary t.bin_artifacts_host ~loc:None "coqc"
+              in
+              match bin with
+              | Ok bin -> (
+                let+ t = Coq_config.make ~bin in
+                match Coq_config.by_name t s with
+                | None ->
+                  User_error.raise
+                    ~loc:(Dune_lang.Template.Pform.loc source)
+                    [ Pp.textf "Unknown Coq configuration variable %S" s ]
+                | Some v -> (
+                  match v with
+                  | Bool x -> string (string_of_bool x)
+                  | Int x -> string (string_of_int x)
+                  | String x -> string x
+                  | Strings x -> strings x
+                  | Path x -> Value.L.paths [ x ]))
+              | Error _ -> User_error.raise Pp.[ textf "coqc not found." ]))))
 
 (* Make sure to delay exceptions *)
 let expand_pform_gen ~context ~bindings ~dir ~source pform =
@@ -624,8 +642,8 @@ let expand_pform_gen ~context ~bindings ~dir ~source pform =
   | exception (User_error.E _ as exn) ->
     Direct
       (Without
-         (let open Memo.Build.O in
-         let+ () = Memo.Build.return () in
+         (let open Memo.O in
+         let+ () = Memo.return () in
          reraise exn))
   | Direct _ as x -> x
   | Need_full_expander f ->
@@ -634,8 +652,8 @@ let expand_pform_gen ~context ~bindings ~dir ~source pform =
         try f t
         with User_error.E _ as exn ->
           Without
-            (let open Memo.Build.O in
-            let+ () = Memo.Build.return () in
+            (let open Memo.O in
+            let+ () = Memo.return () in
             reraise exn))
 
 let describe_source ~source =
@@ -655,7 +673,7 @@ let expand_pform t ~source pform =
         | Need_full_expander f -> f t
       with
       | With x -> x
-      | Without x -> Action_builder.memo_build x)
+      | Without x -> Action_builder.of_memo x)
     ~human_readable_description:(fun () -> describe_source ~source)
 
 let expand_pform_no_deps t ~source pform =
@@ -678,7 +696,9 @@ let expand t ~mode template =
     ~f:(expand_pform t)
 
 let make ~scope ~scope_host ~(context : Context.t) ~lib_artifacts
-    ~lib_artifacts_host ~bin_artifacts_host ~find_package =
+    ~lib_artifacts_host ~bin_artifacts_host =
+  let open Memo.O in
+  let+ sites = Sites.create context in
   { dir = context.build_dir
   ; env = context.env
   ; local_env = Env.Var.Map.empty
@@ -695,7 +715,7 @@ let make ~scope ~scope_host ~(context : Context.t) ~lib_artifacts
       (fun ~dir ->
         Code_error.raise "foreign flags expander is not set"
           [ ("dir", Path.Build.to_dyn dir) ])
-  ; find_package
+  ; sites
   ; expanding_what = Nothing_special
   }
 
@@ -708,7 +728,7 @@ let expand_str t sw =
   Value.to_string v ~dir:(Path.build t.dir)
 
 module No_deps = struct
-  open Memo.Build.O
+  open Memo.O
 
   let expand_pform = expand_pform_no_deps
 
@@ -752,8 +772,10 @@ module With_deps_if_necessary = struct
     E.expand ~dir:(Path.build t.dir) ~mode sw ~f:(expand_pform t)
 
   let expand_path t sw =
-    let+ v = expand t ~mode:Single sw in
-    Value.to_path v ~error_loc:(String_with_vars.loc sw) ~dir:(Path.build t.dir)
+    let+ vs = expand t ~mode:Many sw in
+    List.map vs ~f:(fun v ->
+        Value.to_path_in_build_or_external v
+          ~error_loc:(String_with_vars.loc sw) ~dir:t.dir)
 
   let expand_str t sw =
     let+ v = expand t ~mode:Single sw in
@@ -761,14 +783,14 @@ module With_deps_if_necessary = struct
 end
 
 module With_reduced_var_set = struct
-  open Memo.Build.O
+  open Memo.O
 
   let expand_pform_opt ~context ~bindings ~dir ~source pform =
-    let open Memo.Build.O in
+    let open Memo.O in
     Memo.push_stack_frame
       (fun () ->
         match expand_pform_gen ~context ~bindings ~dir ~source pform with
-        | Need_full_expander _ | Direct (With _) -> Memo.Build.return None
+        | Need_full_expander _ | Direct (With _) -> Memo.return None
         | Direct (Without x) -> x >>| Option.some)
       ~human_readable_description:(fun () -> describe_source ~source)
 
@@ -806,4 +828,15 @@ let expand_and_eval_set t set ~standard =
 let eval_blang t blang =
   Blang.eval ~f:(No_deps.expand_pform t) ~dir:(Path.build t.dir) blang
 
-let find_package t pkg = t.find_package pkg
+let expand_lock ~base expander (Locks.Lock sw) =
+  let open Memo.O in
+  match base with
+  | `Of_expander -> No_deps.expand_path expander sw
+  | `This base ->
+    let+ str = No_deps.expand_str expander sw in
+    Path.relative base str
+
+let expand_locks ~base expander locks =
+  Memo.List.map locks ~f:(expand_lock ~base expander)
+
+let sites t = t.sites
