@@ -1,10 +1,8 @@
-open! Dune_engine
-open! Dune_engine.Import
-open! Stdune
-open Memo.Build.O
+open Import
+open Memo.O
 module Opam_package = Package
-module P = Variant
-module Ps = Variant.Set
+module P = Ocaml.Variant
+module Ps = Ocaml.Variant.Set
 
 let meta_fn = "META"
 
@@ -81,7 +79,7 @@ module Rules = struct
     let set_rules =
       List.map rules.set_rules ~f:Rule.make
       |> List.stable_sort ~compare:(fun a b ->
-             Poly.compare
+             Int.compare
                (Rule.formal_predicates_count b)
                (Rule.formal_predicates_count a))
     in
@@ -117,7 +115,9 @@ module Config = struct
   let load path ~toolchain ~context =
     let path = Path.extend_basename path ~suffix:".d" in
     let conf_file = Path.relative path (toolchain ^ ".conf") in
-    let* conf_file_exists = Fs_memo.file_exists conf_file in
+    let* conf_file_exists =
+      Fs_memo.file_exists (Path.as_outside_build_dir_exn conf_file)
+    in
     if not conf_file_exists then
       User_error.raise
         [ Pp.textf "ocamlfind toolchain %s isn't defined in %s (context: %s)"
@@ -125,7 +125,9 @@ module Config = struct
             (Path.to_string_maybe_quoted path)
             context
         ];
-    let+ meta = Meta.load ~name:None conf_file in
+    let+ meta =
+      Meta.load ~name:None (Path.as_outside_build_dir_exn conf_file)
+    in
     { vars = String.Map.map meta.vars ~f:Rules.of_meta_rules
     ; preds = Ps.of_list [ P.make toolchain ]
     }
@@ -207,9 +209,9 @@ module Loader : sig
   val lookup_and_load :
        DB.t
     -> Package.Name.t
-    -> (Dune_package.t, Unavailable_reason.t) result Memo.Build.t
+    -> (Dune_package.t, Unavailable_reason.t) result Memo.t
 
-  val dummy_package : DB.t -> Lib_name.t -> Dune_package.t Memo.Build.t
+  val dummy_package : DB.t -> Lib_name.t -> Dune_package.t Memo.t
 end = struct
   module Findlib_package : sig
     type t =
@@ -220,9 +222,9 @@ end = struct
       }
 
     val to_dune_library :
-      t -> lib_config:Lib_config.t -> Dune_package.Lib.t Memo.Build.t
+      t -> lib_config:Lib_config.t -> Dune_package.Lib.t Memo.t
 
-    val exists : t -> is_builtin:bool -> bool Memo.Build.t
+    val exists : t -> is_builtin:bool -> bool Memo.t
   end = struct
     type t =
       { meta_file : Path.t
@@ -277,10 +279,11 @@ end = struct
       let exists_if = Vars.get_words t.vars "exists_if" Ps.empty in
       match exists_if with
       | _ :: _ ->
-        Memo.Build.List.for_all exists_if ~f:(fun fn ->
-            Fs_memo.file_exists (Path.relative t.dir fn))
+        Memo.List.for_all exists_if ~f:(fun fn ->
+            Fs_memo.file_exists
+              (Path.as_outside_build_dir_exn (Path.relative t.dir fn)))
       | [] -> (
-        if not is_builtin then Memo.Build.return true
+        if not is_builtin then Memo.return true
         else
           (* The META files for installed packages are sometimes broken, i.e.
              META files for libraries that were not installed by the compiler
@@ -291,9 +294,10 @@ end = struct
              To workaround this problem, for builtin packages we check that at
              least one of the archive is present. *)
           match archives t with
-          | { byte = []; native = [] } -> Memo.Build.return true
+          | { byte = []; native = [] } -> Memo.return true
           | { byte; native } ->
-            Memo.Build.List.exists (byte @ native) ~f:Fs_memo.file_exists)
+            Memo.List.exists (byte @ native) ~f:(fun p ->
+                Path.as_outside_build_dir_exn p |> Fs_memo.file_exists))
 
     let to_dune_library t ~(lib_config : Lib_config.t) =
       let loc = Loc.in_file t.meta_file in
@@ -301,7 +305,9 @@ end = struct
       let dot_dune_file =
         Path.relative t.dir (sprintf "%s.dune" (Lib_name.to_string t.name))
       in
-      let* dot_dune_exists = Fs_memo.file_exists dot_dune_file in
+      let* dot_dune_exists =
+        Fs_memo.file_exists (Path.as_outside_build_dir_exn dot_dune_file)
+      in
       if dot_dune_exists then
         User_warning.emit
           ~loc:(Loc.in_file dot_dune_file)
@@ -312,12 +318,15 @@ end = struct
           ];
       let archives = archives t in
       let obj_dir = Obj_dir.make_external_no_private ~dir:t.dir in
-      let modes : Mode.Dict.Set.t =
+      let modes : Lib_mode.Map.Set.t =
         (* libraries without archives are compatible with all modes. mainly a
            hack for compiler-libs which doesn't have any archives *)
         let discovered = Mode.Dict.map ~f:List.is_non_empty archives in
-        if Mode.Dict.Set.is_empty discovered then Mode.Dict.Set.all
-        else discovered
+        let modes =
+          if Mode.Dict.Set.is_empty discovered then Mode.Dict.Set.all
+          else discovered
+        in
+        { Lib_mode.Map.ocaml = modes; melange = false }
       in
       let+ (info : Path.t Lib_info.t) =
         let kind = kind t in
@@ -356,7 +365,9 @@ end = struct
         let virtual_ = None in
         let default_implementation = None in
         let wrapped = None in
-        let+ dir_contents = Fs_memo.dir_contents t.dir in
+        let+ dir_contents =
+          Fs_memo.dir_contents (Path.as_outside_build_dir_exn t.dir)
+        in
         let foreign_archives, native_archives =
           (* Here we scan [t.dir] and consider all files named [lib*.ext_lib] to
              be foreign archives, and all other files with the extension
@@ -389,6 +400,9 @@ end = struct
                   else Right file
                 else Skip)
         in
+        let foreign_archives =
+          Mode.Map.Multi.create_for_all_modes foreign_archives
+        in
         let entry_modules =
           Lib_info.Source.External
             (match Vars.get_words t.vars "main_modules" Ps.empty with
@@ -472,7 +486,7 @@ end = struct
       let acc =
         Lib_name.Map.add_exn acc (Dune_package.Entry.name entry) entry
       in
-      Memo.Build.List.fold_left meta.subs ~init:acc
+      Memo.List.fold_left meta.subs ~init:acc
         ~f:(fun acc (meta : Meta.Simplified.t) ->
           let full_name =
             match meta.name with
@@ -498,7 +512,9 @@ end = struct
     }
 
   let load_and_convert db ~dir ~meta_file ~name =
-    let* meta = Meta.load meta_file ~name:(Some name) in
+    let* meta =
+      Meta.load (Path.as_outside_build_dir_exn meta_file) ~name:(Some name)
+    in
     dune_package_of_meta db ~dir ~meta_file ~meta
 
   let load_builtin db meta =
@@ -522,53 +538,51 @@ end = struct
     let meta = dummy top_lib subs in
     load_builtin db meta
 
-  let lookup_and_load_one_dir db ~dir ~name =
-    let meta_file = Path.relative dir meta_fn in
-    let* meta_file_exists = Fs_memo.file_exists meta_file in
-    if meta_file_exists then
-      load_and_convert db ~dir ~meta_file ~name >>| Option.some
-    else
-      (* Alternative layout *)
-      match Path.parent dir with
-      | None -> Memo.Build.return None
-      | Some dir ->
-        let meta_file =
-          Path.relative dir (meta_fn ^ "." ^ Package.Name.to_string name)
-        in
-        let* meta_exists = Fs_memo.file_exists meta_file in
-        if meta_exists then
-          load_and_convert db ~dir ~meta_file ~name >>| Option.some
-        else Memo.Build.return None
-
   let lookup_and_load (db : DB.t) name =
-    let rec loop dirs :
-        (Dune_package.t, Unavailable_reason.t) Result.t Memo.Build.t =
+    let rec loop dirs : (Dune_package.t, Unavailable_reason.t) Result.t Memo.t =
       match dirs with
       | [] -> (
         match Package.Name.to_string name with
-        | "dune" -> Memo.Build.return (Ok builtin_for_dune)
-        | _ -> Memo.Build.return (Error Unavailable_reason.Not_found))
+        | "dune" -> Memo.return (Ok builtin_for_dune)
+        | _ -> Memo.return (Error Unavailable_reason.Not_found))
       | dir :: dirs -> (
-        let dir = Path.relative dir (Package.Name.to_string name) in
-        let* dir_exists = Fs_memo.dir_exists dir in
-        if not dir_exists then loop dirs
+        let meta_file =
+          Path.relative dir (meta_fn ^ "." ^ Package.Name.to_string name)
+        in
+        let* file_exists =
+          Fs_memo.file_exists (Path.as_outside_build_dir_exn meta_file)
+        in
+        if file_exists then
+          let+ p = load_and_convert db ~dir ~meta_file ~name in
+          Ok p
         else
-          let dune = Path.relative dir Dune_package.fn in
-          let* exists =
-            let+ exists = Fs_memo.file_exists dune in
-            if exists then Dune_package.Or_meta.load dune
-            else Ok Dune_package.Or_meta.Use_meta
+          let dir = Path.relative dir (Package.Name.to_string name) in
+          let* dir_exists =
+            Fs_memo.dir_exists (Path.as_outside_build_dir_exn dir)
           in
-          match exists with
-          | Error e ->
-            Memo.Build.return
-              (Error (Unavailable_reason.Invalid_dune_package e))
-          | Ok (Dune_package.Or_meta.Dune_package p) -> Memo.Build.return (Ok p)
-          | Ok Use_meta -> (
-            let open Memo.Build.O in
-            lookup_and_load_one_dir db ~dir ~name >>= function
-            | None -> loop dirs
-            | Some p -> Memo.Build.return (Ok p)))
+          if not dir_exists then loop dirs
+          else
+            let dune = Path.relative dir Dune_package.fn in
+            let* exists =
+              let* exists =
+                Fs_memo.file_exists (Path.as_outside_build_dir_exn dune)
+              in
+              if exists then Dune_package.Or_meta.load dune
+              else Memo.return (Ok Dune_package.Or_meta.Use_meta)
+            in
+            match exists with
+            | Error e ->
+              Memo.return (Error (Unavailable_reason.Invalid_dune_package e))
+            | Ok (Dune_package.Or_meta.Dune_package p) -> Memo.return (Ok p)
+            | Ok Use_meta ->
+              let meta_file = Path.relative dir meta_fn in
+              let* meta_file_exists =
+                Fs_memo.file_exists (Path.as_outside_build_dir_exn meta_file)
+              in
+              if meta_file_exists then
+                let+ p = load_and_convert db ~dir ~meta_file ~name in
+                Ok p
+              else loop dirs)
     in
     match Package.Name.Map.find db.builtins name with
     | None -> loop db.paths
@@ -598,11 +612,11 @@ let dummy_lib db ~name =
   | _ -> assert false
 
 let find_root_package db name :
-    (Dune_package.t, Unavailable_reason.t) result Memo.Build.t =
+    (Dune_package.t, Unavailable_reason.t) result Memo.t =
   Memo.exec memo (db, name)
 
 let find t name =
-  let open Memo.Build.O in
+  let open Memo.O in
   let+ p = find_root_package t (Lib_name.package_name name) in
   let open Result.O in
   let* p = p in
@@ -612,9 +626,9 @@ let find t name =
 
 let root_packages (db : DB.t) =
   let+ pkgs =
-    Memo.Build.List.concat_map db.paths ~f:(fun dir ->
-        Fs_memo.dir_contents dir >>= function
-        | Error (ENOENT, _, _) -> Memo.Build.return []
+    Memo.List.concat_map db.paths ~f:(fun dir ->
+        Fs_memo.dir_contents (Path.as_outside_build_dir_exn dir) >>= function
+        | Error (ENOENT, _, _) -> Memo.return []
         | Error (unix_error, _, _) ->
           User_error.raise
             [ Pp.textf "Unable to read directory %s for findlib package"
@@ -623,9 +637,11 @@ let root_packages (db : DB.t) =
             ]
         | Ok dir_contents ->
           let dir_contents = Fs_cache.Dir_contents.to_list dir_contents in
-          Memo.Build.List.filter_map dir_contents ~f:(fun (name, _) ->
+          Memo.List.filter_map dir_contents ~f:(fun (name, _) ->
               let+ exists =
-                Fs_memo.file_exists (Path.relative dir (name ^ "/" ^ meta_fn))
+                Fs_memo.file_exists
+                  (Path.as_outside_build_dir_exn
+                     (Path.relative dir (name ^ "/" ^ meta_fn)))
               in
               if exists then Some (Package.Name.of_string name) else None))
     >>| Package.Name.Set.of_list
@@ -635,7 +651,7 @@ let root_packages (db : DB.t) =
 
 let load_all_packages (t : t) =
   root_packages t >>| Package.Name.Set.to_list
-  >>= Memo.Build.parallel_map ~f:(fun name ->
+  >>= Memo.parallel_map ~f:(fun name ->
           let+ pkg = find_root_package t name in
           (name, pkg))
 
@@ -651,14 +667,13 @@ let all_packages t =
            (Dune_package.Entry.name a)
            (Dune_package.Entry.name b))
 
-let create ~paths ~(lib_config : Lib_config.t) : t =
+let create ~paths ~(lib_config : Lib_config.t) =
   let stdlib_dir = lib_config.stdlib_dir in
   let version = lib_config.ocaml_version in
-  { DB.stdlib_dir
-  ; paths
-  ; builtins = Meta.builtins ~stdlib_dir ~version
-  ; lib_config
-  }
+  let+ builtins = Meta.builtins ~stdlib_dir ~version in
+  { DB.stdlib_dir; paths; builtins; lib_config }
+
+let lib_config (t : t) = t.lib_config
 
 let all_broken_packages t =
   let+ packages = load_all_packages t in
@@ -667,3 +682,21 @@ let all_broken_packages t =
       | Ok _ | Error Unavailable_reason.Not_found -> acc
       | Error (Invalid_dune_package exn) -> (name, exn) :: acc)
   |> List.sort ~compare:(fun (a, _) (b, _) -> Package.Name.compare a b)
+
+let create =
+  let module Input = struct
+    type t = Path.t list * Lib_config.t
+
+    let equal (paths, libs) (paths', libs') =
+      List.equal Path.equal paths paths' && Lib_config.equal libs libs'
+
+    let hash = Tuple.T2.hash (List.hash Path.hash) Lib_config.hash
+
+    let to_dyn = Dyn.pair (Dyn.list Path.to_dyn) Lib_config.to_dyn
+  end in
+  let memo =
+    Memo.create "lib-installed"
+      ~input:(module Input)
+      (fun (paths, lib_config) -> create ~paths ~lib_config)
+  in
+  fun ~paths ~lib_config -> Memo.exec memo (paths, lib_config)
