@@ -1,5 +1,4 @@
-open! Dune_engine
-open Stdune
+open Import
 open Dune_file
 module Library = Dune_file.Library
 
@@ -9,7 +8,7 @@ module Library = Dune_file.Library
    Before this module is removed, there should be a good way to handle new types
    of source files without shoving everything into [Dir_contents].
 
-   Furthemore, this module is also responsible for details such as handling file
+   Furthermore, this module is also responsible for details such as handling file
    extensions and validating filenames. *)
 type t =
   { libraries : Foreign.Sources.t Lib_name.Map.t
@@ -39,11 +38,32 @@ let valid_name language ~loc s =
       ]
   | _ -> s
 
-let eval_foreign_stubs (d : _ Dir_with_dune.t) foreign_stubs
-    ~(sources : Foreign.Sources.Unresolved.t) : Foreign.Sources.t =
-  let multiple_sources_error ~name ~loc ~paths =
+let eval_foreign_stubs foreign_stubs (ctypes : Ctypes_stanza.t option)
+    ~dune_version ~(sources : Foreign.Sources.Unresolved.t) : Foreign.Sources.t
+    =
+  let multiple_sources_error ~name ~mode ~loc ~paths =
+    let hints =
+      [ Pp.text
+          "You can also avoid the name clash by placing the objects into \
+           different foreign archives and building them in different \
+           directories. Foreign archives can be defined using the \
+           (foreign_library ...) stanza."
+      ]
+    in
+    let hints, for_mode =
+      match mode with
+      | Mode.Select.All -> (hints, "")
+      | Mode.Select.Only m ->
+        let mode_hint =
+          Pp.text
+            "You may be missing a mode field that would restrict this stub to \
+             some specific mode."
+        in
+        (mode_hint :: hints, Printf.sprintf " for mode %s" @@ Mode.to_string m)
+    in
     User_error.raise ~loc
-      [ Pp.textf "Multiple sources map to the same object name %S:" name
+      [ Pp.textf "Multiple sources map to the same object name %S%s:" name
+          for_mode
       ; Pp.enumerate (List.sort ~compare:Path.Build.compare paths)
           ~f:(fun path ->
             Pp.text
@@ -54,13 +74,18 @@ let eval_foreign_stubs (d : _ Dir_with_dune.t) foreign_stubs
            names."
           name
       ]
-      ~hints:
-        [ Pp.text
-            "You can also avoid the name clash by placing the objects into \
-             different foreign archives and building them in different \
-             directories. Foreign archives can be defined using the \
-             (foreign_library ...) stanza."
-        ]
+      ~hints
+  in
+  let find_source language (loc, name) =
+    let open Option.O in
+    let* candidates = String.Map.find sources name in
+    match
+      List.filter_map candidates ~f:(fun (l, path) ->
+          Option.some_if (Foreign_language.equal l language) path)
+    with
+    | [ path ] -> Some path
+    | [] -> None
+    | _ :: _ :: _ as paths -> multiple_sources_error ~mode:All ~name ~loc ~paths
   in
   let eval (stubs : Foreign.Stubs.t) =
     let language = stubs.language in
@@ -75,7 +100,7 @@ let eval_foreign_stubs (d : _ Dir_with_dune.t) foreign_stubs
       Ordered_set_lang.Unordered_string.eval_loc stubs.names ~key:Fun.id
         ~standard ~parse:(fun ~loc:_ -> Fun.id)
     in
-    String.Map.map names ~f:(fun (loc, s) ->
+    String.Map.fold names ~init:String.Map.empty ~f:(fun (loc, s) acc ->
         let name = valid_name language ~loc s in
         let basename = Filename.basename s in
         if name <> basename then
@@ -85,32 +110,47 @@ let eval_foreign_stubs (d : _ Dir_with_dune.t) foreign_stubs
                  To include sources in subdirectories, use the \
                  (include_subdirs ...) stanza."
             ];
-        let open Option.O in
-        let source =
-          let* candidates = String.Map.find sources name in
-          match
-            List.filter_map candidates ~f:(fun (l, path) ->
-                Option.some_if (Foreign_language.equal l language) path)
-          with
-          | [ path ] -> Some (loc, Foreign.Source.make ~stubs ~path)
-          | [] -> None
-          | _ :: _ :: _ as paths -> multiple_sources_error ~name ~loc ~paths
-        in
-        match source with
-        | Some source -> source
+        match find_source language (loc, name) with
+        | Some path ->
+          let src = Foreign.Source.make (Stubs stubs) ~path in
+          let new_key = Foreign.Source.object_name src in
+          String.Map.add_exn acc new_key (loc, src)
         | None ->
           User_error.raise ~loc
             [ Pp.textf "Object %S has no source; %s must be present." name
                 (String.enumerate_one_of
-                   (Foreign.possible_sources ~language name
-                      ~dune_version:d.dune_version
+                   (Foreign.possible_sources ~language name ~dune_version
                    |> List.map ~f:(fun s -> sprintf "%S" s)))
             ])
   in
-  let stub_maps = List.map foreign_stubs ~f:eval in
+  let stub_maps =
+    let init = List.map foreign_stubs ~f:eval in
+    match ctypes with
+    | None -> init
+    | Some ctypes ->
+      let ctypes =
+        List.fold_left ~init:String.Map.empty ctypes.function_description
+          ~f:(fun acc (fd : Ctypes_stanza.Function_description.t) ->
+            let loc = Loc.none (* TODO *) in
+            let fname = Ctypes_stanza.c_generated_functions_cout_c ctypes fd in
+            let name = Filename.chop_extension fname in
+            let path =
+              match find_source C (loc, name) with
+              | Some p -> p
+              | None ->
+                (* impossible b/c ctypes stanza generates this *)
+                assert false
+            in
+            let source = Foreign.Source.make (Ctypes ctypes) ~path in
+            String.Map.add_exn acc name (loc, source))
+      in
+      ctypes :: init
+  in
   List.fold_left stub_maps ~init:String.Map.empty ~f:(fun a b ->
-      String.Map.union a b ~f:(fun name (loc, src1) (_, src2) ->
-          multiple_sources_error ~name ~loc
+      String.Map.union a b ~f:(fun _name (loc, src1) (_, src2) ->
+          let name = Foreign.Source.user_object_name src1 in
+          let mode = Foreign.Source.mode src1 in
+          multiple_sources_error ~name ~loc ~mode
             ~paths:Foreign.Source.[ path src1; path src2 ]))
 
 let check_no_qualified (loc, include_subdirs) =
@@ -120,27 +160,31 @@ let check_no_qualified (loc, include_subdirs) =
           "(include_subdirs qualified) is only meant for OCaml and Coq sources"
       ]
 
-let make (d : _ Dir_with_dune.t) ~(sources : Foreign.Sources.Unresolved.t)
+let make stanzas ~(sources : Foreign.Sources.Unresolved.t) ~dune_version
     ~(lib_config : Lib_config.t) =
   let libs, foreign_libs, exes =
     let libs, foreign_libs, exes =
-      List.fold_left d.data ~init:([], [], [])
+      List.fold_left stanzas ~init:([], [], [])
         ~f:(fun ((libs, foreign_libs, exes) as acc) stanza ->
           match (stanza : Stanza.t) with
           | Library lib ->
             let all =
-              eval_foreign_stubs d lib.buildable.foreign_stubs ~sources
+              eval_foreign_stubs ~dune_version lib.buildable.foreign_stubs
+                lib.buildable.ctypes ~sources
             in
             ((lib, all) :: libs, foreign_libs, exes)
           | Foreign_library library ->
-            let all = eval_foreign_stubs d [ library.stubs ] ~sources in
+            let all =
+              eval_foreign_stubs ~dune_version [ library.stubs ] ~sources None
+            in
             ( libs
             , (library.archive_name, (library.archive_name_loc, all))
               :: foreign_libs
             , exes )
           | Executables exe | Tests { exes = exe; _ } ->
             let all =
-              eval_foreign_stubs d exe.buildable.foreign_stubs ~sources
+              eval_foreign_stubs ~dune_version exe.buildable.foreign_stubs
+                ~sources exe.buildable.ctypes
             in
             (libs, foreign_libs, (exe, all) :: exes)
           | _ -> acc)
@@ -214,10 +258,9 @@ let make (d : _ Dir_with_dune.t) ~(sources : Foreign.Sources.Unresolved.t)
   in
   { libraries; archives; executables }
 
-let make (d : _ Dir_with_dune.t) ~include_subdirs ~(lib_config : Lib_config.t)
+let make stanzas ~dune_version ~include_subdirs ~(lib_config : Lib_config.t)
     ~dirs =
   check_no_qualified include_subdirs;
-  let dune_version = d.dune_version in
   let init = String.Map.empty in
   let sources =
     List.fold_left dirs ~init ~f:(fun acc (dir, _local, files) ->
@@ -226,4 +269,4 @@ let make (d : _ Dir_with_dune.t) ~include_subdirs ~(lib_config : Lib_config.t)
         in
         String.Map.Multi.rev_union sources acc)
   in
-  make d ~sources ~lib_config
+  make stanzas ~dune_version ~sources ~lib_config
