@@ -1,4 +1,3 @@
-open! Dune_engine
 open Import
 
 let add_diff sctx loc alias ~dir ~input ~output =
@@ -20,6 +19,44 @@ let depend_on_files ~named dir =
 
 let formatted_dir_basename = ".formatted"
 
+let action =
+  let module Spec = struct
+    type ('path, 'target) t = Dune_lang.Syntax.Version.t * 'path * 'target
+
+    let name = "format-dune-file"
+
+    let version = 1
+
+    let bimap (ver, src, dst) f g = (ver, f src, g dst)
+
+    let is_useful_to ~distribute:_ ~memoize = memoize
+
+    let encode (version, src, dst) path target : Dune_lang.t =
+      List
+        [ Dune_lang.atom_or_quoted_string "format-dune-file"
+        ; Dune_lang.Syntax.Version.encode version
+        ; path src
+        ; target dst
+        ]
+
+    let action (version, src, dst) ~ectx:_ ~eenv:_ =
+      Dune_lang.Format.format_action ~version ~src ~dst;
+      Fiber.return ()
+  end in
+  fun ~version (src : Path.t) (dst : Path.Build.t) ->
+    let module M :
+      Action.Ext.Instance with type path = Path.t and type target = Path.Build.t =
+    struct
+      type path = Path.t
+
+      type target = Path.Build.t
+
+      module Spec = Spec
+
+      let v = (version, src, dst)
+    end in
+    Action.Extension (module M)
+
 let gen_rules_output sctx (config : Format_config.t) ~version ~dialects
     ~expander ~output_dir =
   assert (formatted_dir_basename = Path.Build.basename output_dir);
@@ -41,7 +78,7 @@ let gen_rules_output sctx (config : Format_config.t) ~version ~dialects
         @@
         let open Action_builder.O in
         let+ () = Action_builder.path input in
-        Action.Full.make (Action.format_dune_file ~version input output)
+        Action.Full.make (action ~version input output)
       | _ ->
         let ext = Path.Source.extension file in
         let open Option.O in
@@ -67,42 +104,41 @@ let gen_rules_output sctx (config : Format_config.t) ~version ~dialects
         in
         let open Action_builder.With_targets.O in
         Action_builder.with_no_targets extra_deps
-        >>> Preprocessing.action_for_pp_with_target ~loc ~expander ~action ~src
+        >>> Preprocessing.action_for_pp_with_target
+              ~sandbox:Sandbox_config.default ~loc ~expander ~action ~src
               ~target:output
     in
-    Memo.Build.Option.iter formatter ~f:(fun action ->
-        let open Memo.Build.O in
+    Memo.Option.iter formatter ~f:(fun action ->
+        let open Memo.O in
         Super_context.add_rule sctx ~mode:Standard ~loc ~dir action
         >>> add_diff sctx loc alias_formatted ~dir ~input:(Path.build input)
               ~output)
   in
-  let open Memo.Build.O in
+  let open Memo.O in
   let* () =
     Source_tree.files_of source_dir
-    >>= Memo.Build.parallel_iter_set
-          (module Path.Source.Set)
-          ~f:setup_formatting
+    >>= Memo.parallel_iter_set (module Path.Source.Set) ~f:setup_formatting
   in
   Rules.Produce.Alias.add_deps alias_formatted (Action_builder.return ())
 
 let gen_rules sctx ~output_dir =
-  let open Memo.Build.O in
+  let open Memo.O in
   let dir = Path.Build.parent_exn output_dir in
   let* config = Super_context.format_config sctx ~dir in
-  Memo.Build.when_
+  Memo.when_
     (not (Format_config.is_empty config))
     (fun () ->
       let* expander = Super_context.expander sctx ~dir in
-      let scope = Super_context.find_scope_by_dir sctx output_dir in
+      let* scope = Scope.DB.find_by_dir output_dir in
       let project = Scope.project scope in
       let dialects = Dune_project.dialects project in
       let version = Dune_project.dune_version project in
       gen_rules_output sctx config ~version ~dialects ~expander ~output_dir)
 
 let setup_alias sctx ~dir =
-  let open Memo.Build.O in
+  let open Memo.O in
   let* config = Super_context.format_config sctx ~dir in
-  Memo.Build.when_
+  Memo.when_
     (not (Format_config.is_empty config))
     (fun () ->
       let output_dir = Path.Build.relative dir formatted_dir_basename in
