@@ -1,10 +1,13 @@
-open! Dune_engine
-open! Stdune
 open Import
-open Memo.Build.O
-open! No_io
+open Memo.O
 module Library = Dune_file.Library
 
+let install_file ~(package : Package.Name.t) ~findlib_toolchain =
+  let package = Package.Name.to_string package in
+  match findlib_toolchain with
+  | None -> package ^ ".install"
+  | Some x -> sprintf "%s-%s.install" package (Context_name.to_string x)
+
 module Package_paths = struct
   let opam_file (ctx : Context.t) pkg =
     Path.Build.append_source ctx.build_dir (Package.opam_file pkg)
@@ -35,16 +38,15 @@ end
 
 module Stanzas_to_entries : sig
   val stanzas_to_entries :
-       Super_context.t
-    -> Install.Entry.Sourced.t list Package.Name.Map.t Memo.Build.t
+    Super_context.t -> Install.Entry.Sourced.t list Package.Name.Map.t Memo.t
 end = struct
   let lib_ppxs sctx ~scope ~(lib : Dune_file.Library.t) =
     match lib.kind with
-    | Normal | Ppx_deriver _ -> Memo.Build.return []
+    | Normal | Ppx_deriver _ -> Memo.return []
     | Ppx_rewriter _ ->
       let name = Dune_file.Library.best_name lib in
       let+ ppx_exe =
-        Resolve.Build.read_memo_build (Preprocessing.ppx_exe sctx ~scope name)
+        Resolve.Memo.read_memo (Preprocessing.ppx_exe sctx ~scope name)
       in
       [ ppx_exe ]
 
@@ -63,7 +65,9 @@ end = struct
         let name = Lib_info.name lib in
         let files = Foreign_sources.for_lib foreign_sources ~name in
         Foreign.Sources.object_files files ~dir ~ext_obj
-      else Memo.Build.return (Lib_info.foreign_archives lib)
+      else
+        Memo.return
+          (Mode.Map.Multi.to_flat_list @@ Lib_info.foreign_archives lib)
     in
     List.concat_map
       ~f:(List.map ~f:(fun f -> (Section.Lib, f)))
@@ -91,7 +95,7 @@ end = struct
     let obj_dir = Lib_info.obj_dir info in
     let make_entry section ?sub_dir ?dst fn =
       let entry =
-        Install.Entry.make section fn
+        Install.Entry.make section fn ~kind:`File
           ~dst:
             (let dst =
                match dst with
@@ -130,7 +134,7 @@ end = struct
               make_entry Lib source ?dst))
     in
     let { Lib_config.has_native; ext_obj; _ } = lib_config in
-    let modes = Dune_file.Mode_conf.Set.eval lib.modes ~has_native in
+    let modes = Dune_file.Mode_conf.Set.eval lib.modes.ocaml ~has_native in
     let { Mode.Dict.byte; native } = modes in
     let module_files =
       let inside_subdir f =
@@ -143,14 +147,16 @@ end = struct
       in
       let cm_dir m cm_kind =
         let visibility = Module.visibility m in
-        let dir' = Obj_dir.cm_dir external_obj_dir cm_kind visibility in
+        let dir' = Obj_dir.cm_dir external_obj_dir (Ocaml cm_kind) visibility in
         if Path.equal (Path.build dir) dir' then None
         else Path.basename dir' |> inside_subdir |> Option.some
       in
       let virtual_library = Library.is_virtual lib in
       let modules =
         let common m =
-          let cm_file kind = Obj_dir.Module.cm_file obj_dir m ~kind in
+          let cm_file kind =
+            Obj_dir.Module.cm_file obj_dir m ~kind:(Ocaml kind)
+          in
           let if_ b (cm_kind, f) =
             if b then
               match f with
@@ -176,7 +182,10 @@ end = struct
               common m
               @ List.filter_map Ml_kind.all ~f:(fun ml_kind ->
                     let open Option.O in
-                    let+ cmt = Obj_dir.Module.cmt_file obj_dir m ~ml_kind in
+                    let+ cmt =
+                      Obj_dir.Module.cmt_file obj_dir m ~ml_kind
+                        ~cm_kind:(Ocaml Cmi)
+                    in
                     (Cm_kind.Cmi, cmt))
               |> set_dir m)
         in
@@ -193,7 +202,7 @@ end = struct
       let dll_files = dll_files ~modes ~dynlink:lib.dynlink ~ctx info in
       (lib_files, dll_files)
     in
-    let+ execs = lib_ppxs sctx ~scope ~lib in
+    let+ execs = lib_ppxs ctx ~scope ~lib in
     let install_c_headers =
       List.map lib.install_c_headers ~f:(fun base ->
           Path.Build.relative dir (base ^ Foreign_language.header_extension))
@@ -205,31 +214,31 @@ end = struct
       ; List.map lib_files ~f:(fun (section, file) -> make_entry section file)
       ; List.map execs ~f:(make_entry Libexec)
       ; List.map dll_files ~f:(fun a ->
-            let entry = Install.Entry.make Stublibs a in
+            let entry = Install.Entry.make ~kind:`File Stublibs a in
             Install.Entry.Sourced.create ~loc entry)
       ; List.map ~f:(make_entry Lib) install_c_headers
       ]
 
   let keep_if expander ~scope stanza =
     let+ keep =
+      let open Dune_file in
       match (stanza : Stanza.t) with
-      | Dune_file.Library lib ->
+      | Library lib ->
         let* enabled_if = Expander.eval_blang expander lib.enabled_if in
         if enabled_if then
           if lib.optional then
             Lib.DB.available (Scope.libs scope)
               (Dune_file.Library.best_name lib)
-          else Memo.Build.return true
-        else Memo.Build.return false
-      | Dune_file.Documentation _ -> Memo.Build.return true
-      | Dune_file.Install { enabled_if; _ } ->
-        Expander.eval_blang expander enabled_if
-      | Dune_file.Plugin _ -> Memo.Build.return true
-      | Dune_file.Executables ({ install_conf = Some _; _ } as exes) -> (
+          else Memo.return true
+        else Memo.return false
+      | Documentation _ -> Memo.return true
+      | Install { enabled_if; _ } -> Expander.eval_blang expander enabled_if
+      | Plugin _ -> Memo.return true
+      | Executables ({ install_conf = Some _; _ } as exes) -> (
         Expander.eval_blang expander exes.enabled_if >>= function
-        | false -> Memo.Build.return false
+        | false -> Memo.return false
         | true ->
-          if not exes.optional then Memo.Build.return true
+          if not exes.optional then Memo.return true
           else
             let* compile_info =
               let dune_version =
@@ -240,7 +249,7 @@ end = struct
                   exes.buildable.preprocess
                   ~instrumentation_backend:
                     (Lib.DB.instrumentation_backend (Scope.libs scope))
-                |> Resolve.Build.read_memo_build >>| Preprocess.Per_module.pps
+                |> Resolve.Memo.read_memo >>| Preprocess.Per_module.pps
               in
               Lib.DB.resolve_user_written_deps_for_exes (Scope.libs scope)
                 exes.names exes.buildable.libraries ~pps ~dune_version
@@ -248,8 +257,8 @@ end = struct
             in
             let+ requires = Lib.Compile.direct_requires compile_info in
             Resolve.is_ok requires)
-      | Coq_stanza.Theory.T d -> Memo.Build.return (Option.is_some d.package)
-      | _ -> Memo.Build.return false
+      | Coq_stanza.Theory.T d -> Memo.return (Option.is_some d.package)
+      | _ -> Memo.return false
     in
     Option.some_if keep stanza
 
@@ -257,7 +266,7 @@ end = struct
     List.exists [ "README"; "LICENSE"; "CHANGE"; "HISTORY" ] ~f:(fun prefix ->
         String.is_prefix fn ~prefix)
 
-  let stanza_to_entries ~sctx ~dir ~scope ~expander stanza =
+  let stanza_to_entries ~sites ~sctx ~dir ~scope ~expander stanza =
     let* stanza_and_package =
       let+ stanza = keep_if expander stanza ~scope in
       let open Option.O in
@@ -266,46 +275,63 @@ end = struct
       (stanza, package)
     in
     match stanza_and_package with
-    | None -> Memo.Build.return None
+    | None -> Memo.return None
     | Some (stanza, package) ->
       let new_entries =
+        let open Dune_file in
         match (stanza : Stanza.t) with
-        | Dune_file.Install i
-        | Dune_file.Executables { install_conf = Some i; _ } ->
-          let path_expander =
-            File_binding.Unexpanded.expand ~dir
-              ~f:(Expander.No_deps.expand_str expander)
-          in
+        | Install i | Executables { install_conf = Some i; _ } ->
           let section = i.section in
-          Memo.Build.List.map i.files ~f:(fun unexpanded ->
-              let* fb = path_expander unexpanded in
-              let loc = File_binding.Expanded.src_loc fb in
-              let src = File_binding.Expanded.src fb in
-              let dst = File_binding.Expanded.dst fb in
-              let+ entry =
-                Install.Entry.make_with_site section
-                  (Super_context.get_site_of_packages sctx)
-                  src ?dst
-              in
-              Install.Entry.Sourced.create ~loc entry)
-        | Dune_file.Library lib ->
+          let expand_str = Expander.No_deps.expand_str expander in
+          let* files_expanded =
+            Dune_file.Install_conf.expand_files i ~expand_str ~dir
+          in
+          let* files =
+            Memo.List.map files_expanded ~f:(fun fb ->
+                let loc = File_binding.Expanded.src_loc fb in
+                let src = File_binding.Expanded.src fb in
+                let dst = File_binding.Expanded.dst fb in
+                let+ entry =
+                  Install.Entry.make_with_site ~kind:`File section
+                    (Sites.section_of_site sites)
+                    src ?dst
+                in
+                Install.Entry.Sourced.create ~loc entry)
+          in
+          let* dirs_expanded =
+            Dune_file.Install_conf.expand_dirs i ~expand_str ~dir
+          in
+          let+ files_from_dirs =
+            Memo.List.map dirs_expanded ~f:(fun fb ->
+                let loc = File_binding.Expanded.src_loc fb in
+                let src = File_binding.Expanded.src fb in
+                let dst = File_binding.Expanded.dst fb in
+                let+ entry =
+                  Install.Entry.make_with_site section ~kind:`Directory
+                    (Sites.section_of_site sites)
+                    src ?dst
+                in
+                Install.Entry.Sourced.create ~loc entry)
+          in
+          files @ files_from_dirs
+        | Library lib ->
           let sub_dir = Dune_file.Library.sub_dir lib in
           let* dir_contents = Dir_contents.get sctx ~dir in
           lib_install_files sctx ~scope ~dir ~sub_dir lib ~dir_contents
         | Coq_stanza.Theory.T coqlib ->
           Coq_rules.install_rules ~sctx ~dir coqlib
-        | Dune_file.Documentation d ->
+        | Documentation d ->
           let* dc = Dir_contents.get sctx ~dir in
           let+ mlds = Dir_contents.mlds dc d in
           List.map mlds ~f:(fun mld ->
               let entry =
-                Install.Entry.make
+                Install.Entry.make ~kind:`File
                   ~dst:(sprintf "odoc-pages/%s" (Path.Build.basename mld))
                   Section.Doc mld
               in
               Install.Entry.Sourced.create ~loc:d.loc entry)
-        | Dune_file.Plugin t -> Plugin_rules.install_rules ~sctx ~dir t
-        | _ -> Memo.Build.return []
+        | Plugin t -> Plugin_rules.install_rules ~sctx ~sites ~dir t
+        | _ -> Memo.return []
       in
       let name = Package.name package in
       let+ entries = new_entries in
@@ -313,9 +339,10 @@ end = struct
 
   let stanzas_to_entries sctx =
     let ctx = Super_context.context sctx in
-    let stanzas = Super_context.stanzas sctx in
+    let* stanzas = Only_packages.filtered_stanzas ctx in
+    let* packages = Only_packages.get () in
     let+ init =
-      Package.Name.Map_traversals.parallel_map (Super_context.packages sctx)
+      Package.Name.Map_traversals.parallel_map packages
         ~f:(fun _name (pkg : Package.t) ->
           let init =
             let deprecated_meta_and_dune_files =
@@ -329,11 +356,11 @@ end = struct
                     Package_paths.deprecated_dune_package_file ctx pkg name
                   in
                   [ Install.Entry.Sourced.create
-                      (Install.Entry.make Lib_root meta_file
+                      (Install.Entry.make Lib_root meta_file ~kind:`File
                          ~dst:
                            (Package.Name.to_string name ^ "/" ^ Findlib.meta_fn))
                   ; Install.Entry.Sourced.create
-                      (Install.Entry.make Lib_root dune_package_file
+                      (Install.Entry.make Lib_root dune_package_file ~kind:`File
                          ~dst:
                            (Package.Name.to_string name ^ "/" ^ Dune_package.fn))
                   ])
@@ -341,15 +368,16 @@ end = struct
             let meta_file = Package_paths.meta_file ctx pkg in
             let dune_package_file = Package_paths.dune_package_file ctx pkg in
             Install.Entry.Sourced.create
-              (Install.Entry.make Lib meta_file ~dst:Findlib.meta_fn)
+              (Install.Entry.make Lib meta_file ~kind:`File ~dst:Findlib.meta_fn)
             :: Install.Entry.Sourced.create
-                 (Install.Entry.make Lib dune_package_file ~dst:Dune_package.fn)
+                 (Install.Entry.make Lib dune_package_file ~kind:`File
+                    ~dst:Dune_package.fn)
             ::
             (if not pkg.has_opam_file then deprecated_meta_and_dune_files
             else
               let opam_file = Package_paths.opam_file ctx pkg in
               Install.Entry.Sourced.create
-                (Install.Entry.make Lib opam_file ~dst:"opam")
+                (Install.Entry.make Lib opam_file ~kind:`File ~dst:"opam")
               :: deprecated_meta_and_dune_files)
           in
           let pkg_dir = Package.dir pkg in
@@ -361,18 +389,20 @@ end = struct
             |> String.Set.fold ~init ~f:(fun fn acc ->
                    if is_odig_doc_file fn then
                      let odig_file = Path.Build.relative pkg_dir fn in
-                     let entry = Install.Entry.make Doc odig_file in
+                     let entry = Install.Entry.make Doc ~kind:`File odig_file in
                      Install.Entry.Sourced.create entry :: acc
                    else acc))
     and+ l =
-      Dir_with_dune.deep_fold stanzas ~init:[] ~f:(fun d stanza acc ->
+      let* sites = Sites.create ctx in
+      Dune_file.fold_stanzas stanzas ~init:[] ~f:(fun dune_file stanza acc ->
+          let dir = Path.Build.append_source ctx.build_dir dune_file.dir in
           let named_entries =
-            let { Dir_with_dune.ctx_dir = dir; scope; _ } = d in
             let* expander = Super_context.expander sctx ~dir in
-            stanza_to_entries ~sctx ~dir ~scope ~expander stanza
+            let scope = Expander.scope expander in
+            stanza_to_entries ~sites ~sctx ~dir ~scope ~expander stanza
           in
           named_entries :: acc)
-      |> Memo.Build.all
+      |> Memo.all
     in
     List.fold_left l ~init ~f:(fun acc named_entries ->
         match named_entries with
@@ -403,7 +433,7 @@ end
 
 module Meta_and_dune_package : sig
   val meta_and_dune_package_rules :
-    Super_context.t -> Dune_project.t -> unit Memo.Build.t
+    Super_context.t -> Dune_project.t -> unit Memo.t
 end = struct
   let sections ctx_name files pkg =
     let pkg_name = Package.name pkg in
@@ -441,18 +471,18 @@ end = struct
       Path.Build.L.relative pkg_root subdir
     in
     let* entries =
-      Memo.Build.parallel_map lib_entries ~f:(fun stanza ->
+      Memo.parallel_map lib_entries ~f:(fun (stanza : Scope.DB.Lib_entry.t) ->
           match stanza with
-          | Super_context.Lib_entry.Deprecated_library_name
-              { old_name = _, Deprecated _; _ } -> Memo.Build.return None
-          | Super_context.Lib_entry.Deprecated_library_name
+          | Deprecated_library_name { old_name = _, Deprecated _; _ } ->
+            Memo.return None
+          | Deprecated_library_name
               { old_name = old_public_name, Not_deprecated
               ; new_public_name = _, new_public_name
               ; loc
               ; project = _
               } ->
             let old_public_name = Dune_file.Public_lib.name old_public_name in
-            Memo.Build.return
+            Memo.return
               (Some
                  ( old_public_name
                  , Dune_package.Entry.Deprecated_library_name
@@ -487,7 +517,7 @@ end = struct
               Lib.to_dune_lib lib
                 ~dir:(Path.build (lib_root lib))
                 ~modules ~foreign_objects
-              >>= Resolve.read_memo_build
+              >>= Resolve.read_memo
             in
             Some (name, Dune_package.Entry.Library sub_systems))
     in
@@ -520,9 +550,7 @@ end = struct
     let dune_version =
       Dune_lang.Syntax.greatest_supported_version Stanza.syntax
     in
-    let lib_entries =
-      Super_context.lib_entries_of_package sctx (Package.name pkg)
-    in
+    let* lib_entries = Scope.DB.lib_entries_of_package ctx (Package.name pkg) in
     let action =
       let dune_package_file = Package_paths.dune_package_file ctx pkg in
       let meta_template = Package_paths.meta_template ctx pkg in
@@ -532,15 +560,15 @@ end = struct
           Action_builder.if_file_exists (Path.build meta_template)
             ~then_:(Action_builder.return Dune_package.Or_meta.Use_meta)
             ~else_:
-              (Action_builder.memo_build
-                 (Memo.Build.bind (Memo.Build.return ()) ~f:(fun () ->
+              (Action_builder.of_memo
+                 (Memo.bind (Memo.return ()) ~f:(fun () ->
                       make_dune_package sctx lib_entries pkg)))
         in
         Format.asprintf "%a" (Dune_package.Or_meta.pp ~dune_version) pkg)
     in
     let deprecated_dune_packages =
       List.filter_map lib_entries ~f:(function
-        | Super_context.Lib_entry.Deprecated_library_name
+        | Deprecated_library_name
             ({ old_name = old_public_name, Deprecated _; _ } as t) ->
           Some
             ( Lib_name.package_name (Dune_file.Public_lib.name old_public_name)
@@ -549,8 +577,8 @@ end = struct
       |> Package.Name.Map.of_list_multi
     in
     let* () =
-      Package.Name.Map.foldi pkg.deprecated_package_names
-        ~init:(Memo.Build.return ()) ~f:(fun name _ acc ->
+      Package.Name.Map.foldi pkg.deprecated_package_names ~init:(Memo.return ())
+        ~f:(fun name loc acc ->
           acc
           >>>
           let dune_pkg =
@@ -595,17 +623,18 @@ end = struct
                  (Dune_package.Or_meta.pp ~dune_version)
                  (Dune_package.Or_meta.Dune_package dune_pkg))
           in
-          Super_context.add_rule sctx ~dir:ctx.build_dir action_with_targets)
+          Super_context.add_rule sctx ~dir:ctx.build_dir ~loc
+            action_with_targets)
     in
     Super_context.add_rule sctx ~dir:ctx.build_dir action
 
   let gen_meta_file sctx (pkg : Package.t) =
     let ctx = Super_context.context sctx in
     let pkg_name = Package.name pkg in
-    let deprecated_packages, entries =
-      let entries = Super_context.lib_entries_of_package sctx pkg_name in
+    let* deprecated_packages, entries =
+      let+ entries = Scope.DB.lib_entries_of_package ctx pkg_name in
       List.partition_map entries ~f:(function
-        | Super_context.Lib_entry.Deprecated_library_name
+        | Deprecated_library_name
             { old_name = public, Deprecated { deprecated_package }; _ } as entry
           -> (
           match Dune_file.Public_lib.sub_dir public with
@@ -621,7 +650,7 @@ end = struct
            however. *)
         let vlib =
           List.find_map entries ~f:(function
-            | Super_context.Lib_entry.Library lib ->
+            | Library lib ->
               let info = Lib.Local.info lib in
               Option.some_if (Option.is_some (Lib_info.virtual_ info)) lib
             | Deprecated_library_name _ -> None)
@@ -654,7 +683,7 @@ end = struct
         (let open Action_builder.O in
         (let* template = template in
          let+ meta =
-           Action_builder.memo_build
+           Action_builder.of_memo
              (Gen_meta.gen ~package:pkg ~add_directory_entry:true entries)
          in
          let pp =
@@ -675,9 +704,9 @@ end = struct
       Package.Name.Map.of_list_multi deprecated_packages
     in
     Package.Name.Map_traversals.parallel_iter pkg.deprecated_package_names
-      ~f:(fun name _loc ->
+      ~f:(fun name loc ->
         let meta = Package_paths.deprecated_meta_file ctx pkg name in
-        Super_context.add_rule sctx ~dir:ctx.build_dir
+        Super_context.add_rule sctx ~dir:ctx.build_dir ~loc
           (Action_builder.write_file_dyn meta
              (let open Action_builder.O in
              let+ meta =
@@ -686,7 +715,7 @@ end = struct
                  | None -> []
                  | Some entries -> entries
                in
-               Action_builder.memo_build
+               Action_builder.of_memo
                  (Gen_meta.gen ~package:pkg entries ~add_directory_entry:false)
              in
              let pp =
@@ -725,7 +754,10 @@ let symlink_installed_artifacts_to_build_install sctx
       in
       let rule =
         let { Action_builder.With_targets.targets; build } =
-          Action_builder.symlink ~src:(Path.build entry.src) ~dst
+          (match entry.kind with
+          | `File -> Action_builder.symlink
+          | `Directory -> Action_builder.symlink_dir)
+            ~src:(Path.build entry.src) ~dst
         in
         Rule.make
           ~info:(Rule.Info.of_loc_opt (Some loc))
@@ -747,9 +779,10 @@ let install_entries sctx (package : Package.t) =
 
 let packages =
   let f sctx =
-    let packages = Package.Name.Map.values (Super_context.packages sctx) in
+    let* packages = Only_packages.get () in
+    let packages = Package.Name.Map.values packages in
     let+ l =
-      Memo.Build.parallel_map packages ~f:(fun (pkg : Package.t) ->
+      Memo.parallel_map packages ~f:(fun (pkg : Package.t) ->
           install_entries sctx pkg
           >>| List.map ~f:(fun (e : Install.Entry.Sourced.t) ->
                   (e.entry.src, pkg.id)))
@@ -758,31 +791,29 @@ let packages =
       ~f:Package.Id.Set.add
   in
   let memo =
-    Memo.create "package-map"
+    Memo.create "package-map" f
       ~input:(module Super_context.As_memo_key)
       ~cutoff:(Path.Build.Map.equal ~equal:Package.Id.Set.equal)
-      f
   in
   fun sctx -> Memo.exec memo sctx
 
 let packages_file_is_part_of path =
-  Memo.Build.Option.bind
+  Memo.Option.bind
     (let open Option.O in
     let* ctx_name, _ = Path.Build.extract_build_context path in
     Context_name.of_string_opt ctx_name)
     ~f:Super_context.find
   >>= function
-  | None -> Memo.Build.return Package.Id.Set.empty
+  | None -> Memo.return Package.Id.Set.empty
   | Some sctx ->
-    let open Memo.Build.O in
+    let open Memo.O in
     let+ map = packages sctx in
     Option.value (Path.Build.Map.find map path) ~default:Package.Id.Set.empty
 
 let symlinked_entries sctx package =
   let package_name = Package.name package in
-  let install_paths =
-    Install.Section.Paths.make ~package:package_name ~destdir:Path.root ()
-  in
+  let roots = Install.Section.Paths.Roots.opam_from_prefix Path.root in
+  let install_paths = Install.Section.Paths.make ~package:package_name ~roots in
   let+ entries = install_entries sctx package in
   entries
   |> symlink_installed_artifacts_to_build_install sctx ~install_paths
@@ -800,31 +831,134 @@ let symlinked_entries =
   in
   fun sctx pkg -> Memo.exec memo (sctx, pkg)
 
+let package_deps (pkg : Package.t) files =
+  let rec loop rules_seen (fn : Path.Build.t) =
+    let* pkgs = packages_file_is_part_of fn in
+    if Package.Id.Set.is_empty pkgs || Package.Id.Set.mem pkgs pkg.id then
+      loop_deps rules_seen fn
+    else Memo.return (pkgs, rules_seen)
+  and loop_deps rules_seen fn =
+    Load_rules.get_rule (Path.build fn) >>= function
+    | None -> Memo.return (Package.Id.Set.empty, rules_seen)
+    | Some rule ->
+      if Rule.Set.mem rules_seen rule then
+        Memo.return (Package.Id.Set.empty, rules_seen)
+      else
+        let rules_seen = Rule.Set.add rules_seen rule in
+        let* res = Dune_engine.Build_system.execute_rule rule in
+        loop_files rules_seen
+          (Dep.Facts.paths res.deps |> Path.Map.keys
+          |> (* if this file isn't in the build dir, it doesn't belong to any
+                package and it doesn't have dependencies that do *)
+          List.filter_map ~f:Path.as_in_build_dir)
+  and loop_files rules_seen files =
+    Memo.List.fold_left ~init:(Package.Id.Set.empty, rules_seen) files
+      ~f:(fun (sets, rules_seen) file ->
+        let+ set, rules_seen = loop rules_seen file in
+        (Package.Id.Set.union set sets, rules_seen))
+  in
+  let+ packages, _rules_seen = loop_files Rule.Set.empty files in
+  packages
+
+include (
+  struct
+    module Spec = struct
+      type ('path, 'target) t = Path.t Install.Entry.t list * 'target
+
+      let name = "gen-install-file"
+
+      let version = 1
+
+      let bimap (entries, dst) _ g = (entries, g dst)
+
+      let is_useful_to ~distribute:_ ~memoize = memoize
+
+      let encode (_entries, dst) _path target : Dune_lang.t =
+        List [ Dune_lang.atom_or_quoted_string name; target dst ]
+
+      let read_dir_recursively (entry : _ Install.Entry.t) =
+        let rec loop acc dirs =
+          match dirs with
+          | [] ->
+            List.rev_map acc ~f:(fun (path, comps) ->
+                let comps = List.rev comps in
+                Install.Entry.set_src entry path
+                |> Install.Entry.map_dst ~f:(fun dst ->
+                       Install.Dst.concat_all dst comps))
+            |> List.sort
+                 ~compare:(fun (x : _ Install.Entry.t) (y : _ Install.Entry.t)
+                          -> Path.compare x.src y.src)
+          | (dir, comps) :: dirs -> (
+            match Path.Untracked.readdir_unsorted_with_kinds dir with
+            | Error _ -> Code_error.raise "unable to read directory" []
+            | Ok files ->
+              let files, new_dirs =
+                List.partition_map files ~f:(fun (name, kind) ->
+                    let path = Path.relative dir name in
+                    let comps = name :: comps in
+                    match kind with
+                    | Unix.S_DIR -> Right (path, comps)
+                    | _ -> Left (path, comps))
+              in
+              let acc = List.rev_append files acc in
+              let dirs = List.rev_append new_dirs dirs in
+              loop acc dirs)
+        in
+        loop [] [ (entry.src, []) ]
+
+      let action (entries, dst) ~ectx:_ ~eenv:_ =
+        let entries =
+          List.concat_map entries ~f:(fun (entry : _ Install.Entry.t) ->
+              match entry.kind with
+              | `File -> [ entry ]
+              | `Directory -> read_dir_recursively entry)
+          |> Install.gen_install_file
+        in
+        Io.write_file (Path.build dst) entries;
+        Fiber.return ()
+    end
+
+    let gen_install_file entries ~dst =
+      let module M = struct
+        type path = Path.t
+
+        type target = Path.Build.t
+
+        module Spec = Spec
+
+        let v = (entries, dst)
+      end in
+      Dune_engine.Action.Extension (module M)
+  end :
+    sig
+      val gen_install_file :
+        Path.t Install.Entry.t list -> dst:Path.Build.t -> Action.t
+    end)
+
 let gen_package_install_file_rules sctx (package : Package.t) =
   let package_name = Package.name package in
-  let install_paths =
-    Install.Section.Paths.make ~package:package_name ~destdir:Path.root ()
+  let roots = Install.Section.Paths.Roots.opam_from_prefix Path.root in
+  let install_paths = Install.Section.Paths.make ~package:package_name ~roots in
+  let entries =
+    Action_builder.of_memo (symlinked_entries sctx package >>| fst)
   in
-  let* entries = symlinked_entries sctx package >>| fst in
   let ctx = Super_context.context sctx in
   let pkg_build_dir = Package_paths.build_dir ctx package in
   let files =
-    List.map entries ~f:(fun (e : Install.Entry.Sourced.t) -> e.entry.src)
+    Action_builder.map entries
+      ~f:(List.map ~f:(fun (e : Install.Entry.Sourced.t) -> e.entry.src))
   in
-  let dune_project =
-    let scope = Super_context.find_scope_by_dir sctx pkg_build_dir in
+  let* dune_project =
+    let+ scope = Scope.DB.find_by_dir pkg_build_dir in
     Scope.project scope
   in
   let strict_package_deps = Dune_project.strict_package_deps dune_project in
   let packages =
     let open Action_builder.O in
-    let+ packages =
-      Action_builder.memo_build
-        (Build_system.package_deps package files
-           ~packages_of:packages_file_is_part_of)
-    in
-    match strict_package_deps with
-    | false -> packages
+    let* files = files in
+    let+ packages = Action_builder.of_memo (package_deps package files) in
+    (match strict_package_deps with
+    | false -> ()
     | true ->
       let missing_deps =
         let effective_deps =
@@ -833,8 +967,7 @@ let gen_package_install_file_rules sctx (package : Package.t) =
         in
         Package.missing_deps package ~effective_deps
       in
-      if Package.Name.Set.is_empty missing_deps then packages
-      else
+      if not (Package.Name.Set.is_empty missing_deps) then
         let name = Package.name package in
         User_error.raise
           [ Pp.textf "Package %s is missing the following package dependencies"
@@ -842,74 +975,83 @@ let gen_package_install_file_rules sctx (package : Package.t) =
           ; Package.Name.Set.to_list missing_deps
             |> Pp.enumerate ~f:(fun name ->
                    Pp.text (Package.Name.to_string name))
-          ]
+          ]);
+    packages
+  in
+  let install_file_deps =
+    let open Action_builder.O in
+    let* files = files in
+    Path.Set.of_list_map files ~f:Path.build |> Action_builder.path_set
   in
-  let files = Path.Set.of_list_map files ~f:Path.build in
   let* () =
+    let* all_packages = Only_packages.get () in
     let context = Context.build_context ctx in
-    let target_alias = Alias.package_install ~context ~pkg:package in
+    let target_alias = Dep_conf_eval.package_install ~context ~pkg:package in
     let open Action_builder.O in
     Rules.Produce.Alias.add_deps target_alias
       (Action_builder.dyn_deps
          (let+ packages = packages
-          and+ () = Action_builder.deps (Dep.Set.of_files_set files) in
+          and+ () = install_file_deps in
           ( ()
           , Package.Id.Set.to_list packages
             |> Dep.Set.of_list_map ~f:(fun (pkg : Package.Id.t) ->
                    let pkg =
                      let name = Package.Id.name pkg in
-                     Package.Name.Map.find_exn
-                       (Super_context.packages sctx)
-                       name
+                     Package.Name.Map.find_exn all_packages name
                    in
-                   Alias.package_install ~context ~pkg |> Dep.alias) )))
+                   Dep_conf_eval.package_install ~context ~pkg |> Dep.alias) )))
   in
   let action =
     let install_file =
       Path.Build.relative pkg_build_dir
-        (Utils.install_file ~package:package_name
+        (install_file ~package:package_name
            ~findlib_toolchain:ctx.findlib_toolchain)
     in
     let open Action_builder.O in
-    Action_builder.write_file_dyn install_file
-      (let+ () = Action_builder.path_set files
-       and+ () =
-         if strict_package_deps then
-           Action_builder.map packages ~f:(fun (_ : Package.Id.Set.t) -> ())
-         else Action_builder.return ()
-       in
-       let entries =
-         match ctx.findlib_toolchain with
-         | None -> entries
-         | Some toolchain ->
-           let toolchain = Context_name.to_string toolchain in
-           let prefix = Path.of_string (toolchain ^ "-sysroot") in
-           List.map entries ~f:(fun (e : Install.Entry.Sourced.t) ->
-               { e with
-                 entry =
-                   Install.Entry.add_install_prefix e.entry ~paths:install_paths
-                     ~prefix
-               })
-       in
-       (if not package.allow_empty then
-        if
-          List.for_all entries ~f:(fun (e : Install.Entry.Sourced.t) ->
-              match e.source with
-              | Dune -> true
-              | User _ -> false)
-        then
-          let is_error = Dune_project.dune_version dune_project >= (3, 0) in
-          User_warning.emit ~is_error
-            [ Pp.textf
-                "The package %s does not have any user defined stanzas \
-                 attached to it. If this is intentional, add (allow_empty) to \
-                 the package definition in the dune-project file"
-                (Package.Name.to_string package_name)
-            ]);
-       Install.gen_install_file
-         (List.map entries ~f:(fun (e : Install.Entry.Sourced.t) ->
-              Install.Entry.set_src e.entry (Path.build e.entry.src))))
+    let entries =
+      let+ () = install_file_deps
+      and+ () =
+        if strict_package_deps then
+          Action_builder.map packages ~f:(fun (_ : Package.Id.Set.t) -> ())
+        else Action_builder.return ()
+      and+ entries = entries in
+      let entries =
+        match ctx.findlib_toolchain with
+        | None -> entries
+        | Some toolchain ->
+          let toolchain = Context_name.to_string toolchain in
+          let prefix = Path.of_string (toolchain ^ "-sysroot") in
+          List.map entries ~f:(fun (e : Install.Entry.Sourced.t) ->
+              { e with
+                entry =
+                  Install.Entry.add_install_prefix e.entry ~paths:install_paths
+                    ~prefix
+              })
+      in
+      (if not package.allow_empty then
+       if
+         List.for_all entries ~f:(fun (e : Install.Entry.Sourced.t) ->
+             match e.source with
+             | Dune -> true
+             | User _ -> false)
+       then
+         let is_error = Dune_project.dune_version dune_project >= (3, 0) in
+         User_warning.emit ~is_error
+           [ Pp.textf
+               "The package %s does not have any user defined stanzas attached \
+                to it. If this is intentional, add (allow_empty) to the \
+                package definition in the dune-project file"
+               (Package.Name.to_string package_name)
+           ]);
+      List.map entries ~f:(fun (e : Install.Entry.Sourced.t) ->
+          Install.Entry.set_src e.entry (Path.build e.entry.src))
+    in
+    Action_builder.with_file_targets ~file_targets:[ install_file ]
+      (let+ entries = entries in
+       let action = gen_install_file entries ~dst:install_file in
+       Action.Full.make action)
   in
+
   Super_context.add_rule sctx ~dir:pkg_build_dir
     ~mode:
       (if promote_install_file ctx then
@@ -928,7 +1070,7 @@ let memo =
         (Package.Name.to_string (Package.name pkg)))
     "install-rules-and-pkg-entries"
     (fun (sctx, pkg) ->
-      Memo.Build.return
+      Memo.return
         (let ctx = Super_context.context sctx in
          let context_name = ctx.name in
          Scheme.Approximation
@@ -946,8 +1088,9 @@ let scheme_per_ctx_memo =
     ~input:(module Super_context.As_memo_key)
     "install-rule-scheme"
     (fun sctx ->
-      let packages = Package.Name.Map.values (Super_context.packages sctx) in
-      let* schemes = Memo.Build.sequential_map packages ~f:(scheme sctx) in
+      let* packages = Only_packages.get () in
+      let packages = Package.Name.Map.values packages in
+      let* schemes = Memo.sequential_map packages ~f:(scheme sctx) in
       Scheme.evaluate ~union:Rules.Dir_rules.union (Scheme.all schemes))
 
 let symlink_rules sctx ~dir =
@@ -963,10 +1106,10 @@ let symlink_rules sctx ~dir =
 let gen_install_alias sctx (package : Package.t) =
   let ctx = Super_context.context sctx in
   let name = Package.name package in
-  if ctx.implicit then Memo.Build.return ()
+  if ctx.implicit then Memo.return ()
   else
     let install_fn =
-      Utils.install_file ~package:name ~findlib_toolchain:ctx.findlib_toolchain
+      install_file ~package:name ~findlib_toolchain:ctx.findlib_toolchain
     in
     let path = Package_paths.build_dir ctx package in
     let install_alias = Alias.install ~dir:path in
