@@ -1,5 +1,4 @@
-open! Dune_engine
-open! Stdune
+open Import
 
 module Flags : sig
   type 'flags t =
@@ -45,6 +44,17 @@ module In_buildable : sig
   val default : t
 end
 
+module In_context : sig
+  type t =
+    { flags : Flags.Spec.t
+    ; javascript_files : Path.Build.t list
+    }
+
+  val make : dir:Path.Build.t -> In_buildable.t -> t
+
+  val default : t
+end
+
 module Compilation_mode : sig
   type t =
     | Whole_program
