@@ -1,5 +1,6 @@
 (** Raw library descriptions *)
-open! Dune_engine
+
+open Import
 
 (** This module regroup all information about a library. We call such
     descriptions "raw" as the names, such as the names of dependencies are plain
@@ -7,8 +8,6 @@ open! Dune_engine
 
     The [Lib] module takes care of resolving library names to actual libraries. *)
 
-open Stdune
-
 module Status : sig
   type t =
     | Installed_private
@@ -93,8 +92,10 @@ val archives : 'path t -> 'path list Mode.Dict.t
 (* TODO: Rename [foreign_archives] to [foreign_lib_files] and [native_archives]
    to [native_lib_files] for consistent naming with [foreign_dll_files]. *)
 
-(** All the [lib*.a] files for stubs *)
-val foreign_archives : 'path t -> 'path list
+(** All the [lib*.a] files for stubs. A table indexed by [Mode.Select.t] is used
+    to account for mode-dependent archives. The special key [Mode.Select.All] is
+    used for archives that should be built in every modes. *)
+val foreign_archives : 'path t -> 'path Mode.Map.Multi.t
 
 type 'path native_archives =
   | Needs_module_info of 'path
@@ -147,7 +148,7 @@ val wrapped : _ t -> Wrapped.t Inherited.t option
 
 val special_builtin_support : _ t -> Special_builtin_support.t option
 
-val modes : _ t -> Mode.Dict.Set.t
+val modes : _ t -> Lib_mode.Map.Set.t
 
 val implements : _ t -> (Loc.t * Lib_name.t) option
 
@@ -221,7 +222,7 @@ val create :
   -> plugins:'a list Mode.Dict.t
   -> archives:'a list Mode.Dict.t
   -> ppx_runtime_deps:(Loc.t * Lib_name.t) list
-  -> foreign_archives:'a list
+  -> foreign_archives:'a Mode.Map.Multi.t
   -> native_archives:'a native_archives
   -> foreign_dll_files:'a list
   -> jsoo_runtime:'a list
@@ -234,7 +235,7 @@ val create :
   -> entry_modules:Module_name.t list Or_exn.t Source.t
   -> implements:(Loc.t * Lib_name.t) option
   -> default_implementation:(Loc.t * Lib_name.t) option
-  -> modes:Mode.Dict.Set.t
+  -> modes:Lib_mode.Map.Set.t
   -> wrapped:Wrapped.t Inherited.t option
   -> special_builtin_support:Special_builtin_support.t option
   -> exit_module:Module_name.t option
