@@ -1,8 +1,5 @@
-open! Dune_engine
-open! Stdune
 open Import
-open! No_io
-open Memo.Build.O
+open Memo.O
 module Buildable = Dune_file.Buildable
 module Library = Dune_file.Library
 module Mode_conf = Dune_file.Mode_conf
@@ -18,20 +15,37 @@ let msvc_hack_cclibs =
 
 (* Build an OCaml library. *)
 let build_lib (lib : Library.t) ~native_archives ~sctx ~expander ~flags ~dir
-    ~mode ~cm_files ~scope =
+    ~mode ~cm_files =
   let ctx = Super_context.context sctx in
-  Memo.Build.Result.iter (Context.compiler ctx mode) ~f:(fun compiler ->
+  Memo.Result.iter (Context.compiler ctx mode) ~f:(fun compiler ->
       let target = Library.archive lib ~dir ~ext:(Mode.compiled_lib_ext mode) in
       let stubs_flags =
-        List.concat_map (Library.foreign_archives lib) ~f:(fun archive ->
-            let lname =
-              "-l" ^ Foreign.Archive.(name archive |> Name.to_string)
-            in
-            let cclib = [ "-cclib"; lname ] in
-            let dllib = [ "-dllib"; lname ] in
-            match mode with
-            | Native -> cclib
-            | Byte -> dllib @ cclib)
+        let lib_archive = Library.stubs_archive lib in
+        let foreign_archives = Library.foreign_archives lib in
+        let make_args ~stub_mode archive =
+          let lname =
+            "-l"
+            ^ Foreign.Archive.(name ~mode:stub_mode archive |> Name.to_string)
+          in
+          let cclib = [ "-cclib"; lname ] in
+          let dllib = [ "-dllib"; lname ] in
+          match mode with
+          | Native -> cclib
+          | Byte -> dllib @ cclib
+        in
+        let stub_mode =
+          if Buildable.has_mode_dependent_foreign_stubs lib.buildable then
+            Mode.Select.Only mode
+          else Mode.Select.All
+        in
+        let foreign_archives =
+          List.concat_map foreign_archives
+            ~f:(make_args ~stub_mode:Mode.Select.All)
+        in
+        match lib_archive with
+        | Some lib_archive ->
+          make_args ~stub_mode lib_archive @ foreign_archives
+        | None -> foreign_archives
       in
       let map_cclibs =
         (* https://github.com/ocaml/dune/issues/119 *)
@@ -43,11 +57,8 @@ let build_lib (lib : Library.t) ~native_archives ~sctx ~expander ~flags ~dir
         Action_builder.paths (Cm_files.unsorted_objects_and_cms cm_files ~mode)
       in
       let ocaml_flags = Ocaml_flags.get flags mode in
-      let standard =
-        let project =
-          Super_context.find_scope_by_dir sctx dir |> Scope.project
-        in
-
+      let* standard =
+        let+ project = Scope.DB.find_by_dir dir |> Memo.map ~f:Scope.project in
         match Dune_project.use_standard_c_and_cxx_flags project with
         | Some true when Buildable.has_foreign_cxx lib.buildable ->
           Cxx_flags.get_flags ~for_:Link ctx
@@ -61,8 +72,7 @@ let build_lib (lib : Library.t) ~native_archives ~sctx ~expander ~flags ~dir
         Expander.expand_and_eval_set expander lib.library_flags ~standard
       in
       let ctypes_cclib_flags =
-        Ctypes_rules.ctypes_cclib_flags ~scope ~standard ~expander
-          ~buildable:lib.buildable
+        Ctypes_rules.ctypes_cclib_flags sctx ~expander ~buildable:lib.buildable
       in
       Super_context.add_rule ~dir sctx ~loc:lib.buildable.loc
         (let open Action_builder.With_targets.O in
@@ -91,6 +101,9 @@ let build_lib (lib : Library.t) ~native_archives ~sctx ~expander ~flags ~dir
               ; Dyn
                   (Action_builder.map ctypes_cclib_flags ~f:(fun x ->
                        Command.quote_args "-cclib" (map_cclibs x)))
+              ; Deps
+                  (Foreign.Objects.build_paths lib.buildable.extra_objects
+                     ~ext_obj:ctx.lib_config.ext_obj ~dir)
               ]))
 
 let gen_wrapped_compat_modules (lib : Library.t) cctx =
@@ -124,11 +137,11 @@ let gen_wrapped_compat_modules (lib : Library.t) cctx =
 
 (* Rules for building static and dynamic libraries using [ocamlmklib]. *)
 let ocamlmklib ~loc ~c_library_flags ~sctx ~dir ~o_files ~archive_name
-    ~build_targets_together =
+    ~stubs_mode ~build_targets_together =
   let ctx = Super_context.context sctx in
   let { Lib_config.ext_lib; ext_dll; _ } = ctx.lib_config in
   let static_target =
-    Foreign.Archive.Name.lib_file archive_name ~dir ~ext_lib
+    Foreign.Archive.Name.lib_file archive_name ~dir ~ext_lib ~mode:stubs_mode
   in
   let cclibs =
     Action_builder.map c_library_flags ~f:(fun cclibs ->
@@ -148,7 +161,9 @@ let ocamlmklib ~loc ~c_library_flags ~sctx ~dir ~o_files ~archive_name
         [ A "-g"
         ; (if custom then A "-custom" else Command.Args.empty)
         ; A "-o"
-        ; Path (Path.build (Foreign.Archive.Name.path ~dir archive_name))
+        ; Path
+            (Path.build
+               (Foreign.Archive.Name.path ~dir archive_name ~mode:stubs_mode))
         ; Deps o_files
           (* The [c_library_flags] is needed only for the [dynamic_target] case,
              but we pass them unconditionally for simplicity. *)
@@ -158,7 +173,7 @@ let ocamlmklib ~loc ~c_library_flags ~sctx ~dir ~o_files ~archive_name
       >>| Action.Full.add_sandbox sandbox)
   in
   let dynamic_target =
-    Foreign.Archive.Name.dll_file archive_name ~dir ~ext_dll
+    Foreign.Archive.Name.dll_file archive_name ~dir ~ext_dll ~mode:stubs_mode
   in
   if build_targets_together then
     (* Build both the static and dynamic targets in one [ocamlmklib] invocation,
@@ -168,7 +183,7 @@ let ocamlmklib ~loc ~c_library_flags ~sctx ~dir ~o_files ~archive_name
        [ static_target; dynamic_target ]
       else [ static_target ])
   else
-    let open Memo.Build.O in
+    let open Memo.O in
     (* Build the static target only by passing the [-custom] flag. *)
     let* () =
       build ~sandbox:Sandbox_config.no_special_requirements ~custom:true
@@ -185,7 +200,7 @@ let ocamlmklib ~loc ~c_library_flags ~sctx ~dir ~o_files ~archive_name
        "optional targets", allowing us to run [ocamlmklib] with the [-failsafe]
        flag, which always produces the static target and sometimes produces the
        dynamic target too. *)
-    Memo.Build.when_ ctx.dynamically_linked_foreign_archives (fun () ->
+    Memo.when_ ctx.dynamically_linked_foreign_archives (fun () ->
         build ~sandbox:Sandbox_config.needs_sandboxing ~custom:false
           [ dynamic_target ])
 
@@ -200,15 +215,19 @@ let foreign_rules (library : Foreign.Library.t) ~sctx ~expander ~dir
     >>| Foreign_sources.for_archive ~archive_name
   in
   let* o_files =
-    Foreign_rules.build_o_files ~sctx ~dir ~expander
-      ~requires:(Resolve.return []) ~dir_contents ~foreign_sources
-    |> Memo.Build.parallel_map ~f:(Memo.Build.map ~f:Path.build)
+    let+ o_files_by_mode =
+      Foreign_rules.build_o_files ~sctx ~dir ~expander
+        ~requires:(Resolve.return []) ~dir_contents ~foreign_sources
+    in
+    Mode.Map.Multi.for_all_modes o_files_by_mode
   in
   let* () = Check_rules.add_files sctx ~dir o_files in
-  let standard =
-    let project = Super_context.find_scope_by_dir sctx dir |> Scope.project in
+  let* standard =
+    let+ project =
+      let+ scope = Scope.DB.find_by_dir dir in
+      Scope.project scope
+    in
     let ctx = Super_context.context sctx in
-
     match Dune_project.use_standard_c_and_cxx_flags project with
     | Some true when Foreign.Sources.has_cxx_sources foreign_sources ->
       Cxx_flags.get_flags ~for_:Link ctx
@@ -219,7 +238,7 @@ let foreign_rules (library : Foreign.Library.t) ~sctx ~expander ~dir
       ~standard
   in
   ocamlmklib ~archive_name ~loc:library.stubs.loc ~c_library_flags ~sctx ~dir
-    ~o_files ~build_targets_together:false
+    ~o_files ~build_targets_together:false ~stubs_mode:Mode.Select.All
 
 (* Build a required set of archives for an OCaml library. *)
 let build_stubs lib ~cctx ~dir ~expander ~requires ~dir_contents
@@ -230,25 +249,32 @@ let build_stubs lib ~cctx ~dir ~expander ~requires ~dir_contents
     let name = Library.best_name lib in
     Foreign_sources.for_lib foreign_sources ~name
   in
-  let* lib_o_files =
-    Foreign_rules.build_o_files ~sctx ~dir ~expander ~requires ~dir_contents
-      ~foreign_sources
-    |> Memo.Build.parallel_map ~f:(Memo.Build.map ~f:Path.build)
+  let* o_files =
+    let lib_foreign_o_files =
+      let { Lib_config.ext_obj; _ } = (Super_context.context sctx).lib_config in
+      Foreign.Objects.build_paths lib.buildable.extra_objects ~ext_obj ~dir
+    in
+    let+ tbl =
+      Foreign_rules.build_o_files ~sctx ~dir ~expander ~requires ~dir_contents
+        ~foreign_sources
+    in
+    Mode.Map.Multi.add_all tbl Mode.Select.All lib_foreign_o_files
   in
-  let* () = Check_rules.add_files sctx ~dir lib_o_files in
-  match vlib_stubs_o_files @ lib_o_files with
-  | [] -> Memo.Build.return ()
-  | o_files ->
+  let all_o_files = Mode.Map.Multi.to_flat_list o_files in
+  let* () = Check_rules.add_files sctx ~dir all_o_files in
+  if List.for_all ~f:List.is_empty [ all_o_files; vlib_stubs_o_files ] then
+    Memo.return ()
+  else
     let ctx = Super_context.context sctx in
     let lib_name = Lib_name.Local.to_string (snd lib.name) in
     let archive_name = Foreign.Archive.Name.stubs lib_name in
     let modes = Compilation_context.modes cctx in
     let build_targets_together =
-      modes.native && modes.byte
+      modes.ocaml.native && modes.ocaml.byte
       && Dynlink_supported.get lib.dynlink ctx.supports_shared_libraries
     in
-    let standard =
-      let project = Super_context.find_scope_by_dir sctx dir |> Scope.project in
+    let* standard =
+      let+ project = Scope.DB.find_by_dir dir >>| Scope.project in
       match Dune_project.use_standard_c_and_cxx_flags project with
       | Some true when Foreign.Sources.has_cxx_sources foreign_sources ->
         Cxx_flags.get_flags ~for_:Link ctx
@@ -257,12 +283,37 @@ let build_stubs lib ~cctx ~dir ~expander ~requires ~dir_contents
     let c_library_flags =
       Expander.expand_and_eval_set expander lib.c_library_flags ~standard
     in
-    ocamlmklib ~archive_name ~loc:lib.buildable.loc ~sctx ~dir ~o_files
-      ~c_library_flags ~build_targets_together
+    let lib_o_files_for_all_modes = Mode.Map.Multi.for_all_modes o_files in
+    let for_all_modes =
+      List.rev_append vlib_stubs_o_files lib_o_files_for_all_modes
+    in
+    if
+      Mode.Dict.Set.to_list modes.ocaml
+      |> List.for_all ~f:(fun mode ->
+             List.is_empty
+             @@ Mode.Map.Multi.for_only ~and_all:false o_files mode)
+    then
+      (* if stubs are not mode dependent *)
+      let o_files = for_all_modes in
+      ocamlmklib ~archive_name ~loc:lib.buildable.loc ~sctx ~dir ~o_files
+        ~c_library_flags ~build_targets_together ~stubs_mode:Mode.Select.All
+    else
+      let modes =
+        Mode.Dict.Set.to_list modes.ocaml
+        |> List.map ~f:(fun mode ->
+               let o_files_for_mode =
+                 Mode.Map.Multi.for_only ~and_all:false o_files mode
+               in
+               ( List.rev_append for_all_modes o_files_for_mode
+               , Mode.Select.Only mode ))
+      in
+      Memo.parallel_iter modes ~f:(fun (o_files, stubs_mode) ->
+          ocamlmklib ~archive_name ~loc:lib.buildable.loc ~sctx ~dir ~o_files
+            ~c_library_flags ~build_targets_together ~stubs_mode)
 
 let build_shared lib ~native_archives ~sctx ~dir ~flags =
   let ctx = Super_context.context sctx in
-  Memo.Build.Result.iter ctx.ocamlopt ~f:(fun ocamlopt ->
+  Memo.Result.iter ctx.ocamlopt ~f:(fun ocamlopt ->
       let ext_lib = ctx.lib_config.ext_lib in
       let src =
         let ext = Mode.compiled_lib_ext Native in
@@ -283,6 +334,9 @@ let build_shared lib ~native_archives ~sctx ~dir ~flags =
         Action_builder.with_no_targets
           (Action_builder.paths
              (Library.foreign_lib_files lib ~dir ~ext_lib
+                ~for_mode:(Mode.Select.Only Byte)
+              @ Library.foreign_lib_files lib ~dir ~ext_lib
+                  ~for_mode:Mode.Select.All
              |> List.map ~f:Path.build))
         >>> Command.run ~dir:(Path.build ctx.build_dir) (Ok ocamlopt)
               [ Command.Args.dyn (Ocaml_flags.get flags Native)
@@ -303,28 +357,32 @@ let build_shared lib ~native_archives ~sctx ~dir ~flags =
       in
       Super_context.add_rule sctx build ~dir ~loc:lib.buildable.loc)
 
-let setup_build_archives (lib : Dune_file.Library.t) ~cctx ~expander ~scope =
+let setup_build_archives (lib : Dune_file.Library.t) ~top_sorted_modules ~cctx
+    ~expander ~lib_info =
   let obj_dir = Compilation_context.obj_dir cctx in
+  let dir = Compilation_context.dir cctx in
   let flags = Compilation_context.flags cctx in
   let modules = Compilation_context.modules cctx in
-  let js_of_ocaml = lib.buildable.js_of_ocaml in
+  let js_of_ocaml =
+    Js_of_ocaml.In_context.make ~dir lib.buildable.js_of_ocaml
+  in
   let sctx = Compilation_context.super_context cctx in
   let ctx = Compilation_context.context cctx in
   let { Lib_config.ext_obj; natdynlink_supported; _ } = ctx.lib_config in
-  let impl_only = Modules.impl_only modules in
-  let open Memo.Build.O in
+  let open Memo.O in
   let* () =
     Modules.exit_module modules
-    |> Memo.Build.Option.iter ~f:(fun m ->
+    |> Memo.Option.iter ~f:(fun m ->
            (* These files needs to be alongside stdlib.cma as the compiler
               implicitly adds this module. *)
            [ (Cm_kind.Cmx, Cm_kind.ext Cmx)
            ; (Cmo, Cm_kind.ext Cmo)
            ; (Cmx, ext_obj)
            ]
-           |> Memo.Build.parallel_iter ~f:(fun (kind, ext) ->
+           |> Memo.parallel_iter ~f:(fun (kind, ext) ->
                   let src =
-                    Path.build (Obj_dir.Module.obj_file obj_dir m ~kind ~ext)
+                    Path.build
+                      (Obj_dir.Module.obj_file obj_dir m ~kind:(Ocaml kind) ~ext)
                   in
                   let obj_name = Module.obj_name m in
                   let fname =
@@ -333,23 +391,15 @@ let setup_build_archives (lib : Dune_file.Library.t) ~cctx ~expander ~scope =
                   (* XXX we should get the directory from the dir of the cma
                      file explicitly *)
                   let dst = Path.Build.relative (Obj_dir.dir obj_dir) fname in
-                  Super_context.add_rule sctx
-                    ~dir:(Compilation_context.dir cctx)
-                    ~loc:lib.buildable.loc
+                  Super_context.add_rule sctx ~dir ~loc:lib.buildable.loc
                     (Action_builder.copy ~src ~dst)))
   in
-  let top_sorted_modules =
-    Dep_graph.top_closed_implementations
-      (Compilation_context.dep_graphs cctx).impl impl_only
-  in
   let modes = Compilation_context.modes cctx in
   (* The [dir] below is used as an object directory without going through
      [Obj_dir]. That's fragile and will break if the layout of the object
      directory changes *)
   let dir = Obj_dir.dir obj_dir in
-  let* native_archives =
-    let lib_config = ctx.lib_config in
-    let+ lib_info = Library.to_lib_info lib ~dir ~lib_config in
+  let native_archives =
     Lib_info.eval_native_archives_exn lib_info ~modules:(Some modules)
   in
   let cm_files =
@@ -357,19 +407,19 @@ let setup_build_archives (lib : Dune_file.Library.t) ~cctx ~expander ~scope =
       (* ctypes type_gen and function_gen scripts should not be included in the
          library. Otherwise they will spew stuff to stdout on library load. *)
       match lib.buildable.ctypes with
-      | Some ctypes -> Ctypes_rules.non_installable_modules ctypes
+      | Some ctypes -> Ctypes_stanza.non_installable_modules ctypes
       | None -> []
     in
     Cm_files.make ~excluded_modules ~obj_dir ~ext_obj ~modules
       ~top_sorted_modules ()
   in
   let* () =
-    Mode.Dict.Set.iter_concurrently modes ~f:(fun mode ->
-        build_lib lib ~native_archives ~dir ~sctx ~expander ~flags ~mode ~scope
+    Mode.Dict.Set.iter_concurrently modes.ocaml ~f:(fun mode ->
+        build_lib lib ~native_archives ~dir ~sctx ~expander ~flags ~mode
           ~cm_files)
   and* () =
     (* Build *.cma.js *)
-    Memo.Build.when_ modes.byte (fun () ->
+    Memo.when_ modes.ocaml.byte (fun () ->
         let action_with_targets =
           let src =
             Library.archive lib ~dir ~ext:(Mode.compiled_lib_ext Mode.Byte)
@@ -379,13 +429,13 @@ let setup_build_archives (lib : Dune_file.Library.t) ~cctx ~expander ~scope =
               (Path.Build.basename src)
             |> Path.Build.extend_basename ~suffix:".js"
           in
-          Jsoo_rules.build_cm cctx ~in_buildable:js_of_ocaml ~src ~target
+          Jsoo_rules.build_cm cctx ~in_context:js_of_ocaml ~src ~target
         in
         action_with_targets
         >>= Super_context.add_rule sctx ~dir ~loc:lib.buildable.loc)
   in
-  Memo.Build.when_
-    (Dynlink_supported.By_the_os.get natdynlink_supported && modes.native)
+  Memo.when_
+    (Dynlink_supported.By_the_os.get natdynlink_supported && modes.ocaml.native)
     (fun () -> build_shared ~native_archives ~sctx lib ~dir ~flags)
 
 let cctx (lib : Library.t) ~sctx ~source_modules ~dir ~expander ~scope
@@ -394,50 +444,29 @@ let cctx (lib : Library.t) ~sctx ~source_modules ~dir ~expander ~scope
   and* vimpl = Virtual_rules.impl sctx ~lib ~scope in
   let obj_dir = Library.obj_dir ~dir lib in
   let ctx = Super_context.context sctx in
-  let instrumentation_backend =
-    Lib.DB.instrumentation_backend (Scope.libs scope)
-  in
-  let* preprocess =
-    Resolve.Build.read_memo_build
-      (Preprocess.Per_module.with_instrumentation lib.buildable.preprocess
-         ~instrumentation_backend)
-  in
-  let* instrumentation_deps =
-    Resolve.Build.read_memo_build
-      (Preprocess.Per_module.instrumentation_deps lib.buildable.preprocess
-         ~instrumentation_backend)
-  in
-  (* Preprocess before adding the alias module as it doesn't need
-     preprocessing *)
-  let* pp =
-    Preprocessing.make sctx ~dir ~scope ~preprocess ~expander
-      ~preprocessor_deps:lib.buildable.preprocessor_deps ~instrumentation_deps
-      ~lint:lib.buildable.lint
-      ~lib_name:(Some (snd lib.name))
-  in
-  let* modules =
-    let add_empty_intf = lib.buildable.empty_module_interface_if_absent in
-    Modules.map_user_written source_modules ~f:(fun m ->
-        let* m = Pp_spec.pp_module pp m in
-        if add_empty_intf && not (Module.has m ~ml_kind:Intf) then
-          Module_compilation.with_empty_intf ~sctx ~dir m
-        else Memo.Build.return m)
+  let* modules, pp =
+    Buildable_rules.modules_rules sctx
+      (Library (lib.buildable, snd lib.name))
+      expander ~dir scope source_modules
   in
   let modules = Vimpl.impl_modules vimpl modules in
   let requires_compile = Lib.Compile.direct_requires compile_info in
   let requires_link = Lib.Compile.requires_link compile_info in
   let modes =
     let { Lib_config.has_native; _ } = ctx.lib_config in
-    Dune_file.Mode_conf.Set.eval_detailed lib.modes ~has_native
+    Dune_file.Mode_conf.Lib.Set.eval_detailed lib.modes ~has_native
   in
   let package = Dune_file.Library.package lib in
+  let js_of_ocaml =
+    Js_of_ocaml.In_context.make ~dir lib.buildable.js_of_ocaml
+  in
   Compilation_context.create () ~super_context:sctx ~expander ~scope ~obj_dir
     ~modules ~flags ~requires_compile ~requires_link ~preprocessing:pp
-    ~opaque:Inherit_from_settings ~js_of_ocaml:(Some lib.buildable.js_of_ocaml)
+    ~opaque:Inherit_from_settings ~js_of_ocaml:(Some js_of_ocaml)
     ?stdlib:lib.stdlib ~package ?vimpl ~modes
 
-let library_rules (lib : Library.t) ~cctx ~source_modules ~dir_contents
-    ~compile_info =
+let library_rules (lib : Library.t) ~local_lib ~cctx ~source_modules
+    ~dir_contents ~compile_info =
   let source_modules =
     Modules.fold_user_written source_modules ~init:[] ~f:(fun m acc -> m :: acc)
   in
@@ -450,26 +479,37 @@ let library_rules (lib : Library.t) ~cctx ~source_modules ~dir_contents
   let scope = Compilation_context.scope cctx in
   let* requires_compile = Compilation_context.requires_compile cctx in
   let stdlib_dir = (Compilation_context.context cctx).Context.stdlib_dir in
+  let top_sorted_modules =
+    let impl_only = Modules.impl_only modules in
+    Dep_graph.top_closed_implementations
+      (Compilation_context.dep_graphs cctx).impl impl_only
+  in
   let* () =
-    Memo.Build.Option.iter vimpl
+    Memo.Option.iter vimpl
       ~f:(Virtual_rules.setup_copy_rules_for_impl ~sctx ~dir)
   in
   let* () = Check_rules.add_obj_dir sctx ~obj_dir in
+  let* () = Check_rules.add_cycle_check sctx ~dir top_sorted_modules in
   let* () = gen_wrapped_compat_modules lib cctx
   and* () = Module_compilation.build_all cctx
-  and* expander = Super_context.expander sctx ~dir in
+  and* expander = Super_context.expander sctx ~dir
+  and* lib_info =
+    let lib_config = (Super_context.context sctx).lib_config in
+    Library.to_lib_info lib ~dir ~lib_config
+  in
   let+ () =
-    Memo.Build.when_
+    Memo.when_
       (not (Library.is_virtual lib))
-      (fun () -> setup_build_archives lib ~cctx ~expander ~scope)
+      (fun () ->
+        setup_build_archives lib ~lib_info ~top_sorted_modules ~cctx ~expander)
   and+ () =
     let vlib_stubs_o_files = Vimpl.vlib_stubs_o_files vimpl in
-    Memo.Build.when_
+    Memo.when_
       (Library.has_foreign lib || List.is_non_empty vlib_stubs_o_files)
       (fun () ->
         build_stubs lib ~cctx ~dir ~expander ~requires:requires_compile
           ~dir_contents ~vlib_stubs_o_files)
-  and+ () = Odoc.setup_library_odoc_rules cctx lib
+  and+ () = Odoc.setup_library_odoc_rules cctx local_lib
   and+ () =
     Sub_system.gen_rules
       { super_context = sctx
@@ -480,7 +520,7 @@ let library_rules (lib : Library.t) ~cctx ~source_modules ~dir_contents
       ; compile_info
       }
   and+ preprocess =
-    Resolve.Build.read_memo_build
+    Resolve.Memo.read_memo
       (Preprocess.Per_module.with_instrumentation lib.buildable.preprocess
          ~instrumentation_backend:
            (Lib.DB.instrumentation_backend (Scope.libs scope)))
@@ -490,13 +530,16 @@ let library_rules (lib : Library.t) ~cctx ~source_modules ~dir_contents
       ~preprocess ~libname:(snd lib.name) ~obj_dir
       ~dialects:(Dune_project.dialects (Scope.project scope))
       ~ident:(Lib.Compile.merlin_ident compile_info)
+      ~modes:(`Lib (Lib_info.modes lib_info))
       () )
 
 let rules (lib : Library.t) ~sctx ~dir_contents ~dir ~expander ~scope =
-  let* compile_info =
+  let buildable = lib.buildable in
+  let* local_lib, compile_info =
     Lib.DB.get_compile_info (Scope.libs scope) (Library.best_name lib)
-      ~allow_overlaps:lib.buildable.allow_overlapping_dependencies
+      ~allow_overlaps:buildable.allow_overlapping_dependencies
   in
+  let local_lib = Lib.Local.of_lib_exn local_lib in
   let f () =
     let* source_modules =
       Dir_contents.ocaml dir_contents
@@ -506,14 +549,14 @@ let rules (lib : Library.t) ~sctx ~dir_contents ~dir ~expander ~scope =
       cctx lib ~sctx ~source_modules ~dir ~scope ~expander ~compile_info
     in
     let* () =
-      let buildable = lib.Library.buildable in
-      match buildable.Buildable.ctypes with
-      | None -> Memo.Build.return ()
+      match buildable.ctypes with
+      | None -> Memo.return ()
       | Some _ctypes ->
-        Ctypes_rules.gen_rules ~loc:(fst lib.Library.name) ~cctx ~buildable
-          ~sctx ~scope ~dir
+        Ctypes_rules.gen_rules ~loc:(fst lib.name) ~cctx ~buildable ~sctx ~scope
+          ~dir
     in
-    library_rules lib ~cctx ~source_modules ~dir_contents ~compile_info
+    library_rules lib ~local_lib ~cctx ~source_modules ~dir_contents
+      ~compile_info
   in
   let* () = Buildable_rules.gen_select_rules sctx compile_info ~dir in
   Buildable_rules.with_lib_deps
