@@ -1,12 +1,9 @@
 (** {1 Handle link time code generation} *)
-open! Dune_engine
-
-open! Stdune
 
 type t =
-  { to_link : Lib.Lib_and_module.L.t
+  { to_link : Lib_flags.Lib_and_module.L.t
   ; force_linkall : bool
   }
 
 (** Generate link time code for special libraries such as [findlib.dynload] *)
-val handle_special_libs : Compilation_context.t -> t Resolve.t Memo.Build.t
+val handle_special_libs : Compilation_context.t -> t Resolve.t Memo.t
