@@ -1,4 +1,3 @@
-open! Dune_engine
 open Import
 module Mode = Wrapped
 
@@ -76,7 +75,7 @@ module Stdlib = struct
   let map t ~f = { t with modules = Module_name.Map.map t.modules ~f }
 
   let traverse t ~f =
-    let open Memo.Build.O in
+    let open Memo.O in
     let+ modules =
       Module_name.Map_traversals.parallel_map t.modules ~f:(fun _ -> f)
     in
@@ -148,6 +147,7 @@ module Mangle = struct
   type t =
     | Lib of Lib.t
     | Exe
+    | Melange
 
   let of_lib ~lib_name ~implements ~main_module_name ~modules =
     let kind : Lib.kind =
@@ -173,6 +173,8 @@ module Mangle = struct
         })
     | Exe ->
       sprintf "dune__exe" |> Module_name.of_string |> Visibility.Map.make_both
+    | Melange ->
+      sprintf "melange" |> Module_name.of_string |> Visibility.Map.make_both
 
   let make_alias_module t ~src_dir =
     let prefix = prefix t in
@@ -278,6 +280,22 @@ module Wrapped = struct
     ; wrapped = Simple true
     }
 
+  let melange ~src_dir ~modules =
+    let mangle = Mangle.Melange in
+    let prefix = Mangle.prefix mangle in
+    let alias_module = Mangle.make_alias_module mangle ~src_dir in
+    let modules =
+      Module_name.Map.map modules ~f:(fun m ->
+          Module.with_wrapper m ~main_module_name:prefix.public)
+    in
+    { modules
+    ; wrapped_compat = Module_name.Map.empty
+    ; alias_module
+      (* XXX melange's don't have a main module, but this is harmless *)
+    ; main_module_name = Module.name alias_module
+    ; wrapped = Simple true
+    }
+
   let obj_map
       { modules
       ; wrapped_compat
@@ -395,65 +413,24 @@ let as_singleton m =
   if Module_name.Map.cardinal m <> 1 then None
   else Module_name.Map.choose m |> Option.map ~f:snd
 
-(* Pre-1.11 encoding *)
-module Old_format = struct
-  let decode ~implements ~src_dir =
-    let open Dune_lang.Decoder in
-    fields
-      (let+ loc = loc
-       and+ alias_module = field_o "alias_module" (Module.decode ~src_dir)
-       and+ main_module_name = field_o "main_module_name" Module_name.decode
-       and+ modules =
-         field ~default:[] "modules" (repeat (enter (Module.decode ~src_dir)))
-       and+ wrapped = field "wrapped" Mode.decode in
-       let modules =
-         modules
-         |> List.map ~f:(fun m -> (Module.name m, m))
-         |> Module_name.Map.of_list_exn
-       in
-       match wrapped with
-       | Simple false -> (
-         match as_singleton modules with
-         | Some m -> Singleton m
-         | None -> Unwrapped modules)
-       | Yes_with_transition _ | Simple true -> (
-         match (main_module_name, alias_module, as_singleton modules) with
-         | Some main_module_name, _, Some m
-           when Module.name m = main_module_name && not implements ->
-           Singleton m
-         | Some main_module_name, Some alias_module, _ ->
-           Wrapped
-             { modules
-             ; wrapped_compat = Module_name.Map.empty
-             ; alias_module
-             ; main_module_name
-             ; wrapped
-             }
-         | None, _, _ | _, None, _ ->
-           User_error.raise ~loc
-             [ Pp.text "Cannot wrap without main module name or alias module" ]))
-end
-
 let singleton m = Singleton m
 
-let decode ~version ~src_dir ~implements =
-  if version <= (1, 10) then Old_format.decode ~implements ~src_dir
-  else
-    let open Dune_lang.Decoder in
-    sum
-      [ ( "singleton"
-        , let+ m = Module.decode ~src_dir in
-          Singleton m )
-      ; ( "unwrapped"
-        , let+ modules = Module.Name_map.decode ~src_dir in
-          Unwrapped modules )
-      ; ( "wrapped"
-        , let+ w = Wrapped.decode ~src_dir in
-          Wrapped w )
-      ; ( "stdlib"
-        , let+ stdlib = Stdlib.decode ~src_dir in
-          Stdlib stdlib )
-      ]
+let decode ~src_dir =
+  let open Dune_lang.Decoder in
+  sum
+    [ ( "singleton"
+      , let+ m = Module.decode ~src_dir in
+        Singleton m )
+    ; ( "unwrapped"
+      , let+ modules = Module.Name_map.decode ~src_dir in
+        Unwrapped modules )
+    ; ( "wrapped"
+      , let+ w = Wrapped.decode ~src_dir in
+        Wrapped w )
+    ; ( "stdlib"
+      , let+ stdlib = Stdlib.decode ~src_dir in
+        Stdlib stdlib )
+    ]
 
 let rec to_dyn =
   let open Dyn in
@@ -563,6 +540,17 @@ let exe_wrapped ~src_dir ~modules =
   | Some m -> singleton_exe m
   | None -> Wrapped (Wrapped.exe ~src_dir ~modules)
 
+let singleton_melange m =
+  Singleton
+    (let mangle = Mangle.Melange in
+     let main_module_name = (Mangle.prefix mangle).public in
+     Module.with_wrapper m ~main_module_name)
+
+let melange_wrapped ~src_dir ~modules =
+  match as_singleton modules with
+  | Some m -> singleton_melange m
+  | None -> Wrapped (Wrapped.melange ~src_dir ~modules)
+
 let rec impl_only = function
   | Stdlib w -> Stdlib.impl_only w
   | Singleton m -> if Module.has ~ml_kind:Impl m then [ m ] else []
@@ -656,7 +644,7 @@ let is_user_written m =
   match Module.kind m with
   | Root -> false
   | Wrapped_compat | Alias ->
-    (* Logically, this shold be [acc]. But this is unreachable these are stored
+    (* Logically, this should be [acc]. But this is unreachable these are stored
        separately *)
     assert false
   | _ -> true
@@ -671,8 +659,8 @@ let rec fold_user_written t ~f ~init =
   | Impl { impl; vlib = _ } -> fold_user_written impl ~f ~init
 
 let rec map_user_written t ~f =
-  let f m = if is_user_written m then f m else Memo.Build.return m in
-  let open Memo.Build.O in
+  let f m = if is_user_written m then f m else Memo.return m in
+  let open Memo.O in
   match t with
   | Singleton m ->
     let+ res = f m in
@@ -741,10 +729,7 @@ let rec obj_map : 'a. t -> f:(Sourced_module.t -> 'a) -> 'a Module.Obj_map.t =
           assert false)
 
 let obj_map_build :
-      'a.
-         t
-      -> f:(Sourced_module.t -> 'a Memo.Build.t)
-      -> 'a Module.Obj_map.t Memo.Build.t =
+      'a. t -> f:(Sourced_module.t -> 'a Memo.t) -> 'a Module.Obj_map.t Memo.t =
  fun t ~f ->
   Module.Obj_map_traversals.parallel_map (obj_map t ~f) ~f:(fun _ x -> x)
 
