@@ -1,7 +1,6 @@
 (** OCaml flags *)
-open! Dune_engine
 
-open! Stdune
+open! Import
 
 type t
 
@@ -38,6 +37,8 @@ val prepend_common : string list -> t -> t
 
 val with_vendored_warnings : t -> t
 
+val with_vendored_alerts : t -> t
+
 val common : t -> string list Action_builder.t
 
 val dump : t -> Dune_lang.t list Action_builder.t
