@@ -1,26 +1,46 @@
-open! Dune_engine
-open! Stdune
 open Import
 open Dune_file
-open Memo.Build.O
-module SC = Super_context
+open Memo.O
 
 let ( ++ ) = Path.Build.relative
 
+let find_project_by_key =
+  let memo =
+    let make_map projects =
+      Dune_project.File_key.Map.of_list_map_exn projects ~f:(fun project ->
+          (Dune_project.file_key project, project))
+      |> Memo.return
+    in
+    let module Input = struct
+      type t = Dune_project.t list
+
+      let equal = List.equal Dune_project.equal
+
+      let hash = List.hash Dune_project.hash
+
+      let to_dyn = Dyn.list Dune_project.to_dyn
+    end in
+    Memo.create "project-by-keys" ~input:(module Input) make_map
+  in
+  fun key ->
+    let* { projects; _ } = Dune_load.load () in
+    let+ map = Memo.exec memo projects in
+    Dune_project.File_key.Map.find_exn map key
+
 module Scope_key : sig
-  val of_string : Super_context.t -> string -> Lib_name.t * Lib.DB.t
+  val of_string : Context.t -> string -> (Lib_name.t * Lib.DB.t) Memo.t
 
   val to_string : Lib_name.t -> Dune_project.t -> string
 end = struct
-  let of_string sctx s =
+  let of_string context s =
     match String.rsplit2 s ~on:'@' with
     | None ->
-      (Lib_name.parse_string_exn (Loc.none, s), Super_context.public_libs sctx)
+      let+ public_libs = Scope.DB.public_libs context in
+      (Lib_name.parse_string_exn (Loc.none, s), public_libs)
     | Some (lib, key) ->
-      let scope =
-        Dune_project.File_key.of_string key
-        |> Super_context.find_project_by_key sctx
-        |> Super_context.find_scope_by_project sctx
+      let+ scope =
+        let key = Dune_project.File_key.of_string key in
+        find_project_by_key key >>= Scope.DB.find_by_project context
       in
       (Lib_name.parse_string_exn (Loc.none, lib), Scope.libs scope)
 
@@ -114,7 +134,7 @@ module Dep : sig
 
   (*** [setup_deps ctx target odocs] Adds [odocs] as dependencies for [target].
     These dependencies may be used using the [deps] function *)
-  val setup_deps : Context.t -> target -> Path.Set.t -> unit Memo.Build.t
+  val setup_deps : Context.t -> target -> Path.Set.t -> unit Memo.t
 end = struct
   let html_alias ctx m = Alias.doc ~dir:(Paths.html ctx m)
 
@@ -166,17 +186,30 @@ end = struct
   let odoc_input t = t
 end
 
-let odoc sctx =
-  let dir = (Super_context.context sctx).build_dir in
-  SC.resolve_program sctx ~dir "odoc" ~loc:None ~hint:"opam install odoc"
-
 let odoc_base_flags sctx build_dir =
-  let open Memo.Build.O in
+  let open Memo.O in
   let+ conf = Super_context.odoc sctx ~dir:build_dir in
   match conf.Env_node.Odoc.warnings with
   | Fatal -> Command.Args.A "--warn-error"
   | Nonfatal -> S []
 
+let run_odoc sctx ~dir command ~flags_for args =
+  let build_dir = (Super_context.context sctx).build_dir in
+  let open Memo.O in
+  let* program =
+    Super_context.resolve_program sctx ~dir:build_dir "odoc" ~loc:None
+      ~hint:"opam install odoc"
+  in
+  let+ base_flags =
+    match flags_for with
+    | None -> Memo.return Command.Args.empty
+    | Some path -> odoc_base_flags sctx path
+  in
+  let deps = Action_builder.env_var "ODOC_SYNTAX" in
+  let open Action_builder.With_targets.O in
+  Action_builder.with_no_targets deps
+  >>> Command.run ~dir program [ A command; base_flags; S args ]
+
 let module_deps (m : Module.t) ~obj_dir ~(dep_graphs : Dep_graph.Ml_kind.t) =
   Action_builder.dyn_paths_unit
     (let open Action_builder.O in
@@ -191,49 +224,47 @@ let module_deps (m : Module.t) ~obj_dir ~(dep_graphs : Dep_graph.Ml_kind.t) =
 let compile_module sctx ~obj_dir (m : Module.t) ~includes:(file_deps, iflags)
     ~dep_graphs ~pkg_or_lnu =
   let odoc_file = Obj_dir.Module.odoc obj_dir m in
-  let open Memo.Build.O in
+  let open Memo.O in
   let+ () =
     let* action_with_targets =
       let doc_dir = Path.build (Obj_dir.odoc_dir obj_dir) in
-      let* odoc = odoc sctx in
-      let+ odoc_base_flags = odoc_base_flags sctx odoc_file in
+      let+ run_odoc =
+        run_odoc sctx ~dir:doc_dir "compile" ~flags_for:(Some odoc_file)
+          [ A "-I"
+          ; Path doc_dir
+          ; iflags
+          ; As [ "--pkg"; pkg_or_lnu ]
+          ; A "-o"
+          ; Target odoc_file
+          ; Dep
+              (Path.build
+                 (Obj_dir.Module.cmti_file ~cm_kind:(Ocaml Cmi) obj_dir m))
+          ]
+      in
       let open Action_builder.With_targets.O in
       Action_builder.with_no_targets file_deps
       >>> Action_builder.with_no_targets (module_deps m ~obj_dir ~dep_graphs)
-      >>> Command.run ~dir:doc_dir odoc
-            [ A "compile"
-            ; odoc_base_flags
-            ; A "-I"
-            ; Path doc_dir
-            ; iflags
-            ; As [ "--pkg"; pkg_or_lnu ]
-            ; A "-o"
-            ; Target odoc_file
-            ; Dep (Path.build (Obj_dir.Module.cmti_file obj_dir m))
-            ]
+      >>> run_odoc
     in
     add_rule sctx action_with_targets
   in
   (m, odoc_file)
 
 let compile_mld sctx (m : Mld.t) ~includes ~doc_dir ~pkg =
-  let open Memo.Build.O in
+  let open Memo.O in
   let odoc_file = Mld.odoc_file m ~doc_dir in
   let odoc_input = Mld.odoc_input m in
-  let* odoc = odoc sctx in
-  let* odoc_base_flags = odoc_base_flags sctx odoc_input in
-  let+ () =
-    add_rule sctx
-      (Command.run ~dir:(Path.build doc_dir) odoc
-         [ A "compile"
-         ; odoc_base_flags
-         ; Command.Args.dyn includes
-         ; As [ "--pkg"; Package.Name.to_string pkg ]
-         ; A "-o"
-         ; Target odoc_file
-         ; Dep (Path.build odoc_input)
-         ])
+  let* run_odoc =
+    run_odoc sctx ~dir:(Path.build doc_dir) "compile"
+      ~flags_for:(Some odoc_input)
+      [ Command.Args.dyn includes
+      ; As [ "--pkg"; Package.Name.to_string pkg ]
+      ; A "-o"
+      ; Target odoc_file
+      ; Dep (Path.build odoc_input)
+      ]
   in
+  let+ () = add_rule sctx run_odoc in
   odoc_file
 
 let odoc_include_flags ctx pkg requires =
@@ -259,39 +290,30 @@ let odoc_include_flags ctx pkg requires =
 let link_odoc_rules sctx (odoc_file : odoc_artefact) ~pkg ~requires =
   let ctx = Super_context.context sctx in
   let deps = Dep.deps ctx pkg requires in
-  let open Memo.Build.O in
-  let* odoc = odoc sctx
-  and* odoc_base_flags = odoc_base_flags sctx odoc_file.odoc_file in
+  let open Memo.O in
+  let* run_odoc =
+    run_odoc sctx
+      ~dir:(Path.build (Paths.html_root ctx))
+      "link" ~flags_for:(Some odoc_file.odoc_file)
+      [ odoc_include_flags ctx pkg requires
+      ; A "-o"
+      ; Target odoc_file.odocl_file
+      ; Dep (Path.build odoc_file.odoc_file)
+      ]
+  in
   add_rule sctx
     (let open Action_builder.With_targets.O in
-    Action_builder.with_no_targets deps
-    >>> Command.run
-          ~dir:(Path.build (Paths.html_root ctx))
-          odoc
-          [ A "link"
-          ; odoc_base_flags
-          ; odoc_include_flags ctx pkg requires
-          ; A "-o"
-          ; Target odoc_file.odocl_file
-          ; Dep (Path.build odoc_file.odoc_file)
-          ])
+    Action_builder.with_no_targets deps >>> run_odoc)
 
-let setup_library_odoc_rules cctx (library : Library.t) =
-  let open Memo.Build.O in
-  let* lib =
-    let scope = Compilation_context.scope cctx in
-    Library.best_name library
-    |> Lib.DB.find_even_when_hidden (Scope.libs scope)
-    >>| Option.value_exn
-  in
-  let local_lib = Lib.Local.of_lib_exn lib in
+let setup_library_odoc_rules cctx (local_lib : Lib.Local.t) =
+  let open Memo.O in
   (* Using the proper package name doesn't actually work since odoc assumes that
      a package contains only 1 library *)
-  let pkg_or_lnu = pkg_or_lnu lib in
+  let pkg_or_lnu = pkg_or_lnu (Lib.Local.to_lib local_lib) in
   let sctx = Compilation_context.super_context cctx in
   let ctx = Super_context.context sctx in
   let* requires = Compilation_context.requires_compile cctx in
-  let info = Lib.info lib in
+  let info = Lib.Local.info local_lib in
   let package = Lib_info.package info in
   let odoc_include_flags =
     Command.Args.memo (odoc_include_flags ctx package requires)
@@ -308,9 +330,7 @@ let setup_library_odoc_rules cctx (library : Library.t) =
         in
         compiled :: acc)
   in
-  let* modules_and_odoc_files =
-    Memo.Build.all_concurrently modules_and_odoc_files
-  in
+  let* modules_and_odoc_files = Memo.all_concurrently modules_and_odoc_files in
   Dep.setup_deps ctx (Lib local_lib)
     (Path.Set.of_list_map modules_and_odoc_files ~f:(fun (_, p) -> Path.build p))
 
@@ -320,14 +340,23 @@ let setup_html sctx (odoc_file : odoc_artefact) =
     match odoc_file.source with
     | Mld -> (odoc_file.html_file, [])
     | Module ->
-      (* Dummy target so that the bellow rule as at least one target. We do this
+      (* Dummy target so that the below rule as at least one target. We do this
          because we don't know the targets of odoc in this case. The proper way
          to support this would be to have directory targets. *)
       let dummy = Action_builder.create_file (odoc_file.html_dir ++ ".dummy") in
       (odoc_file.html_dir, [ dummy ])
   in
-  let open Memo.Build.O in
-  let* odoc = odoc sctx in
+  let open Memo.O in
+  let* run_odoc =
+    run_odoc sctx
+      ~dir:(Path.build (Paths.html_root ctx))
+      "html-generate" ~flags_for:None
+      [ A "-o"
+      ; Path (Path.build (Paths.html_root ctx))
+      ; Dep (Path.build odoc_file.odocl_file)
+      ; Hidden_targets [ odoc_file.html_file ]
+      ]
+  in
   add_rule sctx
     (Action_builder.progn
        (Action_builder.with_no_targets
@@ -335,37 +364,29 @@ let setup_html sctx (odoc_file : odoc_artefact) =
              (Action.Full.make
                 (Action.Progn
                    [ Action.Remove_tree to_remove
-                   ; Action.Mkdir (Path.build odoc_file.html_dir)
+                   ; Action.Mkdir odoc_file.html_dir
                    ])))
-       :: Command.run
-            ~dir:(Path.build (Paths.html_root ctx))
-            odoc
-            [ A "html-generate"
-            ; A "-o"
-            ; Path (Path.build (Paths.html_root ctx))
-            ; Dep (Path.build odoc_file.odocl_file)
-            ; Hidden_targets [ odoc_file.html_file ]
-            ]
-       :: dummy))
+       :: run_odoc :: dummy))
 
 let setup_css_rule sctx =
-  let open Memo.Build.O in
+  let open Memo.O in
   let ctx = Super_context.context sctx in
-  let* odoc = odoc sctx in
-  add_rule sctx
-    (Command.run ~dir:(Path.build ctx.build_dir) odoc
-       [ A "support-files"
-       ; A "-o"
-       ; Path (Path.build (Paths.html_root ctx))
-       ; Hidden_targets [ Paths.css_file ctx; Paths.highlight_pack_js ctx ]
-       ])
+  let* run_odoc =
+    run_odoc sctx ~dir:(Path.build ctx.build_dir) "support-files"
+      ~flags_for:None
+      [ A "-o"
+      ; Path (Path.build (Paths.html_root ctx))
+      ; Hidden_targets [ Paths.css_file ctx; Paths.highlight_pack_js ctx ]
+      ]
+  in
+  add_rule sctx run_odoc
 
 let sp = Printf.sprintf
 
 let setup_toplevel_index_rule sctx =
-  let list_items =
-    Super_context.packages sctx
-    |> Package.Name.Map.to_list
+  let* list_items =
+    let+ packages = Only_packages.get () in
+    Package.Name.Map.to_list packages
     |> List.filter_map ~f:(fun (name, pkg) ->
            let name = Package.Name.to_string name in
            let link = sp {|<a href="%s/index.html">%s</a>|} name name in
@@ -403,21 +424,25 @@ let setup_toplevel_index_rule sctx =
   let ctx = Super_context.context sctx in
   add_rule sctx (Action_builder.write_file (Paths.toplevel_index ctx) html)
 
-let libs_of_pkg sctx ~pkg =
-  SC.lib_entries_of_package sctx pkg
-  |> (* Filter out all implementations of virtual libraries *)
-  List.filter_map ~f:(function
-    | Super_context.Lib_entry.Library lib ->
-      let is_impl = Lib.Local.to_lib lib |> Lib.is_impl in
-      Option.some_if (not is_impl) lib
-    | Deprecated_library_name _ -> None)
+let libs_of_pkg ctx ~pkg =
+  let+ entries = Scope.DB.lib_entries_of_package ctx pkg in
+  (* Filter out all implementations of virtual libraries *)
+  List.filter_map entries ~f:(fun (entry : Scope.DB.Lib_entry.t) ->
+      match entry with
+      | Library lib ->
+        let is_impl =
+          Lib.Local.to_lib lib |> Lib.info |> Lib_info.implements
+          |> Option.is_some
+        in
+        Option.some_if (not is_impl) lib
+      | Deprecated_library_name _ -> None)
 
 let load_all_odoc_rules_pkg sctx ~pkg =
-  let pkg_libs = libs_of_pkg sctx ~pkg in
+  let* pkg_libs = libs_of_pkg sctx ~pkg in
   let+ () =
-    Memo.Build.parallel_iter
+    Memo.parallel_iter
       (Pkg pkg :: List.map pkg_libs ~f:(fun lib -> Lib lib))
-      ~f:(fun _ -> Memo.Build.return ())
+      ~f:(fun _ -> Memo.return ())
   in
   pkg_libs
 
@@ -430,14 +455,14 @@ let entry_modules_by_lib sctx lib =
   >>| Modules.entry_modules
 
 let entry_modules sctx ~pkg =
-  let l =
-    libs_of_pkg sctx ~pkg
-    |> List.filter ~f:(fun lib ->
-           Lib.Local.info lib |> Lib_info.status |> Lib_info.Status.is_private
-           |> not)
+  let* l =
+    libs_of_pkg (Super_context.context sctx) ~pkg
+    >>| List.filter ~f:(fun lib ->
+            Lib.Local.info lib |> Lib_info.status |> Lib_info.Status.is_private
+            |> not)
   in
   let+ l =
-    Memo.Build.parallel_map l ~f:(fun l ->
+    Memo.parallel_map l ~f:(fun l ->
         let+ m = entry_modules_by_lib sctx l in
         (l, m))
   in
@@ -513,7 +538,7 @@ let odoc_artefacts sctx target =
         let odoc_file = Obj_dir.Module.odoc obj_dir m in
         create_odoc ctx ~target odoc_file)
       modules
-    |> Memo.Build.return
+    |> Memo.return
 
 let setup_lib_odocl_rules_def =
   let module Input = struct
@@ -527,7 +552,7 @@ let setup_lib_odocl_rules_def =
       && Resolve.equal (List.equal Lib.equal) r1 r2
 
     let hash (sc, l, r) =
-      Hashtbl.hash
+      Poly.hash
         ( Super_context.hash sc
         , Lib.Local.hash l
         , Resolve.hash (List.hash Lib.hash) r )
@@ -537,7 +562,7 @@ let setup_lib_odocl_rules_def =
   let f (sctx, lib, requires) =
     let* odocs = odoc_artefacts sctx (Lib lib) in
     let pkg = Lib_info.package (Lib.Local.info lib) in
-    Memo.Build.parallel_iter odocs ~f:(fun odoc ->
+    Memo.parallel_iter odocs ~f:(fun odoc ->
         link_odoc_rules sctx ~pkg ~requires odoc)
   in
   Memo.With_implicit_output.create "setup_library_odocls_rules"
@@ -560,7 +585,7 @@ let setup_pkg_rules_def memo_name f =
       && Super_context.equal s1 s2
 
     let hash (sctx, p, ls) =
-      Hashtbl.hash
+      Poly.hash
         ( Super_context.hash sctx
         , Package.Name.hash p
         , List.hash Lib.Local.hash ls )
@@ -582,24 +607,23 @@ let setup_pkg_odocl_rules_def =
       let libs = (libs :> Lib.t list) in
       Lib.closure libs ~linking:false
     in
-    let* () =
-      Memo.Build.parallel_iter libs ~f:(setup_lib_odocl_rules sctx ~requires)
+    let* () = Memo.parallel_iter libs ~f:(setup_lib_odocl_rules sctx ~requires)
     and* _ =
       let* pkg_odocs = odoc_artefacts sctx (Pkg pkg) in
       let pkg = Some pkg in
       let+ () =
-        Memo.Build.parallel_iter pkg_odocs ~f:(fun odoc ->
+        Memo.parallel_iter pkg_odocs ~f:(fun odoc ->
             link_odoc_rules sctx ~pkg ~requires odoc)
       in
       pkg_odocs
     and* _ =
-      Memo.Build.parallel_map libs ~f:(fun lib -> odoc_artefacts sctx (Lib lib))
+      Memo.parallel_map libs ~f:(fun lib -> odoc_artefacts sctx (Lib lib))
     in
-    Memo.Build.return ()
+    Memo.return ()
   in
   setup_pkg_rules_def "setup-package-odocls-rules" f
 
-let setup_pkg_odocl_rules sctx ~pkg ~libs : unit Memo.Build.t =
+let setup_pkg_odocl_rules sctx ~pkg ~libs : unit Memo.t =
   Memo.With_implicit_output.exec setup_pkg_odocl_rules_def (sctx, pkg, libs)
 
 let setup_lib_html_rules_def =
@@ -611,16 +635,14 @@ let setup_lib_html_rules_def =
     let equal (sc1, l1) (sc2, l2) =
       Super_context.equal sc1 sc2 && Lib.Local.equal l1 l2
 
-    let hash (sc, l) = Hashtbl.hash (Super_context.hash sc, Lib.Local.hash l)
+    let hash (sc, l) = Poly.hash (Super_context.hash sc, Lib.Local.hash l)
 
     let to_dyn _ = Dyn.Opaque
   end in
   let f (sctx, lib) =
     let ctx = Super_context.context sctx in
     let* odocs = odoc_artefacts sctx (Lib lib) in
-    let* () =
-      Memo.Build.parallel_iter odocs ~f:(fun odoc -> setup_html sctx odoc)
-    in
+    let* () = Memo.parallel_iter odocs ~f:(fun odoc -> setup_html sctx odoc) in
     let html_files = List.map ~f:(fun o -> Path.build o.html_file) odocs in
     let static_html = List.map ~f:Path.build (static_html ctx) in
     Rules.Produce.Alias.add_deps
@@ -638,15 +660,13 @@ let setup_lib_html_rules sctx lib =
 let setup_pkg_html_rules_def =
   let f (sctx, pkg, (libs : Lib.Local.t list)) =
     let ctx = Super_context.context sctx in
-    let* () = Memo.Build.parallel_iter libs ~f:(setup_lib_html_rules sctx)
+    let* () = Memo.parallel_iter libs ~f:(setup_lib_html_rules sctx)
     and* pkg_odocs =
       let* pkg_odocs = odoc_artefacts sctx (Pkg pkg) in
-      let+ () =
-        Memo.Build.parallel_iter pkg_odocs ~f:(fun o -> setup_html sctx o)
-      in
+      let+ () = Memo.parallel_iter pkg_odocs ~f:(fun o -> setup_html sctx o) in
       pkg_odocs
     and* lib_odocs =
-      Memo.Build.parallel_map libs ~f:(fun lib -> odoc_artefacts sctx (Lib lib))
+      Memo.parallel_map libs ~f:(fun lib -> odoc_artefacts sctx (Lib lib))
     in
     let odocs = List.concat (pkg_odocs :: lib_odocs) in
     let html_files = List.map ~f:(fun o -> Path.build o.html_file) odocs in
@@ -657,7 +677,7 @@ let setup_pkg_html_rules_def =
   in
   setup_pkg_rules_def "setup-package-html-rules" f
 
-let setup_pkg_html_rules sctx ~pkg ~libs : unit Memo.Build.t =
+let setup_pkg_html_rules sctx ~pkg ~libs : unit Memo.t =
   Memo.With_implicit_output.exec setup_pkg_html_rules_def (sctx, pkg, libs)
 
 let setup_package_aliases sctx (pkg : Package.t) =
@@ -668,9 +688,11 @@ let setup_package_aliases sctx (pkg : Package.t) =
     let dir = Path.Build.append_source ctx.build_dir pkg_dir in
     Alias.doc ~dir
   in
-  Dep.html_alias ctx (Pkg name)
-  :: (libs_of_pkg sctx ~pkg:name
-     |> List.map ~f:(fun lib -> Dep.html_alias ctx (Lib lib)))
+  let* libs =
+    libs_of_pkg ctx ~pkg:name
+    >>| List.map ~f:(fun lib -> Dep.html_alias ctx (Lib lib))
+  in
+  Dep.html_alias ctx (Pkg name) :: libs
   |> Dune_engine.Dep.Set.of_list_map ~f:(fun f -> Dune_engine.Dep.alias f)
   |> Action_builder.deps
   |> Rules.Produce.Alias.add_deps alias
@@ -717,7 +739,7 @@ let package_mlds =
             let* mlds = Packages.mlds sctx pkg in
             let mlds = check_mlds_no_dupes ~pkg ~mlds in
             let ctx = Super_context.context sctx in
-            if String.Map.mem mlds "index" then Memo.Build.return mlds
+            if String.Map.mem mlds "index" then Memo.return mlds
             else
               let gen_mld = Paths.gen_mld_dir ctx pkg ++ "index.mld" in
               let* entry_modules = entry_modules sctx ~pkg in
@@ -737,7 +759,7 @@ let setup_package_odoc_rules sctx ~pkg =
      back to a package name here. Need to try and change that one day. *)
   let pkg = Package.name pkg in
   let* odocs =
-    Memo.Build.parallel_map (String.Map.values mlds) ~f:(fun mld ->
+    Memo.parallel_map (String.Map.values mlds) ~f:(fun mld ->
         compile_mld sctx (Mld.create mld) ~pkg
           ~doc_dir:(Paths.odocs ctx (Pkg pkg))
           ~includes:(Action_builder.return []))
@@ -753,7 +775,7 @@ let gen_project_rules sctx project =
 
 let setup_private_library_doc_alias sctx ~scope ~dir (l : Dune_file.Library.t) =
   match l.visibility with
-  | Public _ -> Memo.Build.return ()
+  | Public _ -> Memo.return ()
   | Private _ ->
     let ctx = Super_context.context sctx in
     let* lib =
@@ -765,35 +787,52 @@ let setup_private_library_doc_alias sctx ~scope ~dir (l : Dune_file.Library.t) =
       (lib |> Dep.html_alias ctx |> Dune_engine.Dep.alias |> Action_builder.dep)
 
 let has_rules m =
-  let+ rules = Rules.collect_unit (fun () -> m) in
-  Build_config.Rules (Subdir_set.empty, rules)
+  let rules = Rules.collect_unit (fun () -> m) in
+  Memo.return
+    (Build_config.Rules
+       { rules
+       ; build_dir_only_sub_dirs = Subdir_set.empty
+       ; directory_targets = Path.Build.Map.empty
+       })
 
-let with_package sctx pkg ~f =
+let with_package pkg ~f =
   let pkg = Package.Name.of_string pkg in
-  let packages = Super_context.packages sctx in
+  let* packages = Only_packages.get () in
   match Package.Name.Map.find packages pkg with
   | None ->
-    Memo.Build.return (Build_config.Rules (Subdir_set.empty, Rules.empty))
+    Memo.return
+      (Build_config.Rules
+         { rules = Memo.return Rules.empty
+         ; build_dir_only_sub_dirs = Subdir_set.empty
+         ; directory_targets = Path.Build.Map.empty
+         })
   | Some pkg -> has_rules (f pkg)
 
 let gen_rules sctx ~dir:_ rest =
   match rest with
-  | [] -> Memo.Build.return (Build_config.Rules (Subdir_set.All, Rules.empty))
+  | [] ->
+    Memo.return
+      (Build_config.Rules
+         { rules = Memo.return Rules.empty
+         ; build_dir_only_sub_dirs = Subdir_set.All
+         ; directory_targets = Path.Build.Map.empty
+         })
   | [ "_html" ] ->
     has_rules (setup_css_rule sctx >>> setup_toplevel_index_rule sctx)
   | [ "_mlds"; pkg ] ->
-    with_package sctx pkg ~f:(fun pkg ->
+    with_package pkg ~f:(fun pkg ->
         let* _mlds, rules = package_mlds sctx ~pkg in
         Rules.produce rules)
   | [ "_odoc"; "pkg"; pkg ] ->
-    with_package sctx pkg ~f:(fun pkg -> setup_package_odoc_rules sctx ~pkg)
+    with_package pkg ~f:(fun pkg -> setup_package_odoc_rules sctx ~pkg)
   | [ "_odocls"; lib_unique_name_or_pkg ] ->
     has_rules
       ((* TODO we can be a better with the error handling in the case where
           lib_unique_name_or_pkg is neither a valid pkg or lnu *)
-       let lib, lib_db = Scope_key.of_string sctx lib_unique_name_or_pkg in
+       let ctx = Super_context.context sctx in
+       let* lib, lib_db = Scope_key.of_string ctx lib_unique_name_or_pkg in
        let setup_pkg_odocl_rules pkg =
-         let* pkg_libs = load_all_odoc_rules_pkg sctx ~pkg in
+         let* pkg_libs = load_all_odoc_rules_pkg ctx ~pkg in
          setup_pkg_odocl_rules sctx ~pkg ~libs:pkg_libs
        in
        (* jeremiedimino: why isn't [None] some kind of error here? *)
@@ -803,7 +842,7 @@ let gen_rules sctx ~dir:_ rest =
        in
        let+ () =
          match lib with
-         | None -> Memo.Build.return ()
+         | None -> Memo.return ()
          | Some lib -> (
            match Lib_info.package (Lib.Local.info lib) with
            | None ->
@@ -813,11 +852,12 @@ let gen_rules sctx ~dir:_ rest =
              setup_lib_odocl_rules sctx lib ~requires
            | Some pkg -> setup_pkg_odocl_rules pkg)
        and+ () =
+         let* packages = Only_packages.get () in
          match
-           Package.Name.Map.find (SC.packages sctx)
+           Package.Name.Map.find packages
              (Package.Name.of_string lib_unique_name_or_pkg)
          with
-         | None -> Memo.Build.return ()
+         | None -> Memo.return ()
          | Some pkg ->
            let name = Package.name pkg in
            setup_pkg_odocl_rules name
@@ -827,9 +867,12 @@ let gen_rules sctx ~dir:_ rest =
     has_rules
       ((* TODO we can be a better with the error handling in the case where
           lib_unique_name_or_pkg is neither a valid pkg or lnu *)
-       let lib, lib_db = Scope_key.of_string sctx lib_unique_name_or_pkg in
+       let ctx = Super_context.context sctx in
+       let* lib, lib_db = Scope_key.of_string ctx lib_unique_name_or_pkg in
        let setup_pkg_html_rules pkg =
-         let* pkg_libs = load_all_odoc_rules_pkg sctx ~pkg in
+         let* pkg_libs =
+           load_all_odoc_rules_pkg (Super_context.context sctx) ~pkg
+         in
          setup_pkg_html_rules sctx ~pkg ~libs:pkg_libs
        in
        (* jeremiedimino: why isn't [None] some kind of error here? *)
@@ -839,20 +882,21 @@ let gen_rules sctx ~dir:_ rest =
        in
        let+ () =
          match lib with
-         | None -> Memo.Build.return ()
+         | None -> Memo.return ()
          | Some lib -> (
            match Lib_info.package (Lib.Local.info lib) with
            | None -> setup_lib_html_rules sctx lib
            | Some pkg -> setup_pkg_html_rules pkg)
        and+ () =
+         let* packages = Only_packages.get () in
          match
-           Package.Name.Map.find (SC.packages sctx)
+           Package.Name.Map.find packages
              (Package.Name.of_string lib_unique_name_or_pkg)
          with
-         | None -> Memo.Build.return ()
+         | None -> Memo.return ()
          | Some pkg ->
            let name = Package.name pkg in
            setup_pkg_html_rules name
        in
        ())
-  | _ -> Memo.Build.return Build_config.Redirect_to_parent
+  | _ -> Memo.return Build_config.Redirect_to_parent
