@@ -1,7 +1,5 @@
-open! Dune_engine
-open! Stdune
 open Import
-module SC = Super_context
+open Memo.O
 
 (* Encoded representation of a set of library names + scope *)
 module Key : sig
@@ -100,7 +98,7 @@ end = struct
 end
 
 let pped_module m ~f =
-  let open Memo.Build.O in
+  let open Memo.O in
   let pped = Module.pped m in
   let+ () =
     Module.iter m ~f:(fun ml_kind file ->
@@ -108,7 +106,7 @@ let pped_module m ~f =
           Module.file pped ~ml_kind |> Option.value_exn
           |> Path.as_in_build_dir_exn
         in
-        let file = Path.as_in_build_dir_exn file.path in
+        let file = Path.as_in_build_dir_exn (Module.File.path file) in
         f ml_kind file pp_path)
   in
   pped
@@ -116,7 +114,7 @@ let pped_module m ~f =
 module Driver = struct
   module M = struct
     module Info = struct
-      let name = Sub_system_name.make "ppx.driver"
+      let name = Sub_system_name.of_string "ppx.driver"
 
       type t =
         { loc : Loc.t
@@ -182,10 +180,10 @@ module Driver = struct
     let replaces t = t.replaces
 
     let instantiate ~resolve ~get lib (info : Info.t) =
-      let open Memo.Build.O in
+      let open Memo.O in
       let+ replaces =
-        Memo.Build.parallel_map info.replaces ~f:(fun ((loc, name) as x) ->
-            Resolve.Build.bind (resolve x) ~f:(fun lib ->
+        Memo.parallel_map info.replaces ~f:(fun ((loc, name) as x) ->
+            Resolve.Memo.bind (resolve x) ~f:(fun lib ->
                 get ~loc lib >>| function
                 | None ->
                   Resolve.fail
@@ -199,7 +197,7 @@ module Driver = struct
       { info; lib; replaces }
 
     let public_info t =
-      Memo.Build.return
+      Memo.return
         (let open Resolve.O in
         let+ replaces = t.replaces in
         { Info.loc = t.info.loc
@@ -234,7 +232,7 @@ module Driver = struct
            ])
 
   let select libs ~loc =
-    let open Memo.Build.O in
+    let open Memo.O in
     select_replaceable_backend libs ~replaces
     >>| Resolve.bind ~f:(function
           | Ok x -> Resolve.return x
@@ -277,16 +275,16 @@ module Driver = struct
                          Lib_name.to_string (Lib.name (lib t)))))))
 end
 
-let ppx_exe sctx ~key =
-  let build_dir = (Super_context.context sctx).build_dir in
+let ppx_exe (ctx : Context.t) ~key =
+  let build_dir = ctx.build_dir in
   Path.Build.relative build_dir (".ppx/" ^ key ^ "/ppx.exe")
 
 let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
-  let open Memo.Build.O in
-  let ctx = SC.context sctx in
+  let open Memo.O in
+  let ctx = Super_context.context sctx in
   let* driver_and_libs =
-    let ( let& ) t f = Resolve.Build.bind t ~f in
-    let& pps = Resolve.Build.lift pps in
+    let ( let& ) t f = Resolve.Memo.bind t ~f in
+    let& pps = Resolve.Memo.lift pps in
     let& pps = Lib.closure ~linking:true pps in
     Driver.select pps ~loc:(Dot_ppx (target, pp_names))
     >>| Resolve.map ~f:(fun driver -> (driver, pps))
@@ -307,7 +305,7 @@ let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
     |> Option.value_exn |> Path.as_in_build_dir_exn
   in
   let* () =
-    SC.add_rule sctx ~dir
+    Super_context.add_rule sctx ~dir
       (Action_builder.write_file_dyn ml_source
          (Resolve.read
             (let open Resolve.O in
@@ -325,22 +323,24 @@ let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
   let* cctx =
     let* expander = Super_context.expander sctx ~dir in
     let requires_compile = Resolve.map driver_and_libs ~f:snd in
-    let requires_link =
-      Memo.lazy_ (fun () -> Memo.Build.return requires_compile)
-    in
+    let requires_link = Memo.lazy_ (fun () -> Memo.return requires_compile) in
     let flags = Ocaml_flags.of_list [ "-g"; "-w"; "-24" ] in
     let opaque = Compilation_context.Explicit false in
     let modules = Modules.singleton_exe module_ in
     Compilation_context.create ~super_context:sctx ~scope ~expander ~obj_dir
       ~modules ~flags
-      ~requires_compile:(Memo.Build.return requires_compile)
+      ~requires_compile:(Memo.return requires_compile)
       ~requires_link ~opaque ~js_of_ocaml:None ~package:None ~bin_annot:false ()
   in
-  Exe.build_and_link ~program ~linkages cctx ~promote:None
+  let+ (_ : Exe.dep_graphs) =
+    Exe.build_and_link ~program ~linkages cctx ~promote:None
+  in
+  ()
 
 let get_rules sctx key =
-  let exe = ppx_exe sctx ~key in
-  let pp_names, scope =
+  let ctx = Super_context.context sctx in
+  let exe = ppx_exe ctx ~key in
+  let* pp_names, scope =
     match Digest.from_hex key with
     | None ->
       User_error.raise
@@ -349,18 +349,17 @@ let get_rules sctx key =
         ]
     | Some key ->
       let { Key.Decoded.pps; project_root } = Key.decode key in
-      let scope =
+      let+ scope =
         let dir =
           match project_root with
-          | None -> (Super_context.context sctx).build_dir
-          | Some dir ->
-            Path.Build.append_source (Super_context.context sctx).build_dir dir
+          | None -> ctx.build_dir
+          | Some dir -> Path.Build.append_source ctx.build_dir dir
         in
-        Super_context.find_scope_by_dir sctx dir
+        Scope.DB.find_by_dir dir
       in
       (pps, scope)
   in
-  let open Memo.Build.O in
+  let open Memo.O in
   let* pps =
     let lib_db = Scope.libs scope in
     List.map pp_names ~f:(fun x -> (Loc.none, x)) |> Lib.DB.resolve_pps lib_db
@@ -370,17 +369,16 @@ let get_rules sctx key =
 let gen_rules sctx components =
   match components with
   | [ key ] -> get_rules sctx key
-  | _ -> Memo.Build.return ()
+  | _ -> Memo.return ()
 
-let ppx_driver_exe sctx libs =
+let ppx_driver_exe (ctx : Context.t) libs =
   let key = Digest.to_string (Key.Decoded.of_libs libs |> Key.encode) in
   (* Make sure to compile ppx.exe for the compiling host. See: #2252, #2286 and
      #3698 *)
-  let sctx = SC.host sctx in
-  ppx_exe sctx ~key
+  ppx_exe ~key (Context.host ctx)
 
 let get_cookies ~loc ~expander ~lib_name libs =
-  let open Memo.Build.O in
+  let open Memo.O in
   let expander, library_name_cookie =
     match lib_name with
     | None -> (expander, None)
@@ -394,23 +392,20 @@ let get_cookies ~loc ~expander ~lib_name libs =
       )
   in
   let+ cookies =
-    Memo.Build.List.concat_map libs ~f:(fun t ->
+    Memo.List.concat_map libs ~f:(fun t ->
         let info = Lib.info t in
         let kind = Lib_info.kind info in
         match kind with
-        | Normal -> Memo.Build.return []
+        | Normal -> Memo.return []
         | Ppx_rewriter { cookies } | Ppx_deriver { cookies } ->
-          Memo.Build.List.map
+          Memo.List.map cookies
             ~f:(fun { Lib_kind.Ppx_args.Cookie.name; value } ->
               let+ value = Expander.No_deps.expand_str expander value in
-              (name, (value, Lib.name t)))
-            cookies)
+              (name, (value, Lib.name t))))
   in
-  cookies
-  |> (fun l ->
-       match library_name_cookie with
-       | None -> l
-       | Some cookie -> cookie :: l)
+  (match library_name_cookie with
+  | None -> cookies
+  | Some cookie -> cookie :: cookies)
   |> String.Map.of_list_reducei
        ~f:(fun name ((val1, lib1) as res) (val2, lib2) ->
          if String.equal val1 val2 then res
@@ -427,26 +422,41 @@ let get_cookies ~loc ~expander ~lib_name libs =
          [ "--cookie"; sprintf "%s=%S" name value ])
   |> List.concat
 
-let ppx_driver_and_flags_internal sctx ~loc ~expander ~lib_name ~flags libs =
+let ppx_driver_and_flags_internal sctx ~dune_version ~loc ~expander ~lib_name
+    ~flags libs =
   let open Action_builder.O in
   let* flags =
-    Action_builder.List.map ~f:(Expander.expand_str expander) flags
+    if dune_version <= (3, 2) then
+      Action_builder.List.map ~f:(Expander.expand_str expander) flags
+    else
+      (* Allow list expansion *)
+      Action_builder.List.concat_map flags
+        ~f:(Expander.expand ~mode:Many expander)
+      |> Action_builder.map
+           ~f:
+             (List.map
+                ~f:(Value.to_string ~dir:(Path.build @@ Expander.dir expander)))
   in
   let+ cookies =
-    Action_builder.memo_build (get_cookies ~loc ~lib_name ~expander libs)
+    Action_builder.of_memo (get_cookies ~loc ~lib_name ~expander libs)
   in
-  let sctx = SC.host sctx in
-  (ppx_driver_exe sctx libs, flags @ cookies)
+  let ppx_driver_exe =
+    let ctx = Context.host (Super_context.context sctx) in
+    ppx_driver_exe ctx libs
+  in
+  (ppx_driver_exe, flags @ cookies)
 
 let ppx_driver_and_flags sctx ~lib_name ~expander ~scope ~loc ~flags pps =
   let open Action_builder.O in
-  let* libs = Resolve.Build.read (Lib.DB.resolve_pps (Scope.libs scope) pps) in
+  let* libs = Resolve.Memo.read (Lib.DB.resolve_pps (Scope.libs scope) pps) in
   let* exe, flags =
-    ppx_driver_and_flags_internal sctx ~loc ~expander ~lib_name ~flags libs
+    let dune_version = Scope.project scope |> Dune_project.dune_version in
+    ppx_driver_and_flags_internal sctx ~loc ~expander ~dune_version ~lib_name
+      ~flags libs
   in
-  let* libs = Resolve.Build.read (Lib.closure libs ~linking:true) in
+  let* libs = Resolve.Memo.read (Lib.closure libs ~linking:true) in
   let+ driver =
-    Action_builder.memo_build (Driver.select libs ~loc:(User_file (loc, pps)))
+    Action_builder.of_memo (Driver.select libs ~loc:(User_file (loc, pps)))
     >>= Resolve.read
   in
   (exe, driver, flags)
@@ -476,7 +486,7 @@ let promote_correction_with_target fn build ~suffix =
 
 let chdir action = Action_unexpanded.Chdir (workspace_root_var, action)
 
-let action_for_pp ~loc ~expander ~action ~src =
+let action_for_pp ~sandbox ~loc ~expander ~action ~src =
   let action = chdir action in
   let bindings =
     Pform.Map.singleton (Var Input_file) [ Value.Path (Path.build src) ]
@@ -486,27 +496,30 @@ let action_for_pp ~loc ~expander ~action ~src =
   Action_builder.path (Path.build src)
   >>> Action_unexpanded.expand_no_targets action ~loc ~expander ~deps:[]
         ~what:"preprocessing actions"
+  >>| Action.Full.add_sandbox sandbox
 
-let action_for_pp_with_target ~loc ~expander ~action ~src ~target =
-  let action = action_for_pp ~loc ~expander ~action ~src in
+let action_for_pp_with_target ~sandbox ~loc ~expander ~action ~src ~target =
+  let action = action_for_pp ~sandbox ~loc ~expander ~action ~src in
   Action_builder.with_stdout_to target action
 
 (* Generate rules for the dialect modules in [modules] and return a a new module
    with only OCaml sources *)
-let setup_dialect_rules sctx ~dir ~expander (m : Module.t) =
-  let open Memo.Build.O in
+let setup_dialect_rules sctx ~sandbox ~dir ~expander (m : Module.t) =
+  let open Memo.O in
   let ml = Module.ml_source m in
   let+ () =
     Module.iter m ~f:(fun ml_kind f ->
-        Memo.Build.Option.iter (Dialect.preprocess f.dialect ml_kind)
+        Memo.Option.iter
+          (Dialect.preprocess (Module.File.dialect f) ml_kind)
           ~f:(fun (loc, action) ->
-            let src = Path.as_in_build_dir_exn f.path in
+            let src = Path.as_in_build_dir_exn (Module.File.path f) in
             let dst =
               Option.value_exn (Module.file ml ~ml_kind)
               |> Path.as_in_build_dir_exn
             in
-            SC.add_rule sctx ~dir
-              (action_for_pp_with_target ~loc ~expander ~action ~src ~target:dst)))
+            Super_context.add_rule sctx ~dir
+              (action_for_pp_with_target ~sandbox ~loc ~expander ~action ~src
+                 ~target:dst)))
   in
   ml
 
@@ -520,31 +533,36 @@ let driver_flags expander ~corrected_suffix ~driver_flags ~standard =
   let expander = add_corrected_suffix_binding expander corrected_suffix in
   Expander.expand_and_eval_set expander driver_flags ~standard
 
-let lint_module sctx ~dir ~expander ~lint ~lib_name ~scope =
+let lint_module sctx ~sandbox ~dir ~expander ~lint ~lib_name ~scope =
   let open Action_builder.O in
   Staged.stage
     (let alias = Alias.lint ~dir in
-     let add_alias build = SC.add_alias_action sctx alias build ~dir in
+     let add_alias build =
+       Super_context.add_alias_action sctx alias build ~dir
+     in
      let lint =
        Module_name.Per_item.map lint ~f:(function
-         | Preprocess.No_preprocessing ->
-           fun ~source:_ ~ast:_ -> Memo.Build.return ()
+         | Preprocess.No_preprocessing -> fun ~source:_ ~ast:_ -> Memo.return ()
          | Future_syntax loc ->
            User_error.raise ~loc
              [ Pp.text "'compat' cannot be used as a linter" ]
          | Action (loc, action) ->
            fun ~source ~ast:_ ->
              Module.iter source ~f:(fun _ (src : Module.File.t) ->
-                 let src = Path.as_in_build_dir_exn src.path in
+                 let src = Path.as_in_build_dir_exn (Module.File.path src) in
                  add_alias ~loc:(Some loc)
-                   (action_for_pp ~loc ~expander ~action ~src))
+                   (action_for_pp ~sandbox ~loc ~expander ~action ~src))
          | Pps { loc; pps; flags; staged } ->
            if staged then
              User_error.raise ~loc
                [ Pp.text "Staged ppx rewriters cannot be used as linters." ];
            let corrected_suffix = ".lint-corrected" in
            let driver_and_flags =
-             Action_builder.memoize "ppx driver and flags"
+             Action_builder.memoize
+               ~cutoff:
+                 (Tuple.T3.equal Path.Build.equal (List.equal String.equal)
+                    (List.equal String.equal))
+               "ppx driver and flags"
                (let* () = Action_builder.return () in
                 let* exe, driver, flags =
                   ppx_driver_and_flags sctx ~expander ~loc ~lib_name ~flags
@@ -559,7 +577,7 @@ let lint_module sctx ~dir ~expander ~lint ~lib_name ~scope =
            in
            fun ~source ~ast ->
              Module.iter ast ~f:(fun ml_kind src ->
-                 add_alias ~loc:None
+                 add_alias ~loc:(Some loc)
                    (promote_correction ~suffix:corrected_suffix
                       (Path.as_in_build_dir_exn
                          (Option.value_exn (Module.file source ~ml_kind)))
@@ -571,7 +589,7 @@ let lint_module sctx ~dir ~expander ~lint ~lib_name ~scope =
                          (Ok (Path.build exe))
                          [ As args
                          ; Command.Ml_kind.ppx_driver_flag ml_kind
-                         ; Dep src.path
+                         ; Dep (Module.File.path src)
                          ; As flags
                          ]))))
      in
@@ -589,48 +607,58 @@ let make sctx ~dir ~expander ~lint ~preprocess ~preprocessor_deps
   let preprocessor_deps, sandbox =
     Dep_conf_eval.unnamed preprocessor_deps ~expander
   in
+  let sandbox =
+    match
+      Sandbox_config.equal Sandbox_config.no_special_requirements sandbox
+    with
+    | false -> sandbox
+    | true ->
+      let project = Scope.project scope in
+      let dune_version = Dune_project.dune_version project in
+      if dune_version >= (3, 3) then Sandbox_config.needs_sandboxing
+      else sandbox
+  in
   let preprocessor_deps =
     Action_builder.memoize "preprocessor deps" preprocessor_deps
   in
   let lint_module =
-    Staged.unstage (lint_module sctx ~dir ~expander ~lint ~lib_name ~scope)
+    Staged.unstage
+      (lint_module sctx ~sandbox ~dir ~expander ~lint ~lib_name ~scope)
   in
   let open Action_builder.O in
   Module_name.Per_item.map preprocess ~f:(fun pp ->
       match pp with
       | No_preprocessing ->
         fun m ~lint ->
-          let open Memo.Build.O in
-          let* ast = setup_dialect_rules sctx ~dir ~expander m in
-          let+ () =
-            Memo.Build.when_ lint (fun () -> lint_module ~ast ~source:m)
-          in
+          let open Memo.O in
+          let* ast = setup_dialect_rules sctx ~sandbox ~dir ~expander m in
+          let+ () = Memo.when_ lint (fun () -> lint_module ~ast ~source:m) in
           ast
       | Action (loc, action) ->
         fun m ~lint ->
-          let open Memo.Build.O in
+          let open Memo.O in
           let* ast =
             pped_module m ~f:(fun _kind src dst ->
                 let action =
-                  action_for_pp_with_target ~loc ~expander ~action ~src
+                  action_for_pp_with_target ~sandbox ~loc ~expander ~action ~src
                     ~target:dst
                 in
-                SC.add_rule sctx ~loc ~dir
+                Super_context.add_rule sctx ~loc ~dir
                   (let open Action_builder.With_targets.O in
-                  Action_builder.with_no_targets preprocessor_deps
-                  >>> action
-                  >>| Action.Full.add_sandbox sandbox))
-            >>= setup_dialect_rules sctx ~dir ~expander
-          in
-          let+ () =
-            Memo.Build.when_ lint (fun () -> lint_module ~ast ~source:m)
+                  Action_builder.with_no_targets preprocessor_deps >>> action))
+            >>= setup_dialect_rules sctx ~sandbox ~dir ~expander
           in
+          let+ () = Memo.when_ lint (fun () -> lint_module ~ast ~source:m) in
           ast
       | Pps { loc; pps; flags; staged } ->
         if not staged then
           let corrected_suffix = ".ppx-corrected" in
           let driver_and_flags =
-            Action_builder.memoize "ppx driver and flags"
+            Action_builder.memoize
+              ~cutoff:
+                (Tuple.T3.equal Path.Build.equal (List.equal String.equal)
+                   (List.equal String.equal))
+              "ppx driver and flags"
               (let* () = Action_builder.return () in
                let* exe, driver, flags =
                  ppx_driver_and_flags sctx ~expander ~loc ~lib_name ~flags
@@ -644,13 +672,11 @@ let make sctx ~dir ~expander ~lint ~preprocess ~preprocessor_deps
                (exe, ppx_flags, flags))
           in
           fun m ~lint ->
-            let open Memo.Build.O in
-            let* ast = setup_dialect_rules sctx ~dir ~expander m in
-            let* () =
-              Memo.Build.when_ lint (fun () -> lint_module ~ast ~source:m)
-            in
+            let open Memo.O in
+            let* ast = setup_dialect_rules sctx ~sandbox ~dir ~expander m in
+            let* () = Memo.when_ lint (fun () -> lint_module ~ast ~source:m) in
             pped_module ast ~f:(fun ml_kind src dst ->
-                SC.add_rule sctx ~loc ~dir
+                Super_context.add_rule sctx ~loc ~dir
                   (promote_correction_with_target ~suffix:corrected_suffix
                      (Path.as_in_build_dir_exn
                         (Option.value_exn (Module.file m ~ml_kind)))
@@ -673,7 +699,8 @@ let make sctx ~dir ~expander ~lint ~preprocess ~preprocessor_deps
                             >>| Action.Full.add_sandbox sandbox))))
         else
           let dash_ppx_flag =
-            Action_builder.memoize "ppx command"
+            Action_builder.memoize ~cutoff:(List.equal String.equal)
+              "ppx command"
               (let* () = Action_builder.return () in
                let* exe, driver, flags =
                  ppx_driver_and_flags sctx ~expander ~loc ~scope ~flags
@@ -690,7 +717,8 @@ let make sctx ~dir ~expander ~lint ~preprocess ~preprocessor_deps
                  List.map
                    (List.concat
                       [ [ Path.reach (Path.build exe)
-                            ~from:(Path.build (SC.context sctx).build_dir)
+                            ~from:
+                              (Path.build (Super_context.context sctx).build_dir)
                         ]
                       ; driver_flags
                       ; flags
@@ -702,20 +730,20 @@ let make sctx ~dir ~expander ~lint ~preprocess ~preprocessor_deps
           in
           let pp = Some (dash_ppx_flag, sandbox) in
           fun m ~lint ->
-            let open Memo.Build.O in
-            let* ast = setup_dialect_rules sctx ~dir ~expander m in
-            let+ () =
-              Memo.Build.when_ lint (fun () -> lint_module ~ast ~source:m)
-            in
+            let open Memo.O in
+            let* ast = setup_dialect_rules sctx ~sandbox ~dir ~expander m in
+            let+ () = Memo.when_ lint (fun () -> lint_module ~ast ~source:m) in
             Module.set_pp ast pp)
   |> Pp_spec.make
 
 let get_ppx_driver sctx ~loc ~expander ~scope ~lib_name ~flags pps =
   let open Action_builder.O in
-  let* libs = Resolve.Build.read (Lib.DB.resolve_pps (Scope.libs scope) pps) in
-  ppx_driver_and_flags_internal sctx ~loc ~expander ~lib_name ~flags libs
+  let* libs = Resolve.Memo.read (Lib.DB.resolve_pps (Scope.libs scope) pps) in
+  let dune_version = Scope.project scope |> Dune_project.dune_version in
+  ppx_driver_and_flags_internal sctx ~loc ~expander ~dune_version ~lib_name
+    ~flags libs
 
 let ppx_exe sctx ~scope pp =
-  let open Resolve.Build.O in
+  let open Resolve.Memo.O in
   let+ libs = Lib.DB.resolve_pps (Scope.libs scope) [ (Loc.none, pp) ] in
   ppx_driver_exe sctx libs
