@@ -1,8 +1,6 @@
 (** Preprocessing of OCaml source files *)
-open! Dune_engine
 
-open! Stdune
-open! Import
+open Import
 
 val make :
      Super_context.t
@@ -14,7 +12,7 @@ val make :
   -> instrumentation_deps:Dep_conf.t list
   -> lib_name:Lib_name.Local.t option
   -> scope:Scope.t
-  -> Pp_spec.t Memo.Build.t
+  -> Pp_spec.t Memo.t
 
 (** Get a path to a cached ppx driver with some extra flags for cookies. *)
 val get_ppx_driver :
@@ -27,12 +25,13 @@ val get_ppx_driver :
   -> (Loc.t * Lib_name.t) list
   -> (Path.Build.t * string list) Action_builder.t
 
-val gen_rules : Super_context.t -> string list -> unit Memo.Build.t
+val gen_rules : Super_context.t -> string list -> unit Memo.t
 
 val chdir : Action_unexpanded.t -> Action_unexpanded.t
 
 val action_for_pp_with_target :
-     loc:Loc.t
+     sandbox:Sandbox_config.t
+  -> loc:Loc.t
   -> expander:Expander.t
   -> action:Action_unexpanded.t
   -> src:Path.Build.t
@@ -40,4 +39,4 @@ val action_for_pp_with_target :
   -> Action.Full.t Action_builder.With_targets.t
 
 val ppx_exe :
-  Super_context.t -> scope:Scope.t -> Lib_name.t -> Path.Build.t Resolve.Build.t
+  Context.t -> scope:Scope.t -> Lib_name.t -> Path.Build.t Resolve.Memo.t
