@@ -1,22 +1,19 @@
-open! Dune_engine
-open! Dune_engine.Import
-open! Stdune
+open Import
 open Dune_file
-module SC = Super_context
-open Memo.Build.O
+open Memo.O
 
 module Alias_rules = struct
   let add sctx ~alias ~loc build =
     let dir = Alias.dir alias in
-    SC.add_alias_action sctx alias ~dir ~loc build
+    Super_context.add_alias_action sctx alias ~dir ~loc build
 
   let add_empty sctx ~loc ~alias =
     let action = Action_builder.return (Action.Full.make Action.empty) in
     add sctx ~loc ~alias action
 end
 
-let check_filename ~kind =
-  let not_in_dir ~error_loc s =
+let check_filename =
+  let not_in_dir ~kind ~error_loc s =
     User_error.raise ~loc:error_loc
       [ (match kind with
         | Targets_spec.Kind.File ->
@@ -25,37 +22,35 @@ let check_filename ~kind =
           Pp.textf "Directory targets must have exactly one path component.")
       ]
   in
-  fun ~error_loc ~dir -> function
+  fun ~kind ~error_loc ~dir -> function
     | Value.String ("." | "..") ->
       User_error.raise ~loc:error_loc
         [ Pp.text "'.' and '..' are not valid targets" ]
     | String s ->
       if Filename.dirname s <> Filename.current_dir_name then
-        not_in_dir ~error_loc s;
+        not_in_dir ~kind ~error_loc s;
       Path.Build.relative ~error_loc dir s
     | Path p -> (
       match Option.equal Path.equal (Path.parent p) (Some (Path.build dir)) with
       | true -> Path.as_in_build_dir_exn p
-      | false -> not_in_dir ~error_loc (Path.to_string p))
-    | Dir p -> not_in_dir ~error_loc (Path.to_string p)
+      | false -> not_in_dir ~kind ~error_loc (Path.to_string p))
+    | Dir p -> not_in_dir ~kind ~error_loc (Path.to_string p)
 
 type rule_kind =
-  | Alias_only of Alias.Name.t
-  | Alias_with_targets of Alias.Name.t * Path.Build.t
+  | Aliases_only of Alias.Name.t list
+  | Aliases_with_targets of Alias.Name.t list * Path.Build.t
   | No_alias
 
 let rule_kind ~(rule : Rule.t) ~(action : _ Action_builder.With_targets.t) =
-  match rule.alias with
-  | None -> No_alias
-  | Some alias -> (
+  match rule.aliases with
+  | [] -> No_alias
+  | aliases -> (
     match Targets.head action.targets with
-    | None -> Alias_only alias
-    | Some target -> Alias_with_targets (alias, target))
+    | None -> Aliases_only aliases
+    | Some target -> Aliases_with_targets (aliases, target))
 
 let interpret_and_add_locks ~expander locks action =
-  let+ locks =
-    Memo.Build.List.map locks ~f:(Expander.No_deps.expand_path expander)
-  in
+  let+ locks = Expander.expand_locks expander ~base:`Of_expander locks in
   match locks with
   | [] -> action
   | _ ->
@@ -66,24 +61,25 @@ let add_user_rule sctx ~dir ~(rule : Rule.t)
     ~(action : _ Action_builder.With_targets.t) ~expander =
   let* build = interpret_and_add_locks ~expander rule.locks action.build in
   let action = { action with Action_builder.With_targets.build } in
-  SC.add_rule_get_targets sctx ~dir ~mode:rule.mode ~loc:rule.loc action
+  Super_context.add_rule_get_targets sctx ~dir ~mode:rule.mode ~loc:rule.loc
+    action
 
 let user_rule sctx ?extra_bindings ~dir ~expander (rule : Rule.t) =
   Expander.eval_blang expander rule.enabled_if >>= function
-  | false -> (
-    match rule.alias with
-    | None -> Memo.Build.return None
-    | Some name ->
-      let alias = Alias.make ~dir name in
-      let+ () = Alias_rules.add_empty sctx ~alias ~loc:(Some rule.loc) in
-      None)
+  | false ->
+    let aliases = List.map rule.aliases ~f:(Alias.make ~dir) in
+    let+ () =
+      Memo.parallel_iter aliases ~f:(fun alias ->
+          Alias_rules.add_empty sctx ~loc:(Some rule.loc) ~alias)
+    in
+    None
   | true -> (
     let* targets =
       match rule.targets with
-      | Infer -> Memo.Build.return Targets_spec.Infer
+      | Infer -> Memo.return Targets_spec.Infer
       | Static { targets; multiplicity } ->
         let+ targets =
-          Memo.Build.List.concat_map targets ~f:(fun (target, kind) ->
+          Memo.List.concat_map targets ~f:(fun (target, kind) ->
               let error_loc = String_with_vars.loc target in
               (match multiplicity with
               | One ->
@@ -100,11 +96,10 @@ let user_rule sctx ?extra_bindings ~dir ~expander (rule : Rule.t) =
       | None -> expander
       | Some bindings -> Expander.add_bindings expander ~bindings
     in
-    let action =
+    let* (action : _ Action_builder.With_targets.t) =
       Action_unexpanded.expand (snd rule.action) ~loc:(fst rule.action)
         ~expander ~deps:rule.deps ~targets ~targets_dir:dir
     in
-    let* action = action in
     let action =
       if rule.patch_back_source_tree then
         Action_builder.With_targets.map action ~f:(fun action ->
@@ -125,18 +120,22 @@ let user_rule sctx ?extra_bindings ~dir ~expander (rule : Rule.t) =
     | No_alias ->
       let+ targets = add_user_rule sctx ~dir ~rule ~action ~expander in
       Some targets
-    | Alias_with_targets (alias, alias_target) ->
+    | Aliases_with_targets (aliases, alias_target) ->
       let* () =
-        let alias = Alias.make alias ~dir in
-        Rules.Produce.Alias.add_deps alias
-          (Action_builder.path (Path.build alias_target))
+        let aliases = List.map ~f:(Alias.make ~dir) aliases in
+        Memo.parallel_iter aliases ~f:(fun alias ->
+            Rules.Produce.Alias.add_deps alias
+              (Action_builder.path (Path.build alias_target)))
       in
       let+ targets = add_user_rule sctx ~dir ~rule ~action ~expander in
       Some targets
-    | Alias_only name ->
-      let alias = Alias.make ~dir name in
+    | Aliases_only aliases ->
+      let aliases = List.map ~f:(Alias.make ~dir) aliases in
       let* action = interpret_and_add_locks ~expander rule.locks action.build in
-      let+ () = Alias_rules.add sctx ~alias ~loc:(Some rule.loc) action in
+      let+ () =
+        Memo.parallel_iter aliases ~f:(fun alias ->
+            Alias_rules.add sctx ~alias ~loc:(Some rule.loc) action)
+      in
       None)
 
 let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
@@ -144,7 +143,7 @@ let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
   let* glob_in_src =
     let+ src_glob = Expander.No_deps.expand_str expander def.files in
     if Filename.is_relative src_glob then
-      Path.Source.relative src_dir src_glob ~error_loc:loc |> Path.source
+      Path.relative (Path.source src_dir) src_glob ~error_loc:loc
     else
       let since = (2, 7) in
       if def.syntax_version < since then
@@ -163,15 +162,24 @@ let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
            (Path.to_string_maybe_quoted glob_in_src)
            (Path.Source.to_string_maybe_quoted src_dir));
   let src_in_src = Path.parent_exn glob_in_src in
-  let pred =
-    Path.basename glob_in_src |> Glob.of_string_exn loc |> Glob.to_pred
-  in
-  let* exists =
+  let glob = Path.basename glob_in_src |> Glob.of_string_exn loc in
+  let src_in_build =
     match Path.as_in_source_tree src_in_src with
-    | None -> Memo.Build.return (Path.exists src_in_src)
-    | Some src_in_src -> Source_tree.dir_exists src_in_src
+    | None -> src_in_src
+    | Some src_in_src ->
+      let context = Super_context.context sctx in
+      Path.Build.append_source context.build_dir src_in_src |> Path.build
   in
-  if not exists then
+  let* exists_or_generated =
+    match src_in_src with
+    | In_build_dir _ -> assert false
+    | External ext -> Fs_memo.dir_exists (External ext)
+    | In_source_tree src_in_src -> (
+      Source_tree.dir_exists src_in_src >>= function
+      | true -> Memo.return true
+      | false -> Load_rules.is_under_directory_target src_in_build)
+  in
+  if not exists_or_generated then
     User_error.raise ~loc
       [ Pp.textf "Cannot find directory: %s" (Path.to_string src_in_src) ];
   if Path.equal src_in_src (Path.source src_dir) then
@@ -181,29 +189,21 @@ let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
            <dir> is not the current directory."
       ];
   (* add rules *)
-  let src_in_build =
-    match Path.as_in_source_tree src_in_src with
-    | None -> src_in_src
-    | Some src_in_src ->
-      let context = Super_context.context sctx in
-      Path.Build.append_source context.build_dir src_in_src |> Path.build
-  in
   let* files =
-    Build_system.eval_pred (File_selector.create ~dir:src_in_build pred)
+    Build_system.eval_pred (File_selector.of_glob ~dir:src_in_build glob)
   in
   (* CR-someday amokhov: We currently traverse the set [files] twice: first, to
      add the corresponding rules, and then to convert the files to [targets]. To
-     do only one traversal we need [Memo.Build.parallel_map_set]. *)
+     do only one traversal we need [Memo.parallel_map_set]. *)
   let* () =
-    Memo.Build.parallel_iter_set
+    Memo.parallel_iter_set
       (module Path.Set)
       files
       ~f:(fun file_src ->
         let basename = Path.basename file_src in
         let file_dst = Path.Build.relative dir basename in
-        SC.add_rule sctx ~loc ~dir ~mode:def.mode
-          ((if def.add_line_directive then
-            Action_builder.copy_and_add_line_directive
+        Super_context.add_rule sctx ~loc ~dir ~mode:def.mode
+          ((if def.add_line_directive then Copy_line_directive.builder
            else Action_builder.copy)
              ~src:file_src ~dst:file_dst))
   in
@@ -215,7 +215,7 @@ let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
         Path.build file_dst)
   in
   let+ () =
-    Memo.Build.Option.iter def.alias ~f:(fun alias ->
+    Memo.Option.iter def.alias ~f:(fun alias ->
         let alias = Alias.make alias ~dir in
         Rules.Produce.Alias.add_deps alias (Action_builder.path_set targets))
   in
@@ -224,7 +224,7 @@ let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
 let copy_files sctx ~dir ~expander ~src_dir (def : Copy_files.t) =
   Expander.eval_blang expander def.enabled_if >>= function
   | true -> copy_files sctx ~dir ~expander ~src_dir def
-  | false -> Memo.Build.return Path.Set.empty
+  | false -> Memo.return Path.Set.empty
 
 let alias sctx ?extra_bindings ~dir ~expander (alias_conf : Alias_conf.t) =
   let alias = Alias.make ~dir alias_conf.name in
