@@ -1,8 +1,5 @@
-open! Dune_engine
-open! Stdune
 open Import
-open! No_io
-open Memo.Build.O
+open Memo.O
 
 let exe_name = "utop"
 
@@ -31,14 +28,14 @@ module Libs_and_ppxs =
     end))
 
 module Source_tree_map_reduce =
-  Source_tree.Dir.Make_map_reduce (Memo.Build) (Libs_and_ppxs)
+  Source_tree.Dir.Make_map_reduce (Memo) (Libs_and_ppxs)
 
 let libs_and_ppx_under_dir sctx ~db ~dir =
   (match Path.drop_build_context dir with
-  | None -> Memo.Build.return None
+  | None -> Memo.return None
   | Some dir -> Source_tree.find_dir dir)
   >>= function
-  | None -> Memo.Build.return ([], [])
+  | None -> Memo.return ([], [])
   | Some dir ->
     Source_tree_map_reduce.map_reduce dir
       ~traverse:{ data_only = false; vendored = true; normal = true }
@@ -47,14 +44,14 @@ let libs_and_ppx_under_dir sctx ~db ~dir =
           Path.Build.append_source (Super_context.context sctx).build_dir
             (Source_tree.Dir.path dir)
         in
-        match Super_context.stanzas_in sctx ~dir with
-        | None -> Memo.Build.return Libs_and_ppxs.empty
-        | Some (d : _ Dir_with_dune.t) ->
-          Memo.Build.List.fold_left d.data ~init:Libs_and_ppxs.empty
+        Only_packages.stanzas_in_dir dir >>= function
+        | None -> Memo.return Libs_and_ppxs.empty
+        | Some (d : Dune_file.t) ->
+          Memo.List.fold_left d.stanzas ~init:Libs_and_ppxs.empty
             ~f:(fun (acc, pps) -> function
             | Dune_file.Library l -> (
               let+ lib =
-                let open Memo.Build.O in
+                let open Memo.O in
                 let+ resolve =
                   Lib.DB.resolve_when_exists db
                     (l.buildable.loc, Dune_file.Library.best_name l)
@@ -90,16 +87,17 @@ let libs_and_ppx_under_dir sctx ~db ~dir =
                 else (acc, pps))
             | Dune_file.Executables exes -> (
               let* libs =
-                let open Memo.Build.O in
+                let open Memo.O in
                 let* compile_info =
-                  let project = Scope.project d.scope in
+                  let* scope = Scope.DB.find_by_dir dir in
+                  let project = Scope.project scope in
                   let dune_version = Dune_project.dune_version project in
                   let+ pps =
-                    Resolve.Build.read_memo_build
+                    Resolve.Memo.read_memo
                       (Preprocess.Per_module.with_instrumentation
                          exes.buildable.preprocess
                          ~instrumentation_backend:
-                           (Lib.DB.instrumentation_backend (Scope.libs d.scope)))
+                           (Lib.DB.instrumentation_backend (Scope.libs scope)))
                     >>| Preprocess.Per_module.pps
                   in
                   Lib.DB.resolve_user_written_deps_for_exes db exes.names
@@ -112,14 +110,14 @@ let libs_and_ppx_under_dir sctx ~db ~dir =
                 Resolve.peek available
               in
               match libs with
-              | Error () -> Memo.Build.return (acc, pps)
+              | Error () -> Memo.return (acc, pps)
               | Ok libs ->
-                Memo.Build.List.fold_left libs ~init:(acc, pps)
+                Memo.List.fold_left libs ~init:(acc, pps)
                   ~f:(fun (acc, pps) lib ->
                     let info = Lib.info lib in
                     match Lib_info.kind info with
                     | Lib_kind.Ppx_rewriter _ | Ppx_deriver _ ->
-                      Memo.Build.return
+                      Memo.return
                         ( Appendable_list.( @ )
                             (Appendable_list.singleton lib)
                             acc
@@ -128,21 +126,21 @@ let libs_and_ppx_under_dir sctx ~db ~dir =
                                (Lib_info.loc info, Lib_info.name info))
                             pps )
                     | Normal ->
-                      Memo.Build.return
+                      Memo.return
                         ( Appendable_list.( @ )
                             (Appendable_list.singleton lib)
                             acc
                         , pps )))
-            | _ -> Memo.Build.return (acc, pps)))
+            | _ -> Memo.return (acc, pps)))
     >>| fun (libs, pps) ->
     (Appendable_list.to_list libs, Appendable_list.to_list pps)
 
 let libs_under_dir sctx ~db ~dir = libs_and_ppx_under_dir sctx ~db ~dir >>| fst
 
 let setup sctx ~dir =
-  let open Memo.Build.O in
+  let open Memo.O in
   let* expander = Super_context.expander sctx ~dir in
-  let scope = Super_context.find_scope_by_dir sctx dir in
+  let* scope = Scope.DB.find_by_dir dir in
   let db = Scope.libs scope in
   let* libs, pps = libs_and_ppx_under_dir sctx ~db ~dir:(Path.build dir) in
   let pps =
@@ -160,7 +158,7 @@ let setup sctx ~dir =
   let loc = Toplevel.Source.loc source in
   let* modules = Toplevel.Source.modules source preprocessing in
   let requires =
-    let open Resolve.Build.O in
+    let open Resolve.Memo.O in
     (loc, Lib_name.of_string "utop")
     |> Lib.DB.resolve db
     >>| (fun utop -> utop :: libs)
