@@ -1,11 +1,13 @@
 module Config = Config
 module Log = Log
-module Ml_kind = Ml_kind
 module Persistent = Persistent
 module Report_error = Report_error
 module Stringlike = Stringlike
 module Stringlike_intf = Stringlike_intf
 module Value = Value
+module Build_path_prefix_map = Build_path_prefix_map0
+module Flock = Flock
+module Global_lock = Global_lock
 open Stdune
 
 let xdg =
